/**
 * 
 */
package CloneTransforming;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;

/**
 * @author Hoan Nguyen
 *
 */
public class ASTNodeWrapper {
	public static ASTNode createAST(String code)
	{
		ASTParser parser = ASTParser.newParser(AST.JLS3);
    	parser.setSource(code.toCharArray());
    	parser.setKind(ASTParser.K_COMPILATION_UNIT);
    	ASTNode node = parser.createAST(null);
    	if(isParsed(node))
    	{
    		//System.out.println("Unit");
    		return node;
    	}
    	
    	parser = ASTParser.newParser(AST.JLS3);
    	parser.setSource(code.toCharArray());
    	parser.setKind(ASTParser.K_CLASS_BODY_DECLARATIONS);
    	node = parser.createAST(null);
    	if(isParsed(node))
    	{
    		//System.out.println("Body");
    		return node;
    	}

    	parser = ASTParser.newParser(AST.JLS3);
    	parser.setSource(code.toCharArray());
    	parser.setKind(ASTParser.K_STATEMENTS);
    	node = parser.createAST(null);
    	if(isParsed(node))
    	{
    		//System.out.println("Statements");
    		code = "public void MISSING(){\n" + code + "\n}";
    		parser = ASTParser.newParser(AST.JLS3);
        	parser.setSource(code.toCharArray());
        	parser.setKind(ASTParser.K_CLASS_BODY_DECLARATIONS);
        	node = parser.createAST(null);
    		return node;
    	}
    	else if (code.trim().endsWith(")"))
    	{
    		code = "public void MISSING(){\n" + code + ";\n}";
    		parser = ASTParser.newParser(AST.JLS3);
        	parser.setSource(code.toCharArray());
        	parser.setKind(ASTParser.K_CLASS_BODY_DECLARATIONS);
        	node = parser.createAST(null);
    		return node;
    	}
    	System.err.println("Failed to create AST");
    	return null;
	}
	public static String getContent(ASTNode node)
	{
		//System.out.println(node.getClass().getSimpleName() + "\t" + node);
		switch (node.getNodeType())
		{
			case ASTNode.COMPILATION_UNIT:
				//System.out.println("Unit");
	    		CompilationUnit cu = (CompilationUnit)node;
	    		TypeDeclaration type = (TypeDeclaration)(cu.types().get(0));
	    		if(type.getName().toString().equals("MISSING"))
	    		{
	    			if(type.getMethods() != null && type.getMethods().length > 0)
	    			{
		    			MethodDeclaration method = type.getMethods()[0];
						if(method.getName().toString().equals("MISSING"))
						{
							Block block = method.getBody();
							String content = "";
							for(int i = 0; i < block.statements().size(); i++)
								content += block.statements().get(i).toString();
							return content;
						}
						return method.toString();
	    			}
	    			else if(type.getFields() != null && type.getFields().length > 0)
	    			{
	    				FieldDeclaration field = type.getFields()[0];
	    				return field.toString();
	    			}
	    		}
				return cu.toString();
			case ASTNode.TYPE_DECLARATION:
				//System.out.println("Class");
	    		type = (TypeDeclaration)node;
				if(type.getName().toString().equals("MISSING"))
				{
					//return type.bodyDeclarations().get(0).toString();
					if(type.getMethods() != null && type.getMethods().length > 0)
					{
						MethodDeclaration method = type.getMethods()[0];
						if(method.getName().toString().equals("MISSING"))
						{
							Block block = method.getBody();
							String content = "";
							for(int i = 0; i < block.statements().size(); i++)
								content += block.statements().get(i).toString();
							return content;
						}
						return method.toString();
					}
	    			else if(type.getFields() != null && type.getFields().length > 0)
	    			{
	    				FieldDeclaration field = type.getFields()[0];
	    				return field.toString();
	    			}
				}
				return type.toString();
			case ASTNode.BLOCK:
				//System.out.println("Block");
	    		Block block = (Block)node;
				String content = "";
				for(int i = 0; i < block.statements().size(); i++)
					content += block.statements().get(i).toString();
				return content;
			case ASTNode.METHOD_DECLARATION:
				//System.out.println("Block");
				MethodDeclaration method = (MethodDeclaration)node;
				block = method.getBody();
				content = "";
				for(int i = 0; i < block.statements().size(); i++)
					content += block.statements().get(i).toString();
				return content;
			default: return node.toString();
		}
	}
	private static boolean isParsed(ASTNode node)
	{
		switch (node.getNodeType())
		{
			case ASTNode.COMPILATION_UNIT:
				CompilationUnit cu = (CompilationUnit)node;
				if(cu.types() != null && cu.types().size() > 0)
					return true;
				return false;
			case ASTNode.TYPE_DECLARATION:
				TypeDeclaration type = (TypeDeclaration)node;
				if(type.bodyDeclarations() != null && type.bodyDeclarations().size() > 0)
					return true;
				return false;
			case ASTNode.BLOCK:
				Block block = (Block)node;
				if(block.statements() != null && block.statements().size() > 0)
					return true;
				return false;
			default: return false;
		}
	}
}
