/**
 * 
 */
package CloneTransforming;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Stack;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.BooleanLiteral;
import org.eclipse.jdt.core.dom.CharacterLiteral;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.NullLiteral;
import org.eclipse.jdt.core.dom.NumberLiteral;
import org.eclipse.jdt.core.dom.PostfixExpression;
import org.eclipse.jdt.core.dom.PrefixExpression;
import org.eclipse.jdt.core.dom.PrimitiveType;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.StructuralPropertyDescriptor;
import org.eclipse.jdt.core.dom.TypeLiteral;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;

import cleman.Fragment;
import cleman.Main;

import repositoryConnecting.SVNConnector;
import sun.rmi.server.*;
//import sun.rmi.server.Util;
import treeMapping.LineDiff;
import treeMapping.MapSourceFile;
import treeMapping.MapVisitor;
import treeMapping.TreeMapper;
import util.StringProcessor;

/**
 * @author Hoan Anh Nguyen
 *
 */
public class CloneMapper {
	public static final String propertyCloneMap = "cMap", propertySyncMap = "sMap";
	public static final String propertyCloneMapStatus = "cMS";
	/*public static final String propertyStatus = "status";
	public static final int STATUS_UNCHANGED = 0;
	public static final int STATUS_MODIFIED = 1;
	public static final int STATUS_DELETED = 3;
	public static final int STATUS_ADDED = 4;
	public static final int STATUS_MOVED = 5;
	public static final int STATUS_RELABELED = 6;
	public static final int CHANGE_OK = 0;
	public static final int CHANGE_INCONSISTENT_STRUCTURE = 1;
	public static final int CHANGE_INCONSISTENT_NAMING = 2;
	public static final int CHANGE_INCONSISTENT_LABEL = 2;
	public static final int CHANGE_CONFLICT = 4;*/
	public ASTNode ast1M, ast2M, ast1N, ast2N, astSync12;
	public HashMap<ASTNode, ArrayList<ASTNode>> tree1M, tree2M, tree1N, tree2N, treeSync12;
	public HashMap<String, String> naming = new HashMap<String, String>(), tokens = new HashMap<String, String>();
	private HashMap<String, ArrayList<String>> namingPosition1M = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> namingPosition2M = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> namingPosition1N = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<String>> namingPosition2N = new HashMap<String, ArrayList<String>>();
	private HashMap<String, ArrayList<ASTNode>> namingNode1M = new HashMap<String, ArrayList<ASTNode>>();
	private HashMap<String, ArrayList<ASTNode>> namingNode2M = new HashMap<String, ArrayList<ASTNode>>();
	private HashMap<String, ArrayList<ASTNode>> namingNode1N = new HashMap<String, ArrayList<ASTNode>>();
	private HashMap<String, ArrayList<ASTNode>> namingNode2N = new HashMap<String, ArrayList<ASTNode>>();
	
	private int lastNode = 0, sid = -1;	//for end token of vector
	private Stack<Integer> stackCurrentNode = new Stack<Integer>();	//for start token of vector
	
	public CloneMapper(Fragment f1M, Fragment f2M, long r1, long r2)
	{
		SVNConnector svnConnector = new SVNConnector(Main.svnURL, Main.repUserName, Main.repPassword);
		svnConnector.connect();
		
		String fileName1 = f1M.getSourceFile().fileLocation, fileName2 = f2M.getSourceFile().fileLocation;
		
		MapSourceFile file1M = new MapSourceFile(svnConnector.getFile(fileName1, r1));
		MapSourceFile file1N = new MapSourceFile(svnConnector.getFile(fileName1, r2));
		LineDiff lineDiff = new LineDiff(file1M, file1N);
		ArrayList<ArrayList<Integer>> diff1 = lineDiff.doDiffO();
		
		file1M.setFilePath(fileName1);
		file1M.buildBlockCommentLines();
		file1M.setUnchangedLines(diff1.get(0));
		
		file1N.setFilePath(fileName1);
		file1N.buildBlockCommentLines();
		file1N.setUnchangedLines(diff1.get(1));
		
		TreeMapper mapper1 = new TreeMapper(file1M, file1N);
		mapper1.map();
		
		MapSourceFile file2M = new MapSourceFile(svnConnector.getFile(fileName2, r1));
		MapSourceFile file2N = new MapSourceFile(svnConnector.getFile(fileName2, r2));
		lineDiff = new LineDiff(file2M, file2N);
		ArrayList<ArrayList<Integer>> diff2 = lineDiff.doDiffO();
		
		file2M.setFilePath(fileName2);
		file2M.buildBlockCommentLines();
		file2M.setUnchangedLines(diff2.get(0));
		
		file2N.setFilePath(fileName2);
		file2N.buildBlockCommentLines();
		file2N.setUnchangedLines(diff2.get(1));
		
		TreeMapper mapper2 = new TreeMapper(file2M, file2N);
		mapper2.map();
		
		this.tree1M = file1M.getTree(); this.tree2M = file2M.getTree();
		this.tree1N = file1N.getTree(); this.tree2N = file2N.getTree();
		this.ast1M = getASTNode(f1M, tree1M, file1M.getAst());
		this.ast2M = getASTNode(f2M, tree2M, file2M.getAst());
		if(this.ast1M == null || this.ast2M == null)
		{
			System.err.println("Error in CloneMapping");
			System.err.println("f1M: " + f1M.getSourceFile().fileLocation + "\n" + f1M);
			System.err.println("f2M: " + f2M.getSourceFile().fileLocation + "\n" + f2M);
		}
		this.ast1N = (ASTNode)this.ast1M.getProperty(TreeMapper.propertyRevisionMap);
		this.ast2N = (ASTNode)this.ast2M.getProperty(TreeMapper.propertyRevisionMap);
	}
	public CloneMapper(ASTNode ast1, ASTNode ast2)
	{
		this.ast1M = ast1;
		this.ast2M = ast2;
		TreeVisitor treeVisitor = new TreeVisitor(ast1);
		ast1.accept(treeVisitor);
		this.tree1M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(ast2);
		ast2.accept(treeVisitor);
		this.tree2M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
	}
	public CloneMapper(ASTNode ast1, ASTNode ast2, HashMap<ASTNode, ArrayList<ASTNode>> tree1, HashMap<ASTNode, ArrayList<ASTNode>> tree2)
	{
		this.ast1M = ast1;
		this.ast2M = ast2;
		this.tree1M = tree1;
		this.tree2M = tree2;
	}
	/*public CloneMapper(ASTNode ast1M, ASTNode ast2M, ASTNode ast1N, ASTNode ast2N)
	{
		this.ast1M = ast1M;
		this.ast2M = ast2M;
		this.ast1N = ast1N;
		this.ast2N = ast2N;
		TreeVisitor treeVisitor = new TreeVisitor(ast1M);
		ast1M.accept(treeVisitor);
		this.tree1M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(ast2M);
		ast2M.accept(treeVisitor);
		this.tree2M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(ast1N);
		ast1N.accept(treeVisitor);
		this.tree1N = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		if(ast2N != null)
		{
			treeVisitor = new TreeVisitor(ast2N);
			ast2N.accept(treeVisitor);
			this.tree2N = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		}
	}*/
	public CloneMapper(ASTNode ast1M, ASTNode ast2M, ASTNode ast1N, ASTNode ast2N)
	{
		MapSourceFile file1M = new MapSourceFile(ast1M.toString());
		MapSourceFile file1N = new MapSourceFile(ast1N.toString());
		MapSourceFile file2M = new MapSourceFile(ast2M.toString());
		MapSourceFile file2N = new MapSourceFile(ast2N.toString());
		file1M.buildBlockCommentLines();
		file1N.buildBlockCommentLines();
		file2M.buildBlockCommentLines();
		file2N.buildBlockCommentLines();
		
		LineDiff lineDiff = new LineDiff(file1M, file2M);
		ArrayList<ArrayList<Integer>> diff = lineDiff.doDiffO();
		file1M.setUnchangedLines(diff.get(0));
		file2M.setUnchangedLines(diff.get(1));
		TreeMapper mapper = new TreeMapper(file1M, file2M);
		mapper.map();
		CloneMapVisitor cloneMapVisitor = new CloneMapVisitor();
		file1M.getAst().accept(cloneMapVisitor);
		cloneMapVisitor = new CloneMapVisitor();
		file2M.getAst().accept(cloneMapVisitor);
		
		lineDiff = new LineDiff(file1M, file1N);
		ArrayList<ArrayList<Integer>> diff1 = lineDiff.doDiffO();
		//file1M.setFilePath(fileName1);
		file1M.setUnchangedLines(diff1.get(0));
		//file1N.setFilePath(fileName1);
		file1N.setUnchangedLines(diff1.get(1));
		TreeMapper mapper1 = new TreeMapper(file1M, file1N);
		mapper1.map();
		
		lineDiff = new LineDiff(file2M, file2N);
		ArrayList<ArrayList<Integer>> diff2 = lineDiff.doDiffO();
		//file2M.setFilePath(fileName2);
		file2M.setUnchangedLines(diff2.get(0));
		//file2N.setFilePath(fileName2);
		file2N.setUnchangedLines(diff2.get(1));
		TreeMapper mapper2 = new TreeMapper(file2M, file2N);
		mapper2.map();
		
		this.ast1M = file1M.getAst();
		this.ast2M = file2M.getAst();
		this.ast1N = file1N.getAst();
		this.ast2N = file2N.getAst();
		
		TreeVisitor treeVisitor = new TreeVisitor(this.ast1M);
		this.ast1M.accept(treeVisitor);
		this.tree1M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(this.ast2M);
		this.ast2M.accept(treeVisitor);
		this.tree2M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(this.ast1N);
		this.ast1N.accept(treeVisitor);
		this.tree1N = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		if(this.ast2N != null)
		{
			treeVisitor = new TreeVisitor(this.ast2N);
			this.ast2N.accept(treeVisitor);
			this.tree2N = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		}
	}
	private class CloneMapVisitor extends ASTVisitor {
		public void preVisit(ASTNode node)
		{
			node.setProperty(propertyCloneMap, node.getProperty(TreeMapper.propertyRevisionMap));
			node.setProperty(TreeMapper.propertyRevisionMap, null);
			node.setProperty(propertyCloneMapStatus, node.getProperty(MapVisitor.propertyStatus));
			node.setProperty(MapVisitor.propertyStatus, null);
		}
	}
	public void setAst1M(ASTNode ast1) {
		this.ast1M = ast1;
	}
	public void setAst2M(ASTNode ast2) {
		this.ast2M = ast2;
	}
	public void setAst1N(ASTNode ast1N) {
		this.ast1N = ast1N;
	}
	public ASTNode getAst2N() {
		return ast2N;
	}
	public void setTree1M(HashMap<ASTNode, ArrayList<ASTNode>> tree1) {
		this.tree1M = tree1;
	}
	public void setTree2M(HashMap<ASTNode, ArrayList<ASTNode>> tree2) {
		this.tree2M = tree2;
	}
	public void setTree1N(HashMap<ASTNode, ArrayList<ASTNode>> tree1N) {
		this.tree1N = tree1N;
	}
	public void setTree2N(HashMap<ASTNode, ArrayList<ASTNode>> tree2N) {
		this.tree2N = tree2N;
	}
	private ASTNode getASTNode(Fragment fragment, HashMap<ASTNode, ArrayList<ASTNode>> tree, ASTNode root)
	{
		/*this.sid = 0;
		this.lastNode = 0;
		this.stackCurrentNode = new Stack<Integer>();
		return getASTNode(fragment.getSourceFile().fragments.indexOf(fragment.getId()), tree, root);*/
		SequenceIDVisitor sidVisitor = new SequenceIDVisitor(fragment.getSourceFile().fragments.indexOf(fragment.getId()));
		root.accept(sidVisitor);
		return sidVisitor.node;
	}
	
	public boolean map()
	{
		if(tree1M == null)
		{
			TreeVisitor treeVisitor = new TreeVisitor(ast1M);
			ast1M.accept(treeVisitor);
			tree1M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		}
		if(tree2M == null)
		{
			TreeVisitor treeVisitor = new TreeVisitor(ast2M);
			ast2M.accept(treeVisitor);
			tree2M = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		}
		return map(ast1M, ast2M, tree1M, tree2M);
	}
	public boolean mapClone(ASTNode ast1, ASTNode ast2)
	{
		MapSourceFile file1 = new MapSourceFile(ast1);
		MapSourceFile file2 = new MapSourceFile(ast2);
		LineDiff lineDiff = new LineDiff(file1, file2);
		ArrayList<ArrayList<Integer>> diff = lineDiff.doDiffO();
		
		file1.buildBlockCommentLines();
		file1.setUnchangedLines(diff.get(0));
		
		file2.buildBlockCommentLines();
		file2.setUnchangedLines(diff.get(1));
		
		TreeMapper mapper = new TreeMapper(file1, file2);
		mapper.map();
		
		return true;
	}
	public void mapNaming()
	{
		final int threshold = 2;
		NamingVisitor namingVisitor = new NamingVisitor();
		this.ast1M.accept(namingVisitor);
		for (String name : namingVisitor.namingCounts.keySet())
			if (namingVisitor.namingCounts.get(name) < threshold)
				namingVisitor.naming.remove(name);
		for (String token : namingVisitor.tokenCounts.keySet())
			if (namingVisitor.tokenCounts.get(token) < threshold)
				namingVisitor.tokens.remove(token);
		this.naming = namingVisitor.naming;
		this.tokens = namingVisitor.tokens;
	}
	private class NamingVisitor extends ASTVisitor
	{
		public HashMap<String, String> naming = new HashMap<String, String>(), tokens = new HashMap<String, String>();
		public HashMap<String, Integer> namingCounts = new HashMap<String,Integer>(), tokenCounts = new HashMap<String,Integer>();
		
		public void preVisit(ASTNode node)
		{
			if(node.getNodeType() == ASTNode.SIMPLE_NAME)
			{
				if(node.getProperty(propertyCloneMap) != null)
				{
					SimpleName sm = (SimpleName)node;
					SimpleName csm = (SimpleName)node.getProperty(propertyCloneMap);
					String name1 = sm.getIdentifier(), name2 = csm.getIdentifier();
					if (!naming.containsKey(name1))
					{
						naming.put(name1, name2);
						namingCounts.put(name1, 1);
					}
					else
					{
						String name = naming.get(name1);
						if (name2.equals(name))
						{
							namingCounts.put(name1, namingCounts.get(name1)+1);
						}
						else if (name != null)
						{
							naming.put(name1, null);
							namingCounts.put(name1, 0);
						}
					}
					ArrayList<String> tokens1 = StringProcessor.serialize2(name1);
					ArrayList<String> tokens2 = StringProcessor.serialize2(name2);
					if(tokens1.size() == tokens2.size())
					{
						for(int i = 0; i < tokens1.size(); i++)
						{
							String token1 = tokens1.get(i).toLowerCase(), token2 = tokens2.get(i).toLowerCase();
							if(!tokens.containsKey(token1))
							{
								tokens.put(token1, token2);
								tokenCounts.put(token1, 1);
							}
							else
							{
								String token = tokens.get(token1);
								if (token2.equals(token1))
								{
									tokenCounts.put(token1, tokenCounts.get(token1)+1);
								}
								else if (token != null)
								{
									tokens.put(token1, null);
									tokenCounts.put(token1, 0);
								}
							}
						}
					}
				}
			}
		}
	}
	public boolean map(ASTNode root1, ASTNode root2, HashMap<ASTNode, ArrayList<ASTNode>> tree1, HashMap<ASTNode, ArrayList<ASTNode>> tree2)
	{
		if(root1.getNodeType() != root2.getNodeType())
		{
			System.out.println("Not exactly matched clones");
			return false;
		}
		root1.setProperty(propertyCloneMap, root2);
		root2.setProperty(propertyCloneMap, root1);
		ArrayList<ASTNode> children1 = tree1.get(root1);
		ArrayList<ASTNode> children2 = tree2.get(root2);
		if(children1.size() != children2.size())
		{
			System.out.println("Not exactly matched clones");
			return false;
		}
		else
		{
			for(int i = 0; i < children1.size(); i++)
				map(children1.get(i), children2.get(i), tree1, tree2);
		}
		return true;
	}
	private void buildNaming(ASTNode root, HashMap<ASTNode, ArrayList<ASTNode>> tree, 
			HashMap<String, ArrayList<String>> namingPosition, String pid, HashMap<String, ArrayList<ASTNode>> namingNode)
	{
		if(root.getNodeType() == ASTNode.SIMPLE_NAME)
		{
			String name = ((SimpleName)root).getIdentifier();
			ArrayList<String> pids = namingPosition.get(name);
			ArrayList<ASTNode> nodes = namingNode.get(name);
			if(pids == null)
			{
				pids = new ArrayList<String>();
				nodes = new ArrayList<ASTNode>();
			}
			pids.add(pid);
			nodes.add(root);
			namingPosition.put(name, pids);
			namingNode.put(name, nodes);
		}
		else
		{
			ArrayList<ASTNode> children = tree.get(root);
			for(int i = 0; i < children.size(); i++)
				buildNaming(children.get(i), tree, namingPosition, pid + i + ".", namingNode);
		}
	}
	public boolean checkCloning(ASTNode ast1, ASTNode ast2, HashMap<ASTNode, ArrayList<ASTNode>> tree1, HashMap<ASTNode, ArrayList<ASTNode>> tree2)
	{
		buildNaming(ast1, tree1, namingPosition1M, ".", namingNode1M);
		buildNaming(ast2, tree2, namingPosition2M, ".", namingNode2M);
		
		if(namingPosition1M.keySet().size() != namingPosition2M.keySet().size())
			return false;
		else
		{
			for(String name1 : namingPosition1M.keySet())
			{
				ArrayList<String> pids1 = namingPosition1M.get(name1);
				boolean isNamed = false;
				for(String name2 : namingPosition2M.keySet())
					if(namingPosition2M.get(name2).equals(pids1))
					{
						isNamed = true;
						break;
					}
				if(!isNamed)
				{
					System.out.println("Not same naming in cloning: " + name1);
					return false;
				}
			}
		}
		return true;
	}
	public void synchronize()
	{
		astSync12 = ASTNode.copySubtree(ast1N.getAST(), ast1N);
		TreeVisitor treeVisitor = new TreeVisitor(astSync12);
		astSync12.accept(treeVisitor);
		treeSync12 = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		synchronize(ast1N, tree1N, astSync12, treeSync12);
		synchronizeNaming(astSync12, treeSync12);
		synchronizeMoving(astSync12, treeSync12);
	}
	private void synchronize(ASTNode node1, HashMap<ASTNode, ArrayList<ASTNode>> tree1, ASTNode node2, HashMap<ASTNode, ArrayList<ASTNode>> tree2)
	{
		node2.setProperty(propertySyncMap, node1.getProperty(TreeMapper.propertyRevisionMap));
		if (node1.getProperty(TreeMapper.propertyRevisionMap) != null)
		{
			ASTNode node =(ASTNode)node1.getProperty(TreeMapper.propertyRevisionMap);
			node.setProperty(propertySyncMap, node2);
		}
		ArrayList<ASTNode> children1 = tree1.get(node1), children2 = tree2.get(node2);
		for(int i = 0; i < children1.size(); i++)
			synchronize(children1.get(i), tree1, children2.get(i), tree2);
	}
	private void relabel(ASTNode node, ASTNode rep)
	{
		switch (node.getNodeType())
		{
		case ASTNode.SIMPLE_NAME:
			SimpleName sn1 = (SimpleName)node, sn2 = (SimpleName)rep;
			sn1.setIdentifier(sn2.getIdentifier());
			break;
		case ASTNode.STRING_LITERAL:
			StringLiteral sl1 = (StringLiteral)node, sl2 = (StringLiteral)rep;
			sl1.setLiteralValue(sl2.getLiteralValue());
			break;
		case ASTNode.BOOLEAN_LITERAL:
			BooleanLiteral bl1 = (BooleanLiteral)node, bl2 = (BooleanLiteral)rep;
			bl1.setBooleanValue(bl2.booleanValue());
			break;
		case ASTNode.CHARACTER_LITERAL:
			CharacterLiteral cl1 =(CharacterLiteral)node, cl2 = (CharacterLiteral)rep;
			cl1.setCharValue(cl2.charValue());
			break;
		case ASTNode.INFIX_EXPRESSION:
			InfixExpression ie1 = (InfixExpression)node, ie2 = (InfixExpression)rep;
			ie1.setOperator(ie2.getOperator());
			break;
		case ASTNode.POSTFIX_EXPRESSION:
			PostfixExpression se1 = (PostfixExpression)node, se2 = (PostfixExpression)rep;
			se1.setOperator(se2.getOperator());
			break;
		case ASTNode.PREFIX_EXPRESSION:
			PrefixExpression pe1 = (PrefixExpression)node, pe2 = (PrefixExpression)rep;
			pe1.setOperator(pe2.getOperator());
			break;
		case ASTNode.MODIFIER:
			Modifier mod1 = (Modifier)node, mod2 = (Modifier)rep;
			mod1.setKeyword(mod2.getKeyword());
			break;
		case ASTNode.NULL_LITERAL:
			break;
		case ASTNode.NUMBER_LITERAL:
			NumberLiteral nl1 = (NumberLiteral)node, nl2 = (NumberLiteral)rep;
			nl1.setToken(nl2.getToken());
			break;
		case ASTNode.PRIMITIVE_TYPE:
			PrimitiveType pt1 = (PrimitiveType)node, pt2 = (PrimitiveType)rep;
			pt1.setPrimitiveTypeCode(pt2.getPrimitiveTypeCode());
			break;
		case ASTNode.TYPE_LITERAL:
			TypeLiteral tl1 = (TypeLiteral)node, tl2 = (TypeLiteral)rep;
			tl1.setType(tl2.getType());
			break;
		}
	}
	private void synchronizeNaming(ASTNode node12, HashMap<ASTNode, ArrayList<ASTNode>> tree12)
	{
		ArrayList<ASTNode> children = tree12.get(node12);
		if(!children.isEmpty())
			for(ASTNode child : children)
				synchronizeNaming(child, tree12);
		
		if(node12.getProperty(propertySyncMap) != null)
		{
			ASTNode node1M = (ASTNode)node12.getProperty(propertySyncMap);
			if(node1M.getProperty(propertyCloneMap) != null)
			{
				ASTNode node2M = (ASTNode)node1M.getProperty(propertyCloneMap);
				if(children.isEmpty())
				{
					if(node1M.toString().equals(node12.toString()))
					{
						relabel(node12, node2M);
					}
					else 
					{
						if(node12.getNodeType() == ASTNode.SIMPLE_NAME)
						{
							SimpleName sm12 = (SimpleName)node12, sm2M = (SimpleName)node2M;
							String name12 = sm12.getIdentifier(), name2M = sm2M.getIdentifier();
							if(this.naming.get(name12) != null)
								sm12.setIdentifier(this.naming.get(name12));
							else if(!node1M.toString().equals(name2M))
							{
								ArrayList<String> tokens12 = StringProcessor.serialize2(name12);
								ArrayList<String> tokens1M = StringProcessor.serialize2(node1M.toString());
								ArrayList<String> tokens2M = StringProcessor.serialize2(name2M);
								if (tokens12.size() > 1 && tokens1M.size() == tokens2M.size())
								{
									String name2N = "";
									ArrayList<Integer> match = StringProcessor.match(tokens1M, tokens12);
									for(int i = 0; i < match.size(); i++)
									{
										String token12 = tokens12.get(i);
										if(match.get(i) > 0)
										{
											String token1M = tokens1M.get(match.get(i));
											String token2M = tokens2M.get(match.get(i));
											if(token12.equals(token1M))
												name2N += token2M;
											else if(this.tokens.get(token12.toLowerCase()) != null)
												name2N += StringProcessor.getToken(token12, this.tokens.get(token12.toLowerCase()));
											else if(token1M.equals(token2M))
												name2N += token12;
											else
												name2N += token2M;
										}
										else if(this.tokens.get(token12.toLowerCase()) != null)
											name2N += StringProcessor.getToken(token12, this.tokens.get(token12.toLowerCase()));
										else
											name2N += token12;
									}
									sm12.setIdentifier(name2N);
								}
								else
									sm12.setIdentifier(name2M);
							}
						}
					}
				}
				else
				{
					int nodeType = node12.getNodeType();
					if(nodeType == ASTNode.INFIX_EXPRESSION)
					{
						if(((InfixExpression)node1M).getOperator().equals(((InfixExpression)node12).getOperator()))
							((InfixExpression)node12).setOperator(((InfixExpression)node2M).getOperator());
						else if(!((InfixExpression)node1M).getOperator().equals(((InfixExpression)node2M).getOperator()))
								((InfixExpression)node12).setOperator(((InfixExpression)node2M).getOperator());
					}
					else if(nodeType == ASTNode.PREFIX_EXPRESSION)
					{
						if(((PrefixExpression)node1M).getOperator().equals(((PrefixExpression)node12).getOperator()))
							((PrefixExpression)node12).setOperator(((PrefixExpression)node2M).getOperator());
						else if(!((PrefixExpression)node1M).getOperator().equals(((PrefixExpression)node2M).getOperator()))
							((PrefixExpression)node12).setOperator(((PrefixExpression)node2M).getOperator());
					}
					else if(nodeType == ASTNode.POSTFIX_EXPRESSION)
					{
						if(((PostfixExpression)node1M).getOperator().equals(((PostfixExpression)node12).getOperator()))
							((PostfixExpression)node12).setOperator(((PostfixExpression)node2M).getOperator());
						else if(!((PostfixExpression)node1M).getOperator().equals(((PostfixExpression)node2M).getOperator()))
							((PostfixExpression)node12).setOperator(((PostfixExpression)node2M).getOperator());
					}
				}
			}
			else
			{
				if(node12.getNodeType() == ASTNode.SIMPLE_NAME)
				{
					SimpleName sm12 = (SimpleName)node12;
					String name12 = sm12.getIdentifier();
					if(this.naming.get(name12) != null)
						sm12.setIdentifier(this.naming.get(name12));
					else
					{
						String name2N = "";
						ArrayList<String> tokens12 = StringProcessor.serialize2(name12);
						boolean modified = false;
						if (tokens12.size() > 1)
							for(String token12 : tokens12)
								if(this.tokens.get(token12.toLowerCase()) != null)
								{
									modified = true;
									name2N += StringProcessor.getToken(token12, this.tokens.get(token12.toLowerCase()));
								}
								else
									name2N += token12;
						if(modified)
							sm12.setIdentifier(name2N);
					}
				}
			}
		}
		else
		{
			if(node12.getNodeType() == ASTNode.SIMPLE_NAME)
			{
				SimpleName sm12 = (SimpleName)node12;
				String name12 = sm12.getIdentifier();
				if(this.naming.get(name12) != null)
					sm12.setIdentifier(this.naming.get(name12));
				else
				{
					String name2N = "";
					ArrayList<String> tokens12 = StringProcessor.serialize2(name12);
					boolean modified = false;
					if (tokens12.size() > 1)
						for(String token12 : tokens12)
							if(this.tokens.get(token12.toLowerCase()) != null)
							{
								modified = true;
								name2N += StringProcessor.getToken(token12, this.tokens.get(token12.toLowerCase()));
							}
							else
								name2N += token12;
					if(modified)
						sm12.setIdentifier(name2N);
				}
			}
		}
	}
	private void synchronizeMoving(ASTNode node12, HashMap<ASTNode, ArrayList<ASTNode>> tree12)
	{
		ArrayList<ASTNode> children = tree12.get(node12);
		if(!children.isEmpty())
		{
			for(ASTNode child : children)
				synchronizeMoving(child, tree12);
			if (node12.getNodeType() == ASTNode.EXPRESSION_STATEMENT)
				return;
			ArrayList<HashMap<Integer, ASTNode>> moveNodesList = new ArrayList<HashMap<Integer,ASTNode>>();
			ArrayList<List<ASTNode>> lists = new ArrayList<List<ASTNode>>();
			int index12 = 0, c = 0;
			while (index12 < children.size())
			{
				ASTNode child12 = children.get(index12);
				if(child12.getNodeType() == ASTNode.EXPRESSION_STATEMENT)
					child12 = ((ExpressionStatement)child12).getExpression();
				if (child12.getProperty(propertySyncMap) != null)
				{
					ASTNode child1M = (ASTNode)child12.getProperty(propertySyncMap);
					ArrayList<ASTNode> children1M = tree1M.get(child1M.getParent());
					if(child1M.getParent().getNodeType() == ASTNode.EXPRESSION_STATEMENT)
						children1M = tree1M.get(child1M.getParent().getParent());
					int cloneStatus = (Integer)child1M.getProperty(propertyCloneMapStatus);
					if(cloneStatus == MapVisitor.STATUS_MOVED)
					{
						ASTNode child2M = (ASTNode)child1M.getProperty(propertyCloneMap);
						if (child2M.getParent().getNodeType() == ASTNode.EXPRESSION_STATEMENT)
							child2M = child2M.getParent();
						ArrayList<ASTNode> children2M = tree2M.get(child2M.getParent());
						int index2M = children2M.indexOf(child2M);
						int index2N = index2M;
						for(int i = index2M; i >= 0; i--)
						{
							ASTNode child21 = children1M.get(index2M);
							if(child21.getNodeType() == ASTNode.EXPRESSION_STATEMENT)
								child21 = ((ExpressionStatement)child21).getExpression();
							if(child21.getProperty(propertySyncMap) != null)
							{
								ASTNode child2N = (ASTNode)child21.getProperty(propertySyncMap);
								if (child2N.getParent().getNodeType() == ASTNode.EXPRESSION_STATEMENT)
									child2N = child2N.getParent();
								index2N = children.indexOf(child2N);
								break;
							}
						}
						if (child12.getParent().getNodeType() == ASTNode.EXPRESSION_STATEMENT)
							child12 = child12.getParent();
						List<ASTNode> list = (List<ASTNode>)node12.getStructuralProperty(child12.getLocationInParent());
						int listId = lists.indexOf(list);
						HashMap<Integer, ASTNode> moveNodes = null;
						if (listId == -1)
						{
							listId = lists.size();
							lists.add(list);
							moveNodesList.add(null);
							moveNodes = new HashMap<Integer, ASTNode>();
						}
						else
							moveNodes = moveNodesList.get(listId);
						moveNodes.put(list.indexOf(children.get(index2N)), child12);
						moveNodesList.set(listId, moveNodes);
					}
				}
				index12++;
			}
			for (int id = 0; id < lists.size(); id++)
			{
				List<ASTNode> list = lists.get(id);
				HashMap<Integer, ASTNode> moveNodes = moveNodesList.get(id);
				c = 0;
				for(ASTNode child12 : moveNodes.values())
				{
					c++;
					ASTNode newNode = child12;
					for(int index = list.indexOf(child12); index < list.size()-c; index++)
					{
						newNode = list.get(index + 1);
						ASTNode copiedNode = ASTNode.copySubtree(newNode.getAST(), newNode);
						list.set(index + 1, copiedNode);
						list.set(index, newNode);
						if(children.indexOf(newNode) == -1)
							System.err.println("Error in sync moving");
						children.set(children.indexOf(newNode)-1, newNode);
					}
					
				}
			}
			for (int id = 0; id < lists.size(); id++)
			{
				List<ASTNode> list = lists.get(id);
				HashMap<Integer, ASTNode> moveNodes = moveNodesList.get(id);
				for(int i = 0; i < list.size(); i++)
				{
					if (moveNodes.containsKey(i))
					{
						ASTNode child12 = moveNodes.get(i);
						index12 = children.indexOf(list.get(i));
						moveNodes.remove(i);
						ASTNode newNode = child12;
						for(int index = list.size()-moveNodes.size()-1; index > i; index--)
						{
							newNode = list.get(index-1);
							ASTNode copiedNode = ASTNode.copySubtree(newNode.getAST(), newNode);
							list.set(index-1, copiedNode);
							list.set(index, newNode);
							children.set(children.indexOf(newNode)+1, newNode);
						}
						list.set(i, child12);
						children.set(index12, child12);
					}
				}
			}
		}
	}
	/*private void synchronize(ASTNode node12)
	{
		ArrayList<ASTNode> children = tree1N.get(node12);
		if(!children.isEmpty())
			for(ASTNode child : children)
				synchronize(child);
		
		ASTNode node1M = (ASTNode)node12.getProperty(TreeMapper.propertyRevisionMap);
		if(node1M != null)
		{
			boolean isModified = false;
			ASTNode node2M = (ASTNode)node1M.getProperty(propertyCloneMap);
			ASTNode node2N = (ASTNode)node12.getProperty(propertyCloneMap);
			if(children.isEmpty())
			{
				if(node2M == null)
					return;
				if(!node1M.toString().equals(node2M.toString()))
				{
					if(node1M.getNodeType() == ASTNode.SIMPLE_NAME)
						((SimpleName)node2N).setIdentifier(((SimpleName)node2M).getIdentifier());
					else
						replaceNode(node2N, node2M);
				}
				else
					if(node1M.getNodeType() == ASTNode.SIMPLE_NAME)
						((SimpleName)node2N).setIdentifier(((SimpleName)node12).getIdentifier());
					else
						replaceNode(node2N, node1N);
		 	    if(!node1M.toString().equals(node12.toString()))
		 	    	isModified = true;
			}
			else
			{
				int nodeType = node12.getNodeType();
				if(nodeType == ASTNode.INFIX_EXPRESSION)
				{
					if(((InfixExpression)node1M).getOperator().equals(((InfixExpression)node2M).getOperator()))
						((InfixExpression)node2N).setOperator(((InfixExpression)node12).getOperator());
					else
						((InfixExpression)node2N).setOperator(((InfixExpression)node2M).getOperator());
					if(((InfixExpression)node1M).getOperator().equals(((InfixExpression)node12).getOperator()))
						isModified = true;
				}
				else if(nodeType == ASTNode.PREFIX_EXPRESSION)
				{
					if(((PrefixExpression)node1M).getOperator().equals(((PrefixExpression)node2M).getOperator()))
						((PrefixExpression)node2N).setOperator(((PrefixExpression)node12).getOperator());
					else
						((PrefixExpression)node2N).setOperator(((PrefixExpression)node2M).getOperator());
					if(((PrefixExpression)node1M).getOperator().equals(((PrefixExpression)node12).getOperator()))
						isModified = true;
				}
				else if(nodeType == ASTNode.POSTFIX_EXPRESSION)
				{
					if(((PostfixExpression)node1M).getOperator().equals(((PostfixExpression)node2M).getOperator()))
						((PostfixExpression)node2N).setOperator(((PostfixExpression)node12).getOperator());
					else
						((PostfixExpression)node2N).setOperator(((PostfixExpression)node2M).getOperator());
					if(((PostfixExpression)node1M).getOperator().equals(((PostfixExpression)node12).getOperator()))
						isModified = true;
				}
			}
			if(isModified)
			{
				node1M.setProperty(propertyStatus, STATUS_MODIFIED);
				node1N.setProperty(propertyStatus, STATUS_MODIFIED);
				node2M.setProperty(propertyStatus, STATUS_MODIFIED);
				node2N.setProperty(propertyStatus, STATUS_MODIFIED);
			}
			else
			{
				node1M.setProperty(propertyStatus, STATUS_UNCHANGED);
				node1N.setProperty(propertyStatus, STATUS_UNCHANGED);
				node2M.setProperty(propertyStatus, STATUS_UNCHANGED);
				node2N.setProperty(propertyStatus, STATUS_UNCHANGED);
			}
		}
		else
		{
			if(node12.getNodeType() == ASTNode.SIMPLE_NAME)
			{
				String name = ((SimpleName)node12).getIdentifier();
				ArrayList<ASTNode> nodes1 = namingNode1N.get(name);
				for(ASTNode node : nodes1)
					if(node.getProperty(TreeMapper.propertyRevisionMap) != null)
					{
						node1M = node;
						break;
					}
				if(node1M != null)
				{
					((SimpleName)node12.getProperty(propertyCloneMap)).setIdentifier(((SimpleName)node1M.getProperty(propertyCloneMap)).getIdentifier());
				}
			}
			node12.setProperty(propertyStatus, STATUS_ADDED);
			((ASTNode)node12.getProperty(propertyCloneMap)).setProperty(propertyStatus, STATUS_ADDED);
		}
	}*/
	private void replaceNode(ASTNode oldNode, ASTNode newNode)
	{
		//ASTNode copiedNode = ASTNode.copySubtree(newNode.getAST(), newNode);
		ASTNode parent = oldNode.getParent();
	    StructuralPropertyDescriptor location = oldNode.getLocationInParent();
 	    if (location.isChildProperty()) {
 	        parent.setStructuralProperty(location, oldNode);
 	    } else if (location.isChildListProperty()) {
 	        List list = (List)parent.getStructuralProperty(location);
 	        int index = list.indexOf(oldNode);
 	        list.set(index, newNode);
 	    }
	}
	/*public int checkConsistency()
	{
		if(!map(ast1N, ast2N, tree1N, tree2N))
		{
			System.out.println("Inconsitent Structure");
			return CHANGE_INCONSISTENT_STRUCTURE;
		}
		if(!checkCloning(ast1N, ast2N, tree1N, tree2N))
		{
			System.out.println("Inconsitent Labeling");
			return CHANGE_INCONSISTENT_NAMING;
		}
		if(!checkConsistentRelabeling(ast1N, ast2N))
		{
			System.out.println("Inconsitent Labeling");
			return CHANGE_INCONSISTENT_NAMING;
		}
		return CHANGE_OK;
	}*/
	/*private boolean checkConsistentRelabeling(ASTNode node1N, ASTNode node2N)
	{
		ArrayList<ASTNode> children1N = tree1N.get(node1N);
		ArrayList<ASTNode> children2N = tree2N.get(node2N);
		if(!children1N.isEmpty())
			for(int i =0; i < children1N.size(); i++)
				return checkConsistentRelabeling(children1N.get(i), children2N.get(i));
		if(!node1N.getProperty(MapVisitor.propertyStatus).equals(node2N.getProperty(MapVisitor.propertyStatus)))
			return false;
		return true;
	}*/
	/*public int merge()
	{
		int result = CHANGE_OK;
		int consistency = checkConsistency();
		if(consistency == CHANGE_OK)
			return CHANGE_OK;
		
		
		
		return result;
	}*/
	
	private class SequenceIDVisitor extends ASTVisitor {
		private int sid = 0;
		private int lastNode = 0;	//for end token of vector
		private Stack<Integer> stackCurrentNode = new Stack<Integer>();	//for start token of vector
		public ASTNode node = null;

		public SequenceIDVisitor(int sid)
		{
			this.sid = sid;
		}
		public void preVisit(ASTNode node) {
			int nodeType = node.getNodeType(); 
			if (nodeType == ASTNode.JAVADOC || nodeType == ASTNode.BLOCK_COMMENT || nodeType == ASTNode.LINE_COMMENT)
				return;
			stackCurrentNode.push(++lastNode); 
		}
		public void postVisit(ASTNode node) {
			int nodeType = node.getNodeType();
			if (nodeType == ASTNode.JAVADOC || nodeType == ASTNode.BLOCK_COMMENT || nodeType == ASTNode.LINE_COMMENT)
				return;
			int currentNode = stackCurrentNode.pop();
			if(nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.BLOCK && nodeType != ASTNode.PARENTHESIZED_EXPRESSION && 
					lastNode-currentNode+1 >= Main.minFragmentSize)
			{
				if(this.sid == 0)
					this.node = node;
				this.sid--;
			}

		}
	}
	
	/*private ASTNode getASTNode(int sid, HashMap<ASTNode, ArrayList<ASTNode>> tree, ASTNode root)
	{
		int nodeType = root.getNodeType(); 
    	if (nodeType == ASTNode.JAVADOC || nodeType == ASTNode.BLOCK_COMMENT || nodeType == ASTNode.LINE_COMMENT)
    		return null;
    	stackCurrentNode.push(++lastNode); 
    	ArrayList<ASTNode> children = tree.get(root);
		if(children != null)
			for(ASTNode child : children)
			{
				ASTNode node = getASTNode(sid, tree, child);
				if(node != null)
					return node;
			}
		int currentNode = stackCurrentNode.pop();
    	if(nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.BLOCK && nodeType != ASTNode.PARENTHESIZED_EXPRESSION && 
    			lastNode-currentNode+1 >= Main.minFragmentSize)
		{
    		if(this.sid == sid)
    			return root;
    		this.sid++;
		}
		return null;
	}*/
	private class TreeVisitor extends ASTVisitor {
		private ASTNode root;
		private HashMap<ASTNode, ArrayList<ASTNode>> tree = new HashMap<ASTNode, ArrayList<ASTNode>>();
		
		public TreeVisitor(ASTNode root)
		{
			this.root = root;
		}
		public HashMap<ASTNode, ArrayList<ASTNode>> getTree() {
			return tree;
		}
		public void preVisit(ASTNode node)
		{
			tree.put(node, new ArrayList<ASTNode>());
			if(node!= root)
			{
				tree.get(node.getParent()).add(node);
			}
		}
	}
	
	/*private class TreeVisitor extends ASTVisitor 
	{
		public static final String propertyLocationID = "lid";
		public static final String propertyOffset = "offset";
		public static final String propertyStartLine = "sl";
		public static final String propertyEndLine = "el";
		private ASTNode root;
		private HashMap<ASTNode, ArrayList<ASTNode>> tree = new HashMap<ASTNode, ArrayList<ASTNode>>();
		
		public TreeVisitor(ASTNode root)
		{
			this.root = root;
		}
		public HashMap<ASTNode, ArrayList<ASTNode>> getTree() {
			return tree;
		}
		public void preVisit(ASTNode node)
		{
			if(node.getNodeType() != ASTNode.BLOCK && node.getNodeType() != ASTNode.EXPRESSION_STATEMENT && node.getNodeType() != ASTNode.PARENTHESIZED_EXPRESSION)
			{
				tree.put(node, new ArrayList<ASTNode>());
				ASTNode pNode = getParent(node, root);
				if(pNode != null)
				{
					tree.get(pNode).add(node);
				}
			}
		}
		private ASTNode getParent(ASTNode node, ASTNode root)
		{
			if(node.equals(root))
				return null;
			ASTNode pNode = node.getParent();
			while(pNode != null && !pNode.equals(root) &&
					(pNode.getNodeType() == ASTNode.BLOCK || pNode.getNodeType() == ASTNode.EXPRESSION_STATEMENT || pNode.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION))
				pNode = pNode.getParent();
			
			return pNode;
		}
	}*/
	
}
