package CloneTransforming;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

public class CloneSystemSyncInfo {
	public String systemName;
	public HashMap<Integer, ArrayList<ClonePairSyncInfo>> allPairs = new HashMap<Integer, ArrayList<ClonePairSyncInfo>>();
	public int numOfUncheckeds = 0, numOfPairs = 0, numOf1SidePairs = 0, numOf2SidePairs = 0, numOfLateSync = 0,
				numOfSIPairs = 0, numOfNIPairs = 0, numOfLIPairs = 0, 
				numOfSyncPairs = 0, numOfCorrectSyncPairs = 0, numOfLocs = 0, numOfCorrectLocs = 0, numOfMissLocs = 0, numOfWrongIn1F = 0;

	public void readSyncInfo(String path)
	{
		int index = path.lastIndexOf("\\");
		String name = path.substring(index+1);
		String[] parts = name.split("_");
		systemName = parts[0];
		ArrayList<Integer> fixRevs = CloneSync.getFixRevisions(path+ "\\..\\" + systemName + "_fix_bug_revisions.txt");
		ArrayList<Integer> cloneRevs = CloneSync.getCloneRevisions(path + "\\inconsistencies");
		if (fixRevs != null)
			cloneRevs.retainAll(fixRevs);
		Collections.sort(cloneRevs);
		System.out.println(systemName);
		for(int rev : cloneRevs)
		{
			System.out.println("\t" + rev);
			ArrayList<ClonePairSyncInfo> syncPairs = new ArrayList<ClonePairSyncInfo>();
			File dir = new File(path + "\\inconsistencies\\" + rev);
			File[] files = dir.listFiles();
			Arrays.sort(files);
			HashSet<String> dupPairs = new HashSet<String>();
			for(File sub : files)
			{
				if(!dupPairs.contains(sub.getName()))
				{
					String[] pairs = sub.getName().split("_");
					dupPairs.add(pairs[1] + "_" + pairs[0]);
					System.out.println("\t\t" + sub.getName());
					ClonePairSyncInfo info = new ClonePairSyncInfo();
					boolean done = info.readSyncInfoFile(sub.getAbsolutePath() + "\\info.txt");
					if(done)
					{
						info.revision = rev;
						info.id1 = Integer.parseInt(pairs[0]); info.id2 = Integer.parseInt(pairs[1]);
						this.numOfPairs++;
						if (info.change == 1)
							this.numOf1SidePairs++;
						else if (info.change == 2)
							this.numOf2SidePairs++;
						else
							System.err.println("Neither 1 nor 2 side change");
						String[] inconsistencies = info.inconsistent.split(" ");
						for (String i : inconsistencies)
						{
							if (i.equals("S"))
								this.numOfSIPairs++;
							else if (i.equals("N"))
								this.numOfNIPairs++;
							else if (i.equals("L"))
								this.numOfLIPairs++;
							else
								System.err.println("Unknown inconsistency");
						}
						if(info.lateSync)
							this.numOfLateSync++;
						if (info.jSync)
							this.numOfCorrectSyncPairs++;
						//if (info.change == 2 || info.lateSync)
						{
							this.numOfLocs += info.locations;
							this.numOfCorrectLocs += info.correct;
							this.numOfMissLocs += info.miss;
						}
						if (info.change == 1 && !info.lateSync && !info.jSync)
							this.numOfWrongIn1F += info.miss;
					}
					else
						this.numOfUncheckeds++;
				}
			}
			if (!syncPairs.isEmpty())
				allPairs.put(rev, syncPairs);
		}
	}
	public String print()
	{
		StringBuffer buf = new StringBuffer();
		buf.append(this.systemName + "\r\n");
		buf.append("Checked pairs: " + this.numOfPairs + "\tUnchecked: " + this.numOfUncheckeds + "\r\n");
		buf.append("1-side: " + this.numOf1SidePairs + "\t2-side: " + this.numOf2SidePairs + "\tlate-sync: " + this.numOfLateSync + "\r\n");
		buf.append("SI: " + this.numOfSIPairs + "\tNI: " + this.numOfNIPairs + "\tLI: " + this.numOfLIPairs + "\r\n");
		buf.append("Correctly-synced pairs: " + this.numOfCorrectSyncPairs + "\r\n");
		buf.append("Locations: " + this.numOfLocs + "\tCorrect: " + this.numOfCorrectLocs + "\tWrong: " + this.numOfMissLocs + "\tWrong in 1F: " + this.numOfWrongIn1F);
		
		return buf.toString();
	}
	
	public static void main(String[] args)
	{
		
	}
}
