package CloneTransforming;

import java.util.ArrayList;
import java.util.HashMap;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;

import treeMapping.MapVisitor;
import treeMapping.TreeMapper;

public class ScriptGen {
	private ASTNode astM, astN;
	private HashMap<ASTNode, ArrayList<ASTNode>> treeM, treeN;
	
	public ScriptGen(ASTNode astM, ASTNode astN) {
		this.astM = astM;
		this.astN = astN;
		TreeVisitor treeVisitor = new TreeVisitor(astM);
		astM.accept(treeVisitor);
		this.treeM = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
		treeVisitor = new TreeVisitor(astN);
		astN.accept(treeVisitor);
		this.treeN = new HashMap<ASTNode, ArrayList<ASTNode>>(treeVisitor.getTree());
	}
	private class TreeVisitor extends ASTVisitor {
		public static final String propertyLocationID = "lid";
		public static final String propertyOffset = "offset";
		public static final String propertyStartLine = "sl";
		public static final String propertyEndLine = "el";
		private ASTNode root;
		private HashMap<ASTNode, ArrayList<ASTNode>> tree = new HashMap<ASTNode, ArrayList<ASTNode>>();
		
		public TreeVisitor(ASTNode root)
		{
			this.root = root;
		}
		public HashMap<ASTNode, ArrayList<ASTNode>> getTree() {
			return tree;
		}
		public void preVisit(ASTNode node)
		{
			if(node.getNodeType() != ASTNode.BLOCK && node.getNodeType() != ASTNode.EXPRESSION_STATEMENT && node.getNodeType() != ASTNode.PARENTHESIZED_EXPRESSION)
			{
				tree.put(node, new ArrayList<ASTNode>());
				ASTNode pNode = getParent(node, root);
				if(pNode != null)
				{
					tree.get(pNode).add(node);
				}
			}
		}
		private ASTNode getParent(ASTNode node, ASTNode root)
		{
			if(node.equals(root))
				return null;
			ASTNode pNode = node.getParent();
			while(pNode != null && !pNode.equals(root) &&
					(pNode.getNodeType() == ASTNode.BLOCK || pNode.getNodeType() == ASTNode.EXPRESSION_STATEMENT || pNode.getNodeType() == ASTNode.PARENTHESIZED_EXPRESSION))
				pNode = pNode.getParent();
			
			return pNode;
		}
	}
	public int count1(ASTNode root)
	{
		int count = 0;
		ArrayList<ASTNode> children = treeM.get(root);
		for(ASTNode child : children)
		{
			count += count1(child);
		}
		/*System.err.println(root);
		System.err.println(astM.getRoot());*/
		int status = (Integer)root.getProperty(MapVisitor.propertyStatus);
		if(status > MapVisitor.STATUS_FULLY_CHANGED && root.getProperty(TreeMapper.propertyRevisionMap) != null)
		{
			count++;
		}
		return count;
	}
	public int count2(ASTNode root)
	{
		int count = 0;
		ArrayList<ASTNode> children = treeN.get(root);
		for(ASTNode child : children)
		{
			count += count2(child);
		}
		int status = (Integer)root.getProperty(MapVisitor.propertyStatus);
		if(status == MapVisitor.STATUS_ADDED)
		{
			count++;
		}
		return count;
	}
}
