package cleman;

import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.ArrayList;

import org.eclipse.jdt.core.dom.ASTNode;

/**
 * @author Nguyen Anh Hoan
 *
 */
public class Fragment implements Serializable {
	protected static final long serialVersionUID = 1L;
	
	public static int maxSizeOfGram = 4;	// now < 8
	//public static int mergeWindowSize = Integer.MAX_VALUE/2;
	public static int mergeWindowSize = 1;
	
	/**
	 * Categories of AST node types
	 */
	public static final byte ClassFragment = 1;
	public static final byte MethodFragment = 2;
	public static final byte LoopStatementFragment = 3;
	public static final byte IfStatementFragment = 4;
	public static final byte SwitchStatementFragment = 5;
	public static final byte BlockFragment = 6;
	public static final byte MethodState = 7;
	public static final byte DeclarationState = 8;
	public static final byte AssertState = 9;
	public static final byte AssignState = 10;
	
	public static final byte ArrayState = 13;
	public static final byte Expression = 14;
	public static final byte DeclarationExp = 17;
	public static final byte SimpleName = 18;
	public static final byte Literal = 19;
	public static final byte StatementsGroupFragment = 27;
	public static final byte MethodsGroupFragment = 28;
	public static final byte OtherStatementFragment = 29;
	public static final byte OtherFragments = 30;
	public static final byte NotConsideredFrags = 31;

	public static String[] typeName = {"", "Class", "Method", "Loop", "If", "Switch", "Block", "MethodState", "Decl", "Assert", "Assign",
		"Arr", "Expr", "DeclEx", "Name", "Lit"};
	/**
	 * Global Maps between an n-gram and its index and vice versa
	 */
	public static HashMap<Integer, Integer> gram2Index = new HashMap<Integer, Integer>();
	public static HashMap<Integer, Integer> index2Gram = new HashMap<Integer, Integer>();

	public static int nextID = 1;
	public static HashMap<Integer, HashMap<Integer, Fragment>> allRevisions = new HashMap<Integer, HashMap<Integer, Fragment>>();
	public static HashMap<Integer, Fragment> prev = new HashMap<Integer, Fragment>();
	public static HashMap<Integer, Fragment> all = new HashMap<Integer, Fragment>();
	public static HashMap<Integer, Fragment> unchangeds = new HashMap<Integer, Fragment>();
	public static HashMap<Integer, Fragment> addeds = new HashMap<Integer, Fragment>();
	public static HashMap<Integer, Fragment> modifieds = new HashMap<Integer, Fragment>();
	public static HashMap<Integer, Fragment> deleteds = new HashMap<Integer, Fragment>();
	
	private int id;
	private SourceFile sourceFile;
	private ASTNode astNode = null;
	private byte type = OtherFragments; 
	private String code = null;
	private int startChar;	//start character
	private int length;	//in characters including CrLf
	private int startLine;
	private int endLine;
	private HashSet<Bucket> buckets;	// = new HashSet<Bucket>();
	private HashSet<Integer> clones = new HashSet<Integer>();
	/**
	 * 
	 */
	public Fragment() {
		this.id = nextID++;
	}
	public Fragment(int id)
	{
		this.id = id;
	}
	public Fragment clone()
	{
		Fragment fragment = new Fragment(this.id);
		fragment.setBuckets(buckets);
		fragment.setClones(clones);
		fragment.setCode(code);
		fragment.setGramVector(gramVector);
		fragment.setSourceFile(sourceFile);
		fragment.setStartChar(startChar);
		fragment.setLength(length);
		fragment.setType(type);
		fragment.setVectorLength();
		
		return fragment;
	}
	/**
	 * @return the sourceFile
	 */
	public SourceFile getSourceFile() {
		return sourceFile;
	}
	/**
	 * @param sourceFile the sourceFile to set
	 */
	public void setSourceFile(SourceFile sourceFile) {
		this.sourceFile = sourceFile;
	}
	public ASTNode getAstNode() {
	
		return astNode;
	}
	public void setAstNode(ASTNode astNode) {
		this.astNode = astNode;
	}
	/**
	 * @return the buckets
	 */
	public HashSet<Bucket> getBuckets() {
		return buckets;
	}
	/**
	 * @param buckets the buckets to set
	 */
	public void setBuckets(HashSet<Bucket> buckets) {
		this.buckets = buckets;
	}
	public HashSet<Integer> getClones() {
		return clones;
	}
	public void setClones(HashSet<Integer> clones) {
		this.clones = clones;
	}

	HashMap<Integer, Integer> gramVector = new HashMap<Integer, Integer>();
	double vectorLength;
	
	/**
	 * @return the vectorLength
	 */
	public double getVectorLength() {
		return vectorLength;
	}
	/**
	 * @param vectorLength the vectorLength to set
	 */
	public void setVectorLength() {
		double len = 0;
		for(int val : this.gramVector.values())
			len += val * val;
		len  = Math.sqrt(len);
		this.vectorLength = len;
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.id;
	}
	/**
	 * @return the type
	 */
	public byte getType() {
		return type;
	}
	/**
	 * @param type the type to set
	 */
	public void setType(byte type) {
		this.type = type;
	}
	public int getId() {
		return id;
	}
	public void setId(int id) {
		this.id = id;
	}
	/*public String getFileName() {
		return fileName;
	}
	public void setFileName(String fileName) {
		this.fileName = fileName;
	}*/
	public String getCode() {
		return code;
	}
	public void setCode(String code) {
		this.code = code;
	}
	public int getStartChar() {
		return startChar;
	}
	public void setStartChar(int startChar) {
		this.startChar = startChar;
	}
	public int getStartLine() {
		return startLine;
	}
	public void setStartLine(int startLine) {
		this.startLine = startLine;
	}
	public int getEndLine() {
		return endLine;
	}
	public void setEndLine(int endLine) {
		this.endLine = endLine;
	}
	public int getLength() {
		return length;
	}
	public void setLength(int length) {
		this.length = length;
	}
	/**
	 * This map contains only the grams appearing in this fragment
	 */
	public HashMap<Integer, Integer> getSparseGramVector() {
		return this.gramVector;
	}
	/**
	 * This vector has the full size of all grams in the project. Thus, there might be many zeroed elements
	 * @return
	 */
	public int[] getFullGramVector() {
		int[] full = new int[gram2Index.size()];
		for(int index = 0; index < gram2Index.size(); index++)
			full[index] = (this.gramVector.containsKey(index)?this.gramVector.get(index):0);
		return full;
	}
	public void setGramVector(HashMap<Integer, Integer> gramVector) {
		this.gramVector = new HashMap<Integer, Integer>(gramVector);
	}
	/*
	 * 
	 */
	public boolean encloses(Fragment other)
	{
		//if(this.sourceFile.equals(other.getSourceFile()) && (this.startChar <= other.getStartChar()) && 
		if(this.sourceFile == other.getSourceFile() && (this.startChar <= other.getStartChar()) &&
				((this.startChar+this.length) >= (other.getStartChar()+other.getLength())))
			return true;
		return false;
	}
	public boolean overlap(Fragment other)
	{
		if(this.sourceFile == other.getSourceFile())
		{
			int start1 = this.startChar, end1 = start1 + this.length;
			int start2 = other.getStartChar(), end2 = start2 + other.getLength();
			if(start1 >= start2 && start1 <= end2)
				return true;
			if(end1 >= start2 && end1 <= end2)
				return true;
			if(start2 >= start1 && start2 <= end1)
				return true;
			if(end2 >= start1 && end2 <= end1)
				return true;
		}
		return false;
	}
	/**
	 * 
	 * @param other
	 * @return match[0] = common, match[1] = union, match[2] = min length, match[3] = max length
	 */
	public int[] match(Fragment other)
	{
		int[] matches = new int[4];
		int start1 = this.startChar, end1 = this.startChar + this.length - 1;
		int start2 = other.getStartChar(), end2 = other.getStartChar() + other.getLength() - 1;
		matches[0] = 0;
		matches[1] = end1 - start1 + end2 - start2 + 2;
		if(this.sourceFile == other.getSourceFile())
		{
			if(start1 >= start2 && start1 <= end2)
			{
				matches[0] = Math.min(end1, end2) - start1 + 1;
				matches[1] = Math.max(end1, end2) - start2 + 1;
			}
			else if(end1 >= start2 && end1 <= end2)
			{
				matches[0] = end1 - Math.max(start1, start2) + 1;
				matches[1] = end2 - Math.min(start1, start2) + 1;
			}
			if(start2 >= start1 && start2 <= end1)
			{
				matches[0] = Math.min(end1, end2) - start2 + 1;
				matches[1] = Math.max(end1, end2) - start1 + 1;
			}
			if(end2 >= start1 && end2 <= end1)
			{
				matches[0] = end2 - Math.max(start1, start2) + 1;
				matches[1] = end1 - Math.min(start1, start2) + 1;
			}
		}
		
		return matches;
	}
	/*
	 * 
	 */
	public double similarity(Fragment fragment)
	{
		if(fragment.getType() != this.type || this.encloses(fragment) || fragment.encloses(this))
			return 0;
		else
		{
			HashMap<Integer, Integer> v1 = new HashMap<Integer, Integer>(this.gramVector);
			HashMap<Integer, Integer> v2 = new HashMap<Integer, Integer>(fragment.getSparseGramVector());
			HashSet<Integer> keys = new HashSet<Integer>(v1.keySet());
			keys.addAll(v2.keySet());
			
			double d = 0;
			for(int key : keys)
			{
				int temp = (v1.containsKey(key) ? v1.get(key) : 0) - (v2.containsKey(key) ? v2.get(key) : 0);
				d += temp * temp;
			}
			d = Math.sqrt(d) * 2.0 / (this.vectorLength + fragment.getVectorLength());
			return (1 - d);
		}
	}
	/*
	 * 
	 */
	/*public boolean isClonedTo(Fragment fragment)
	{
		//if(fragment.getType() != this.type || this.descendants.contains(fragment.getId()) || fragment.descendants.contains(this.id))
		//if(fragment.getType() != this.type || this.encloses(fragment) || fragment.encloses(this))
		if(fragment.getType() != this.type || this.overlap(fragment))
			return false;
		//else if(this.group != null && this.group.equals(fragment.getGroup()))
		else if(this.clones.contains(fragment.getId()))
			return true;
		else
		{
			HashMap<Integer, Integer> v1 = new HashMap<Integer, Integer>(this.gramVector);
			HashMap<Integer, Integer> v2 = new HashMap<Integer, Integer>(fragment.getSparseGramVector());
			HashSet<Integer> keys = new HashSet<Integer>(v1.keySet());
			keys.addAll(v2.keySet());
			
			double d = 0;
			for(int key : keys)
			{
				int temp = (v1.containsKey(key) ? v1.get(key) : 0) - (v2.containsKey(key) ? v2.get(key) : 0);
				d += temp * temp;
			}
			d = Math.sqrt(d) * 2.0 / (this.vectorLength + fragment.getVectorLength());
			//d = Math.sqrt(d) / Math.min(this.vectorLength, fragment.getVectorLength());
			return (d <= Main.threshold);
		}
	}*/
	public double clonedTo(Fragment fragment)
	{
		//if(fragment.getType() != this.type || this.descendants.contains(fragment.getId()) || fragment.descendants.contains(this.id))
		//if(fragment.getType() != this.type || this.encloses(fragment) || fragment.encloses(this))
		if(fragment.getType() != this.type || this.overlap(fragment))
			return 1.0;
		//else if(this.group != null && this.group.equals(fragment.getGroup()))
		else if(this.clones.contains(fragment.getId()))
			return 1.0;
		else
		{
			HashMap<Integer, Integer> v1 = new HashMap<Integer, Integer>(this.gramVector);
			HashMap<Integer, Integer> v2 = new HashMap<Integer, Integer>(fragment.getSparseGramVector());
			HashSet<Integer> keys = new HashSet<Integer>(v1.keySet());
			keys.addAll(v2.keySet());
			
			double d = 0;
			for(int key : keys)
			{
				int temp = (v1.containsKey(key) ? v1.get(key) : 0) - (v2.containsKey(key) ? v2.get(key) : 0);
				d += temp * temp;
			}
			d = Math.sqrt(d) * 2.0 / (this.vectorLength + fragment.getVectorLength());
			//d = Math.sqrt(d) / Math.min(this.vectorLength, fragment.getVectorLength());
			return d;
		}
	}
	static public String nGramToString(int gram) {
		StringBuffer strGram = new StringBuffer();
		
		if (gram % (1 << 24) == 0) {
			strGram.append((byte)(gram >> 24));
		}
		else {
			String separator;
			if ((gram >>24) == 0) 
				separator = "--";
			else 
				separator = "<-";
			//gram = Math.abs(gram);
			gram %= 1 << 24;
			while (gram != 0) {
				if(strGram.length() != 0)
					strGram.insert(0, separator);
				strGram.insert(0, typeName[gram % 16]);
				gram = gram >> 4;
			}
		}
		
		return strGram.toString();
	}
	static public ArrayList<Byte> nGramToArray(int gram) {
		ArrayList<Byte> list = new java.util.ArrayList<Byte>();
		
		if (gram % (1 << 24) == 0) {
			list.add((byte)(gram >> 24));
		}
		else {
			if (gram > 0) 
				list.add((byte)0);
			else 
				list.add((byte)1);
			gram = Math.abs(gram);
			while (gram != 0) {
				list.add(1, (byte)(gram % 16));
				gram = gram >> 4;
			}
		}
		
		return list;
	}
	/*private byte getSizeOfGram(int gram) {
		byte i = 0;
		if (gram % (1 << 24) == 0) {
			i = 1;
		}
		else {			
	    	gram = Math.abs(gram);
	    	while (gram != 0) {
	    		i++;
	    		gram = gram / 16;
	    	}
		}
    	
    	return i;
    }*/
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\r\n");
		buf.append("FragID: " + this.id + "\t");
		buf.append(length + " characters from " + startChar + "\r\n");
		buf.append("File: " + this.sourceFile.fileLocation + "\t" + this.sourceFile.fileID + "\r\n");
		//buf.append(tokens.length + " tokens from " + startToken + ", ");
		//buf.append("Tokens from " + startToken + " to " + endToken + ", ");
		//buf.append("From line " + this.startLine + " to " + this.endLine + "\r\n");
		//buf.append("Vector: " + gramVector.size() + "/" + gram2Index.size() + "\r\n");
		/*for (int j = 0; j < Data.numOfNodeTypes - 1; j++) 
			buf.append((nGramVector.containsKey(j)?nGramVector.get(j):0)+",");
		buf.append(nGramVector.containsKey(Data.numOfNodeTypes - 1)?nGramVector.get(Data.numOfNodeTypes - 1):0);*/
		/*buf.append("Sparse: ");
		for (int i = 0; i < maxSizeOfGram; i++)
			for (int j = 0; j < gram2Index.size(); j++) 
				if (gramVector.containsKey(j) && (getSizeOfGram(index2Gram.get(j)) == i+1))
					buf.append(nGramToString(index2Gram.get(j)) + "(" + j + ")" + " = " + gramVector.get(j) + ",");*/
		/*buf.append("\r\n");
 		buf.append("Full: ");
 		int[] full = this.getFullGramVector();
 		for(int index = 0; index < gram2Index.size(); index++)
 			buf.append(full[index] + ", ");*/
		//buf.append("\r\nVector leng: " + this.getVectorLength());
		/*buf.append("\r\nClones: ");
 		for(Fragment f : this.clones)
 		{
 			HashMap<Integer, Integer> v1 = this.gramVector;
			HashMap<Integer, Integer> v2 = f.getSparseGramVector();
			HashSet<Integer> keys = new HashSet<Integer>();
			keys.addAll(v1.keySet());
			keys.addAll(v2.keySet());
			
			double d = 0;
			for(int key : keys)
			{
				int temp = (v1.containsKey(key) ? v1.get(key) : 0) - (v2.containsKey(key) ? v2.get(key) : 0);
				d += temp * temp;
			}
			d = Math.sqrt(d) * 2.0 / (this.vectorLength + f.getVectorLength());
 			
 			buf.append(f.id + ":" + d + "@" + keys.size() + " ");
 		}
 		buf.append("\r\n");*/
		//buf.append("\r\nDescendants: " + this.descendants + "\r\n");
		/*buf.append("\r\nHashcodes: ");
 		for(Bucket b : this.buckets)
 		{
 			buf.append(b.id + " ");
 		}
 		buf.append("\r\n");*/
		buf.append("\r\n-----------------------------------------------------------\r\n");
 		buf.append(code + "\r\n");
 		return buf.toString();
	}
}
