package cleman;

import java.util.Collections;
import java.util.HashSet;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Stack;
import java.io.*;

import util.Pair;
import util.PairDescendingOrder;

public class Group implements Serializable {
	private static final long serialVersionUID = 1L;
	
	public static int nextID = 1;
	
	public static HashMap<Integer, Group> all = new HashMap<Integer, Group>();
	public static HashSet<Integer> clones = new HashSet<Integer>(), changedClones = new HashSet<Integer>();
	//public static HashMap<Integer, HashSet<Integer>> revisionClones = new HashMap<Integer, HashSet<Integer>>();
	public static HashSet<Integer> prevClones = new HashSet<Integer>();
	public static HashSet<HashSet<Integer>> prevGroups = new HashSet<HashSet<Integer>>();
	public static HashSet<HashSet<Integer>> currentGroups = new HashSet<HashSet<Integer>>();
	
	public int id;
	public HashSet<Fragment> fragments = new HashSet<Fragment>();
	
	public Group(HashSet<Fragment> group)
	{
		this.id = nextID++;
		this.fragments = new HashSet<Fragment>(group);
		/*for(Fragment f : group)
			f.setGroup(this);*/
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.id;
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Group other = (Group) obj;
		if (id != other.id)
			return false;
		return true;
	}
	/**
	 * 
	 */
	public static void mapPairs(HashSet<Bucket> buckets)
	{
		ArrayList<Pair> pairs = new ArrayList<Pair>();
		PairDescendingOrder comparator = new PairDescendingOrder();
		HashMap<Fragment, HashSet<Pair>> mapPairs1 = new HashMap<Fragment, HashSet<Pair>>(), mapPairs2 = new HashMap<Fragment, HashSet<Pair>>();
		HashMap<Integer, HashSet<Integer>> checkeds = new HashMap<Integer, HashSet<Integer>>();
		for(Bucket b : buckets)
		{
			b.fragments.retainAll(Fragment.all.keySet());
			ArrayList<Integer> fragmentIDs = new ArrayList<Integer>(b.fragments);
			for(int i = 0; i < fragmentIDs.size()-1; i++)
				for(int j = i+1; j < fragmentIDs.size(); j++)
				{
					int ii = fragmentIDs.get(i), jj = fragmentIDs.get(j);
					if(checkeds.containsKey(ii) && checkeds.get(ii).contains(jj))
						continue;
					HashSet<Integer> checked = checkeds.get(ii);
					if(checked == null)
						checked = new HashSet<Integer>();
					checked.add(jj);
					checkeds.put(ii, checked);
					checked = checkeds.get(jj);
					if(checked == null)
						checked = new HashSet<Integer>();
					checked.add(ii);
					checkeds.put(jj, checked);
					Fragment fi = Fragment.all.get(ii);
					Fragment fj = Fragment.all.get(jj);
					if(fj == null)
					{
						System.err.println(Fragment.deleteds.containsKey(jj));
						System.err.println(Fragment.addeds.containsKey(jj));
						System.err.println(Fragment.modifieds.containsKey(jj));
						System.err.println(Fragment.unchangeds.containsKey(jj));
						System.err.println(b.fragments);
						System.err.println("jj: " + jj);
						for(Bucket bb : Fragment.deleteds.get(jj).getBuckets())
							System.err.println(bb.id + ": " + bb.fragments);
					}
					if(!fi.getClones().contains(jj))
					{
						double distance = fi.clonedTo(fj);
						if(distance <= Main.threshold)
						{
							fi.getClones().add(jj);
							fj.getClones().add(ii);
							clones.add(ii);
							clones.add(jj);
							Pair pair = new Pair(fi, fj, 1 - distance);
							int index = Collections.binarySearch(pairs, pair, comparator);
							if(index < 0)
								pairs.add(-1-index, pair);
							else
								pairs.add(index, pair);
							HashSet<Pair> pairs1 = mapPairs1.get(fi);
							if(pairs1 == null)
								pairs1 = new HashSet<Pair>();
							pairs1.add(pair);
							mapPairs1.put(fi, pairs1);
							HashSet<Pair> pairs2 = mapPairs2.get(fj);
							if(pairs2 == null)
								pairs2 = new HashSet<Pair>();
							pairs2.add(pair);
							mapPairs2.put(fj, pairs2);
						}
					}
				}
		}
		while(!pairs.isEmpty())
		{
			Pair pair = pairs.get(0);
			pairs.remove(0);
			Fragment fi = (Fragment)pair.getObj1(), fj = (Fragment)pair.getObj2();
			mapPairs1.get(fi).remove(pair);
			mapPairs2.get(fj).remove(pair);
			HashSet<Pair> tmpMapPairs = new HashSet<Pair>(mapPairs1.get(fi));
			for(Pair p : tmpMapPairs)
			{
				Fragment other = (Fragment)p.getObj2();
				if(other != fj)
				{
					int[] matches = fj.match(other);
					//if(matches[0]*1.0 / matches[1] >= 0.9)
					if(matches[0]*1.0 / Math.min(fj.getLength(), other.getLength()) >= 0.9)
					{
						pairs.remove(p);
						mapPairs1.get(fi).remove(p);
						mapPairs2.get(other).remove(p);
						fi.getClones().remove(other.getId());
						other.getClones().remove(fi.getId());
						if(other.getClones().isEmpty())
						{
							//other.setClones(null);
							clones.remove(other.getId());
						}
					}
				}
			}
			tmpMapPairs = new HashSet<Pair>(mapPairs2.get(fj));
			for(Pair p : tmpMapPairs)
			{
				Fragment other = (Fragment)p.getObj1();
				if(other != fi)
				{
					int[] matches = fi.match(other);
					//if(matches[0]*1.0 / matches[1] >= 0.9)
					if(matches[0]*1.0 / Math.min(fi.getLength(), other.getLength()) >= 0.9)
					{
						pairs.remove(p);
						mapPairs2.get(fj).remove(p);
						mapPairs1.get(other).remove(p);
						fj.getClones().remove(other.getId());
						other.getClones().remove(fj.getId());
						if(other.getClones().isEmpty())
						{
							//other.setClones(null);
							clones.remove(other.getId());
						}
					}
				}
			}
		}
	}
	
	public static void mapPairs2All(HashSet<Bucket> buckets)
	{
		ArrayList<Pair> pairs = new ArrayList<Pair>();
		PairDescendingOrder comparator = new PairDescendingOrder();
		HashMap<Integer, HashSet<Integer>> checkeds = new HashMap<Integer, HashSet<Integer>>();
		HashMap<Fragment, HashSet<Pair>> mapPairs1 = new HashMap<Fragment, HashSet<Pair>>(), mapPairs2 = new HashMap<Fragment, HashSet<Pair>>();
		for(Bucket b : buckets)
		{
			b.fragments.retainAll(Fragment.all.keySet());
			ArrayList<Integer> fragmentIDs = new ArrayList<Integer>(b.fragments);
			for(int i = 0; i < fragmentIDs.size()-1; i++)
				for(int j = i+1; j < fragmentIDs.size(); j++)
				{
					int ii = fragmentIDs.get(i), jj = fragmentIDs.get(j);
					if(checkeds.containsKey(ii) && checkeds.get(ii).contains(jj))
						continue;
					HashSet<Integer> checked = checkeds.get(ii);
					if(checked == null)
						checked = new HashSet<Integer>();
					checked.add(jj);
					checkeds.put(ii, checked);
					checked = checkeds.get(jj);
					if(checked == null)
						checked = new HashSet<Integer>();
					checked.add(ii);
					checkeds.put(jj, checked);
					Fragment fi = Fragment.all.get(ii);
					Fragment fj = Fragment.all.get(jj);
					if(fj == null)
					{
						System.err.println(Fragment.deleteds.containsKey(jj));
						System.err.println(Fragment.addeds.containsKey(jj));
						System.err.println(Fragment.modifieds.containsKey(jj));
						System.err.println(Fragment.unchangeds.containsKey(jj));
						System.err.println(b.fragments);
						System.err.println("jj: " + jj);
						for(Bucket bb : Fragment.deleteds.get(jj).getBuckets())
							System.err.println(bb.id + ": " + bb.fragments);
					}
					if(!fi.getClones().contains(jj))
					{
						double distance = fi.clonedTo(fj);
						if(distance <= Main.threshold)
						{
							fi.getClones().add(jj);
							fj.getClones().add(ii);
							changedClones.add(ii);
							changedClones.add(jj);
							clones.remove(ii);
							clones.remove(jj);
							Pair pair = new Pair(fi, fj, 1 - distance);
							int index = Collections.binarySearch(pairs, pair, comparator);
							if(index < 0)
								pairs.add(-1-index, pair);
							else
								pairs.add(index, pair);
							HashSet<Pair> pairs1 = mapPairs1.get(fi);
							if(pairs1 == null)
								pairs1 = new HashSet<Pair>();
							pairs1.add(pair);
							mapPairs1.put(fi, pairs1);
							HashSet<Pair> pairs2 = mapPairs2.get(fj);
							if(pairs2 == null)
								pairs2 = new HashSet<Pair>();
							pairs2.add(pair);
							mapPairs2.put(fj, pairs2);
						}
					}
				}
		}
		while(!pairs.isEmpty())
		{
			Pair pair = pairs.get(0);
			pairs.remove(0);
			Fragment fi = (Fragment)pair.getObj1(), fj = (Fragment)pair.getObj2();
			mapPairs1.get(fi).remove(pair);
			mapPairs2.get(fj).remove(pair);
			HashSet<Pair> tmpMapPairs = new HashSet<Pair>(mapPairs1.get(fi));
			for(Pair p : tmpMapPairs)
			{
				Fragment other = (Fragment)p.getObj2();
				if(other != fj)
				{
					int[] matches = fj.match(other);
					//if(matches[0]*1.0 / matches[1] >= 0.9)
					if(matches[0]*1.0 / Math.min(fj.getLength(), other.getLength()) >= 0.9)
					{
						pairs.remove(p);
						mapPairs1.get(fi).remove(p);
						mapPairs2.get(other).remove(p);
						fi.getClones().remove(other.getId());
						other.getClones().remove(fi.getId());
						if(other.getClones().isEmpty())
						{
							//other.setClones(null);
							changedClones.remove(other.getId());
						}
					}
				}
			}
			tmpMapPairs = new HashSet<Pair>(mapPairs2.get(fj));
			for(Pair p : tmpMapPairs)
			{
				Fragment other = (Fragment)p.getObj1();
				if(other != fi)
				{
					int[] matches = fi.match(other);
					//if(matches[0]*1.0 / matches[1] >= 0.9)
					if(matches[0]*1.0 / Math.min(fi.getLength(), other.getLength()) >= 0.9)
					{
						pairs.remove(p);
						mapPairs2.get(fj).remove(p);
						mapPairs1.get(other).remove(p);
						fj.getClones().remove(other.getId());
						other.getClones().remove(fj.getId());
						if(other.getClones().isEmpty())
						{
							//other.setClones(null);
							changedClones.remove(other.getId());
						}
					}
				}
			}
		}
	}
	
	public static HashSet<Integer> filterPairs(HashSet<Integer> clones)
	{
		HashSet<Integer> result = new HashSet<Integer>(), remain = new HashSet<Integer>(clones);
		int count = 0;
		for(int id1 : clones)
		{
			Fragment f1 = Fragment.all.get(id1);
			remain.remove(id1);
			for(int cid1 : new HashSet<Integer>(f1.getClones()))
			{
				Fragment cf1 = Fragment.all.get(cid1);
				SourceFile sf1 = f1.getSourceFile();
				HashSet<Integer> tmpIDs = new HashSet<Integer>(sf1.fragments);
				tmpIDs.retainAll(remain);
				for(int id2 : tmpIDs)
				{
					if(id2 != cid1)
					{
						Fragment f2 = Fragment.all.get(id2);
						for(int cid2 : new HashSet<Integer>(f2.getClones()))
						{
							Fragment cf2 = Fragment.all.get(cid2);
							if((f1.encloses(f2) && cf1.encloses(cf2)) || (f1.encloses(cf2) && cf1.encloses(f2)))
							{
								f2.getClones().remove(cid2);
								cf2.getClones().remove(id2);
							}
							else if((f2.encloses(f1) && cf2.encloses(cf1)) || (f2.encloses(cf1) && cf2.encloses(f1)))
							{
								f1.getClones().remove(cid1);
								cf1.getClones().remove(id1);
							}
						}
					}
				}
			}
			if(!f1.getClones().isEmpty())
				result.add(id1);
		}
		
		return result;
	}
	
	public static void filterChangedPairs()
	{
		HashSet<Integer> result = new HashSet<Integer>();//, remain = new HashSet<Integer>(all);
		changedClones = filterPairs(changedClones);
		for(int id1 : changedClones)
		{
			Fragment f1 = Fragment.all.get(id1);
			for(int cid1 : new HashSet<Integer>(f1.getClones()))
			{
				Fragment cf1 = Fragment.all.get(cid1);
				SourceFile sf1 = f1.getSourceFile();
				HashSet<Integer> tmpIDs = new HashSet<Integer>(sf1.fragments);
				tmpIDs.retainAll(clones);
				for(int id2 : tmpIDs)
				{
					if(id2 != cid1)
					{
						Fragment f2 = Fragment.all.get(id2);
						for(int cid2 : new HashSet<Integer>(f2.getClones()))
						{
							Fragment cf2 = Fragment.all.get(cid2);
							if((f1.encloses(f2) && cf1.encloses(cf2)) || (f1.encloses(cf2) && cf1.encloses(f2)))
							{
								f2.getClones().remove(cid2);
								cf2.getClones().remove(id2);
							}
							else if((f2.encloses(f1) && cf2.encloses(cf1)) || (f2.encloses(cf1) && cf2.encloses(f1)))
							{
								f1.getClones().remove(cid1);
								cf1.getClones().remove(id1);
							}
						}
					}
				}
			}
			if(!f1.getClones().isEmpty())
				result.add(id1);
		}
		if(!result.isEmpty())
		{
			for(int id : clones)
			{
				Fragment f = Fragment.all.get(id);
				if(!f.getClones().isEmpty())
					result.add(id);
			}
			clones = result;
		}
	}
	
	public static HashMap<Integer, Group> group(HashSet<Integer> clones)
	{
		HashMap<Integer, Group> groups = new HashMap<Integer, Group>();
		HashSet<Integer> tmpClones = new HashSet<Integer>(clones);
		while(!tmpClones.isEmpty())
		{
			int id = (Integer)tmpClones.toArray()[0];
			HashSet<Integer> group = new HashSet<Integer>();
			Stack<Integer> stack = new Stack<Integer>();
			stack.push(id);
			//System.err.print("Start a group:");
			while(!stack.isEmpty())
			{
				int id1 = stack.pop();
				group.add(id1);
				//System.err.print(" " + id1);
				tmpClones.remove(id1);
				for(int id2 : Fragment.all.get(id1).getClones())
					if(tmpClones.contains(id2))
						stack.push(id2);
			}
			//System.err.println();
			HashSet<Fragment> fs = new HashSet<Fragment>();
			for(int fid : group)
				fs.add(Fragment.all.get(fid));
			Group g = new Group(fs);
			groups.put(g.id, g);
		}
		
		return groups;
	}
	/*public void removeFragment(Fragment fragment)
	{
		fragment.setGroup(null);
		this.fragments.remove(fragment);
	}*/
	public void selfFilter()
	{
		ArrayList<Fragment> fragments = new ArrayList<Fragment>(this.fragments);
		for(int i = 0; i < fragments.size()-1; i++)
			for(int j = i+1; j < fragments.size(); j++)
			{
				Fragment fi = fragments.get(i);
				Fragment fj = fragments.get(j);
				if(fi.encloses(fj))
					this.fragments.remove(fj);
				if(fj.encloses(fi))
					this.fragments.remove(fi);
			}
	}
	public boolean isEnclosedBy(Group other)
	{
		if(this == other)
			return false;
		boolean isEnclosed = true;
		for(Fragment thisFrag : this.fragments)
		{
			boolean oneEnclosed = false;
			for(Fragment otherFrag : other.fragments)
			{
				oneEnclosed = otherFrag.encloses(thisFrag);
				if(oneEnclosed)
					break;
			}
			if(!oneEnclosed)
			{
				isEnclosed = false;
				break;
			}
		}
		return (isEnclosed);
	}
	/*public boolean isEnclosedBy(Group group)
	{
		HashSet<Integer> otherDescendants = new HashSet<Integer>();
		for(Fragment f : group.fragments)
			otherDescendants.addAll(f.getDescendants());
		
		HashSet<Integer> thisFragments = new HashSet<Integer>();
		for(Fragment f : this.fragments)
			thisFragments.add(f.getId());
		
		return otherDescendants.containsAll(thisFragments);
	}*/
	public boolean isEnclosedBy(HashMap<Integer, Group> groups)
	{
		for(Group g : groups.values())
			if(isEnclosedBy(g))
			{
				return true;
			}
		return false;
	}
	public static HashMap<Integer, Group> crossFilter(HashMap<Integer, Group> groups)
	{
		for(Group g : groups.values())
		{
			g.selfFilter();
		}
		HashMap<Integer, Group> result = new HashMap<Integer, Group>(groups);
		for(Group g : groups.values())
		{
			//System.out.println("Group " + g.id + " has size " + g.fragments.size() + "\t" + g.isEnclosedBy(result));
			if(g.fragments.size() < 2 || g.isEnclosedBy(result))
			{
				/*for(Fragment f : g.fragments)
					f.setGroup(null);*/
				result.remove(g.id);
				for(Fragment f : g.fragments)
				{
					for(Fragment cf : g.fragments)
						f.getClones().remove(cf.getId());
					if(f.getClones().isEmpty())
					{
						//f.setClones(null);
						clones.remove(f.getId());
					}
				}
			}
		}
		return result;
	}
	public void report(String path)
	{
		File dir = new File(path);
		if (!dir.exists()) dir.mkdirs();
		dir = new File(path);
		if (!dir.exists()) dir.mkdirs();
		try {
			BufferedWriter fout = new BufferedWriter(new FileWriter(dir.getAbsolutePath() + "\\" + String.format("%05d", this.id) + ".txt"));
			fout.write("Group " + String.format("%05d", this.id) + " containing " + fragments.size() + " clones.\r\n");		
			for (Fragment fr : this.fragments) {			
		 		fout.write(fr.toString());
			}
			fout.flush();
			fout.close();
		}
		catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static void reportPairs(String path)
	{
		try {
			BufferedWriter fout = new BufferedWriter(new FileWriter(path));
			for(int id : clones)
			{
				Fragment f = Fragment.all.get(id);
				fout.write(id + ": ");
				for(int cid : f.getClones())
				{
					fout.write("\t" + cid);
				}
				fout.write("\n");
			}
			fout.flush();
			fout.close();
		}
		catch (IOException e) {
			e.printStackTrace();
		}

	}
}