/**
 * 
 */
package cleman;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;

import report.RptSourceFile;

/**
 * @author Nguyen Anh Hoan
 *
 */

public class ReportLines implements Serializable {
	protected static final long serialVersionUID = 1L;
	
	public String projectURL;
	public String reportName;
	public long revision;
	public String logFile;
	public int numOfFiles;
	public int numOfGroups;
	public int numOfLOC = 0;
	public int numOfCLOC = 0;
	public int numOfClones = 0;
	public int numOfPortions = 0;
	public HashMap<String, ArrayList<Integer>> linesOfFiles = new HashMap<String, ArrayList<Integer>>();
	public HashMap<String, Integer> totalClonedLinesOfFiles = new HashMap<String, Integer>();
	public HashMap<String, ArrayList<int[]>> allClonedLinesOfFiles = new HashMap<String, ArrayList<int[]>>(); // 0-based
	
	public ReportLines()
	{

	}
	
	/*public Report(HashMap<Integer, Group> groups, HashMap<Integer, Fragment> fragments, HashMap<String, SourceFile> files, int revision)
	{
		this.projectURL = Main.projectURL;
		this.revision = revision;
		this.logFile = System.out.toString();
		this.numOfFiles = files.size();
		this.numOfGroups = groups.size();
		
		for(String fileName : files.keySet())
		{
			SourceFile file = files.get(fileName);
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			try {
				Main.repository.getFile(fileName, revision, null, out);
			}
			catch (SVNException e) {
				if(e.getMessage().contains("path not found"))
					continue;
				else
					e.printStackTrace();
			}
			ArrayList<Integer> linesOfFile = new ArrayList<Integer>();
			String fileContent = out.toString();
			int charCount = -1;
	    	do{
	    		linesOfFile.add(++charCount);
	    		charCount = fileContent.indexOf("\n", charCount);
	    	} while(charCount > -1);
	    	numOfLOC += linesOfFile.size();
	    	linesOfFiles.put(fileName, linesOfFile);
	    	
	    	boolean[] lines = new boolean[linesOfFile.size()];
	    	for(int i = 0; i < lines.length; i++)
	    		lines[i] = false;
	    	for(Fragment f : file.fragments)
	    	{
	    		if(f.getGroup() != null)
	    		{
	    			int start = Main.binSearch(f.getStartChar(), linesOfFile);
	    			int end = Main.binSearch(f.getStartChar() + f.getLength() - 1, linesOfFile);
	    			//System.err.println(start + " -> " + end + ": " + f.getStartChar() + " + " + f.getLength());
	    			for(int i = start; i <= end; i++)
	    				lines[i] = true;
	    			this.numOfClones++;
	    		}
	    	}
	    	int CLsinFile = 0;
    		for(int i = 0; i < lines.length; i++)
    		{
    			if(lines[i])
    			{
    				numOfCLOC++;
    				CLsinFile++;
    			}
    		}
    		totalClonedLinesOfFiles.put(fileName, CLsinFile);
    		
    		ArrayList<int[]> allClonedLinesOfFile = new ArrayList<int[]>();
    		int j = 0;
    		
    		while(j < lines.length)
    		{
    			while(j < lines.length && !lines[j])
    				j++;
    			if(j < lines.length)
    			{
	    			int start = j++;
	    			while(j < lines.length && lines[j])
	    				j++;
	    			int[] pair = new int[2];
	    			pair[0] = start;
	    			pair[1] = j-1;
	    			allClonedLinesOfFile.add(pair);
    			}
    		}
    		allClonedLinesOfFiles.put(fileName, allClonedLinesOfFile);
    		numOfPortions += allClonedLinesOfFile.size();
		}
	}*/
	/*
	 * 
	 */
	public void summarize()
	{
		System.out.println("Files: " + this.numOfFiles);
		System.out.println("Groups: " + this.numOfGroups);
		System.out.println("Lines: " + this.numOfCLOC + " / " + this.numOfLOC);
	}
	public void compareTo(ReportLines other)
	{
		System.out.println(this.reportName + ": " + this.numOfCLOC + " / " + this.numOfLOC);
		System.out.println(other.getReportName() + ": " + other.getNumOfCLOC() + " / " + other.getNumOfLOC());
		System.out.println(this.reportName + ": " + this.numOfFiles + " files");
		System.out.println(other.getReportName() + ": " + other.getNumOfFiles() + " files");
		System.out.println(this.reportName + ": " + this.numOfGroups + " groups");
		System.out.println(other.getReportName() + ": " + other.getNumOfGroups() + " groups");
		
		int inThis = 0, inOther = 0, inBoth = 0;
		for(String fileName : this.linesOfFiles.keySet())
		{
			boolean[] thisLines = new boolean[linesOfFiles.get(fileName).size()];
			for(int i = 0; i < thisLines.length; i++)
				thisLines[i] = false;
			if(this.allClonedLinesOfFiles.containsKey(fileName))
				for(int[] pair : this.allClonedLinesOfFiles.get(fileName))
					for(int i = pair[0]; i <= pair[1]; i++)
						thisLines[i] = true;
			
			boolean[] otherLines = new boolean[linesOfFiles.get(fileName).size()];
			for(int i = 0; i < otherLines.length; i++)
				otherLines[i] = false;
			if(other.getAllClonedLinesOfFiles().containsKey(fileName) && other.getAllClonedLinesOfFiles().get(fileName).size() > 0)
			{
				for(int index = 0; index < other.getAllClonedLinesOfFiles().get(fileName).size()-1; index++)
				{
					int[] pair = other.getAllClonedLinesOfFiles().get(fileName).get(index);
					for(int i = pair[0]; i <= pair[1]; i++)
						otherLines[i] = true;
				}
				int[] pair = other.getAllClonedLinesOfFiles().get(fileName).get(other.getAllClonedLinesOfFiles().get(fileName).size()-1);
				for(int i = pair[0]; i <= Math.min(pair[1], otherLines.length-1); i++)
					otherLines[i] = true;
			}
			
			for(int i = 0; i < thisLines.length; i++)
			{
				if(thisLines[i] && otherLines[i])
					inBoth++;
				if(thisLines[i] && !otherLines[i])
				{
					inThis++;
					//System.out.println("In " + this.reportName + " not in " + other.getReportName() + ": " + fileName + "@line " + i);
				}
				if(!thisLines[i] && otherLines[i])
				{
					inOther++;
					//System.out.println("In " + other.getReportName() + " not in " + this.reportName + ": " + fileName + "@line " + i);
				}
			}
			/*System.out.println(fileName + ": " + 
					this.allClonedLinesOfFiles.get(fileName).size() + " - " + other.getAllClonedLinesOfFiles().get(fileName).size() + "\t" +
					this.totalClonedLinesOfFiles.get(fileName) + " - " + other.getTotalClonedLinesOfFiles().get(fileName));*/
		}
		System.out.println("In both: " + inBoth);
		System.out.println("In " + this.reportName + " not in " + other.getReportName() + ": " + inThis);
		System.out.println("In " + other.getReportName() + " not in " + this.reportName + ": " + inOther);
		System.out.println("Clones: " + this.numOfClones + " - " + other.getNumOfClones());
		System.out.println("Portions: " + this.numOfPortions + " - " + other.getNumOfPortions());
	}
	/*This method will be used to read from a file*/
	public static Object loadObjectFromFile(String rFileName)
	{	   
		Object result = null;
		File data = new File(rFileName);
		if (data.exists()) {
			try{
				ObjectInputStream in = new ObjectInputStream(new FileInputStream(data)); 			     
				result =  in.readObject();
				in.close();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		return result;
	}
	public static HashMap<String, RptSourceFile> getImportRanges(String srcDir, String relativePath)
	{
		HashMap<String, RptSourceFile> importRanges = new HashMap<String, RptSourceFile>();
		File file = new File(srcDir + "\\" + relativePath);
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				importRanges.putAll(getImportRanges(srcDir, relativePath + "\\" + sub.getName()));
		}
		else if(file.getName().endsWith(".java"))
		{
			StringBuffer strBuf = new StringBuffer();
			ArrayList<Integer> lines = new ArrayList<Integer>();
			lines.add(0);
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	String line = "";
		    	int charCount = 0;
		    	while ((line = in.readLine()) != null) { 
		    		strBuf.append(line + "\n");
		    		lines.add(charCount);
		    		charCount += line.length() + 1;
		    	}
		    	in.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			ASTParser parser = ASTParser.newParser(AST.JLS3);
	    	parser.setSource(strBuf.toString().toCharArray());
	    	CompilationUnit ast = (CompilationUnit)parser.createAST(null); 
	    	TypeDeclaration td = (TypeDeclaration)ast.types().get(0);
	    	RptSourceFile rsf = new RptSourceFile();
	    	rsf.endImportLine = ast.getLineNumber(td.getStartPosition()) - 1;
	    	rsf.content = strBuf.toString();
	    	rsf.cu = ast;
	    	rsf.lineStarts = lines;
	    	importRanges.put(relativePath, rsf);
		}
		
		return importRanges;
	}
	public static ReportLines loadDeckardReport(String logName)
	{
		int[] counts = new int[5];
		HashMap<String, RptSourceFile> importRanges = ReportLines.getImportRanges(new File(logName).getParent(), "src");
		String DK_LINE = "LINE:", DK_FILE = "FILE ";
		ReportLines report = new ReportLines();
		report.setReportName("Deckard");
		report.setLogFile(logName);
		HashMap<String, ArrayList<Boolean>> linesOfFiles = new HashMap<String, ArrayList<Boolean>>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) {
	    		while((line = in.readLine()) != null && line.trim().isEmpty())
		    		line = in.readLine();
	    		if(line == null)
	    			break;
	    		ArrayList<CloneFragment> group = new ArrayList<CloneFragment>();
	    		do
	    		{
	    			int start = line.indexOf(DK_FILE);
	    			if(start == -1)
	    				continue;
	    			start += DK_FILE.length();
	    			int end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			String fileName = line.substring(start, end);
	    			fileName = Util.replaceAll(fileName, '/', '\\');
	    			start = line.indexOf(DK_LINE, end+1);
	    			if(start == -1)
	    				continue;
	    			start += DK_LINE.length();
	    			end = line.indexOf(':', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int sl = Integer.parseInt(line.substring(start, end));
	    			/*if(sl <= importRanges.get(fileName).endImportLine)
	    				continue;*/
	    			start = end + 1;
	    			end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int el = sl + Integer.parseInt(line.substring(start, end)) - 1;
	    			CloneFragment cf = new CloneFragment();
	    			cf.fileName = fileName;
	    			cf.startLine = sl; cf.endLine = el;
	    			group.add(cf);
	    		} while ((line = in.readLine()) != null && !line.trim().isEmpty());
	        	for(int i = 0; i < group.size()-1; i++)
	    		{
	    			CloneFragment cf1 = group.get(i);
	    			RptSourceFile rsf1 = importRanges.get(cf1.fileName);
	    			int fragmentType1 = ReportPairs.INDEX_GOOD;
	    			if (cf1.startLine <= rsf1.endImportLine)
	    				fragmentType1 = ReportPairs.INDEX_IMPORT;
	    			else
	    			{
	    				if(Util.isCrossMethod(rsf1.cu, cf1.startLine, cf1.endLine))
	    					fragmentType1 = ReportPairs.INDEX_METHOD;
	    			}
	    			for(int j = i+1; j < group.size(); j++)
	    			{
	    				CloneFragment cf2 = group.get(j);
	    				if(!cf1.isOverlaped(cf2))
	    				{
	    					/*counts[ReportPairs.INDEX_TOTAL]++;
	    					RptSourceFile rsf2 = importRanges.get(cf2.fileName);
	    	    			int fragmentType2 = ReportPairs.INDEX_GOOD;
	    	    			if (cf2.startLine <= rsf2.endImportLine)
	    	    				fragmentType2 = ReportPairs.INDEX_IMPORT;
	    	    			else
	    	    			{
	    	    				if(Util.isCrossMethod(rsf2.cu, cf2.startLine, cf2.endLine))
	    	    					fragmentType2 = ReportPairs.INDEX_METHOD;
	    	    			}
	    	    			if(fragmentType1 == ReportPairs.INDEX_IMPORT || fragmentType2 == ReportPairs.INDEX_IMPORT)
	    	    				counts[ReportPairs.INDEX_IMPORT]++;
	    	    			else if(fragmentType1 == ReportPairs.INDEX_METHOD || fragmentType2 == ReportPairs.INDEX_METHOD)
	    	    				counts[ReportPairs.INDEX_METHOD]++;
	    	    			else*/
	    					{
		    					counts[ReportPairs.INDEX_GOOD]++;
		    					ArrayList<Boolean> lines = linesOfFiles.get(cf1.fileName);
		    	    			if(lines == null)
		    	    				lines = new ArrayList<Boolean>();
		    	    			int cl = lines.size();
		        				for(int ii = cl-1; ii < cf1.endLine; ii++)
		        					lines.add(false);
		    	    			for(int ii = 0; ii < cf1.endLine-cf1.startLine+1; ii++)
		    	    			{
		    	    				lines.set(cf1.startLine + ii - 1, true);
		    	    			}
		    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
		    	    			linesOfFiles.put(cf1.fileName, lines);
		    	    			
		    	    			lines = linesOfFiles.get(cf2.fileName);
		    	    			if(lines == null)
		    	    				lines = new ArrayList<Boolean>();
		    	    			cl = lines.size();
		        				for(int ii = cl-1; ii < cf2.endLine; ii++)
		        					lines.add(false);
		    	    			for(int ii = 0; ii < cf2.endLine-cf2.startLine+1; ii++)
		    	    			{
		    	    				lines.set(cf2.startLine + ii - 1, true);
		    	    			}
		    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
		    	    			linesOfFiles.put(cf2.fileName, lines);
	    					}
	    				}
	    			}
	    		}
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	/*HashMap<String, RptSourceFile> importRanges = getImportRanges(new File(logName).getParent(), "src");
		String DK_LINE = "LINE:", DK_FILE = "FILE ";
		ReportLines report = new ReportLines();
		report.setReportName("Deckard");
		report.setLogFile(logName);
		HashMap<String, ArrayList<Boolean>> linesOfFiles = new HashMap<String, ArrayList<Boolean>>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
	    		if(!line.trim().isEmpty())
	    		{
	    			int start = line.indexOf(DK_FILE);
	    			if(start == -1)
	    				continue;
	    			start += DK_FILE.length();
	    			int end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			String fileName = line.substring(start, end);
	    			fileName = Util.replaceAll(fileName, '/', '\\');
	    			ArrayList<Boolean> lines = linesOfFiles.get(fileName);
	    			if(lines == null)
	    				lines = new ArrayList<Boolean>();
	    			start = line.indexOf(DK_LINE, end+1);
	    			if(start == -1)
	    				continue;
	    			start += DK_LINE.length();
	    			end = line.indexOf(':', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int sl = Integer.parseInt(line.substring(start, end));
	    			if(sl <= importRanges.get(fileName).endImportLine)
	    				continue;
	    			start = end + 1;
	    			end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int len = Integer.parseInt(line.substring(start, end));
	    			int cl = lines.size();
    				for(int i = cl-1; i < sl+len-1; i++)
    					lines.add(false);
	    			for(int i = 0; i < len; i++)
	    			{
	    				lines.set(sl + i - 1, true);
	    			}
	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
	    			linesOfFiles.put(fileName, lines);
	    		}
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}*/
    	for(String fileName : linesOfFiles.keySet())
    	{
    		ArrayList<Boolean> lines = linesOfFiles.get(fileName);
			ArrayList<int[]> allClonedLinesOfFile = new ArrayList<int[]>();
    		int j = 0;
    		
    		while(j < lines.size())
    		{
    			while(j < lines.size() && !lines.get(j))
    				j++;
    			if(j < lines.size())
    			{
    				int start = j++;
	    			while(j < lines.size() && lines.get(j))
	    				j++;
	    			int[] pair = new int[2];
	    			pair[0] = start;
	    			pair[1] = j-1;
	    			allClonedLinesOfFile.add(pair);
	    			report.numOfCLOC += j - start;
    			}
    		}
    		report.allClonedLinesOfFiles.put(fileName, allClonedLinesOfFile);
    	}
		
		return report;
	}
	public static int parseHex(char ch)
	{
		if(ch >= '0' && ch <= '9')
			return ch - '0';
		ch = Character.toLowerCase(ch);
		return 10 + ch - 'a';
	}
	public static int parseHex(String str)
	{
		int v = 0;
		for(int i = 0; i < str.length(); i++)
			v = v * 16 + parseHex(str.charAt(i));
		return v;
	}
	private static HashMap<String, ArrayList<Integer>> readTokensOfFiles(String root, File file)
	{
		HashMap<String, ArrayList<Integer>> tokensOfFiles = new HashMap<String, ArrayList<Integer>>();
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				tokensOfFiles.putAll(readTokensOfFiles(root, sub));
		}
		else if(file.getName().endsWith(".java.2_0_0_0.default.ccfxprep"))
		{
			ArrayList<Integer> tokensOfFile = new ArrayList<Integer>();
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	String line = "";
		    	while ((line = in.readLine()) != null) {
		    		int end = line.indexOf('.');
		    		tokensOfFile.add(parseHex(line.substring(0, end)));
		    	}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
			String fileName = file.getAbsolutePath().substring(root.length()+1, (int)(file.getAbsolutePath().length()-".java.2_0_0_0.default.ccfxprep".length()));
			tokensOfFiles.put(fileName, tokensOfFile);
		}
		
		return tokensOfFiles;
	}
	public static ReportLines loadCCFinderReport(String logName)
	{
		ReportLines report = new ReportLines();
		report.setReportName("CCFinderX");
		report.setLogFile(logName);
		String srcDir = "";
		HashMap<String, ArrayList<Integer>> tokensOfFiles = new HashMap<String, ArrayList<Integer>>();
		ArrayList<String> fileNames = new ArrayList<String>();
		HashMap<String, ArrayList<Boolean>> linesOfFiles = new HashMap<String, ArrayList<Boolean>>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("option: -n "))
		    	{
		    		srcDir = line.substring("option: -n ".length()).trim();
		    		tokensOfFiles = readTokensOfFiles(srcDir + "\\.ccfxprepdir", new File(srcDir + "\\.ccfxprepdir"));
		    		break;
		    	}
	    	}
	    	
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("source_files {"))
		    		break;
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.trim().equals("}"))
		    		break;
		    	String[] parts = line.split("\t");
		    	fileNames.add(parts[1].trim());
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("clone_pairs {"))
		    		break;
	    	}
	    	while ((line = in.readLine()) != null) { 
	    		if(line.trim().equals("}"))
		    		break;
		    	String[] parts = line.split("\t");
		    	String cloneInfo = parts[1].trim();
	    		int start = 0, end = cloneInfo.indexOf('.');
	    		int fid = Integer.parseInt(cloneInfo.substring(start, end));
	    		String fileName = fileNames.get(fid-1).substring(srcDir.length()+1);
	    		ArrayList<Boolean> lines = linesOfFiles.get(fileName);
    			if(lines == null)
    				lines = new ArrayList<Boolean>();
    			start = end+1;
    			end = cloneInfo.indexOf('-', start);
    			int sl = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start, end)));
    			start = end + 1;
    			int el = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start)));
    			int cl = lines.size();
				for(int i = cl; i < el; i++)
					lines.add(false);
    			for(int i = sl; i <= el; i++)
    			{
    				lines.set(i - 1, true);
    			}
    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
    			linesOfFiles.put(fileName, lines);
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	for(String fileName : linesOfFiles.keySet())
    	{
    		ArrayList<Boolean> lines = linesOfFiles.get(fileName);
			ArrayList<int[]> allClonedLinesOfFile = new ArrayList<int[]>();
    		int j = 0;
    		
    		while(j < lines.size())
    		{
    			while(j < lines.size() && !lines.get(j))
    				j++;
    			if(j < lines.size())
    			{
	    			int start = j++;
	    			while(j < lines.size() && lines.get(j))
	    				j++;
	    			int[] pair = new int[2];
	    			pair[0] = start;
	    			pair[1] = j-1;
	    			allClonedLinesOfFile.add(pair);
	    			report.numOfCLOC += j - start;
    			}
    		}
    		report.allClonedLinesOfFiles.put(fileName, allClonedLinesOfFile);
    	}
		
		return report;
	}
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub
		/*String path = "D:\\Iowa State Uni\\Projects\\FSE2009\\test\\";
		Report sta = (Report)loadObjectFromFile(path + "_clemanX_Standalone_28_28_0.9_16_4_4_50.dat");
		Report inc = (Report)loadObjectFromFile(path + "_clemanX_Incremental_20_28_0.9_16_4_4_50.dat");*/
		String path = "D:\\Iowa State Uni\\Projects\\FSE2009\\columba\\";
		ReportLines sta = (ReportLines)loadObjectFromFile(path + "_clemanX_Standalone_362_362_1_0.9_16_4_4_50.dat");
		ReportLines inc = (ReportLines)loadObjectFromFile(path + "_clemanX_Incremental_21_362_1_0.9_16_4_4_50.dat");
		/*String path = "D:\\Iowa State Uni\\Projects\\FSE2009\\jEdit\\";
		Report sta = (Report)loadObjectFromFile(path + "_clemanX_Standalone_14000_14000_1000_0.9_16_4_4_50.dat");
		Report inc = (Report)loadObjectFromFile(path + "_clemanX_Incremental_10000_14000_1000_0.9_16_4_4_50.dat");*/
		
		System.out.println("Usage: projectPath standaloneReport incrementalReport");
		if(args.length > 0)
		{
			path = args[0];
			sta = (ReportLines)loadObjectFromFile(path + "\\" + args[1]);
			inc = (ReportLines)loadObjectFromFile(path + "\\" + args[2]);
		}
		else
			return;
		String logName = args[2];
		logName = logName.replace("_Incremental_", "_reportComparision_");
		logName = logName.replace(".dat", ".txt");
		
		File logFile = new File(path + "\\" + logName);
		try{
			if(!logFile.exists())
				logFile.createNewFile();
			//System.setOut(new PrintStream(logFile));
		}
		catch (IOException e) {
			e.printStackTrace();
		}
		sta.setReportName("Standalone");
		inc.setReportName("Incremental");
		//System.out.println(sta.getAllClonedLinesOfFiles().size());
		/*for(ArrayList<int[]> a : sta.getAllClonedLinesOfFiles().values())
			System.out.println(a.size());*/
			
		inc.compareTo(sta);
		System.out.flush();
	}

	/**
	 * @return the projectURL
	 */
	public String getProjectURL() {
		return projectURL;
	}

	/**
	 * @param projectURL the projectURL to set
	 */
	public void setProjectURL(String projectURL) {
		this.projectURL = projectURL;
	}

	/**
	 * @return the reportName
	 */
	public String getReportName() {
		return reportName;
	}

	/**
	 * @param reportName the reportName to set
	 */
	public void setReportName(String reportName) {
		this.reportName = reportName;
	}
	
	/**
	 * @return the revision
	 */
	public long getRevision() {
		return revision;
	}

	/**
	 * @param revision the revision to set
	 */
	public void setRevision(long revision) {
		this.revision = revision;
	}

	/**
	 * @return the logFile
	 */
	public String getLogFile() {
		return logFile;
	}

	/**
	 * @param logFile the logFile to set
	 */
	public void setLogFile(String logFile) {
		this.logFile = logFile;
	}

	/**
	 * @return the numOfFiles
	 */
	public int getNumOfFiles() {
		return numOfFiles;
	}

	/**
	 * @param numOfFiles the numOfFiles to set
	 */
	public void setNumOfFiles(int numOfFiles) {
		this.numOfFiles = numOfFiles;
	}

	/**
	 * @return the numOfGroups
	 */
	public int getNumOfGroups() {
		return numOfGroups;
	}

	/**
	 * @param numOfGroups the numOfGroups to set
	 */
	public void setNumOfGroups(int numOfGroups) {
		this.numOfGroups = numOfGroups;
	}

	/**
	 * @return the numOfLOC
	 */
	public int getNumOfLOC() {
		return numOfLOC;
	}

	/**
	 * @param numOfLOC the numOfLOC to set
	 */
	public void setNumOfLOC(int numOfLOC) {
		this.numOfLOC = numOfLOC;
	}

	/**
	 * @return the numOfCLOC
	 */
	public int getNumOfCLOC() {
		return numOfCLOC;
	}

	/**
	 * @param numOfCLOC the numOfCLOC to set
	 */
	public void setNumOfCLOC(int numOfCLOC) {
		this.numOfCLOC = numOfCLOC;
	}

	/**
	 * @return the linesOfFiles
	 */
	public HashMap<String, ArrayList<Integer>> getLinesOfFiles() {
		return linesOfFiles;
	}

	/**
	 * @param linesOfFiles the linesOfFiles to set
	 */
	public void setLinesOfFiles(HashMap<String, ArrayList<Integer>> linesOfFiles) {
		this.linesOfFiles = linesOfFiles;
	}

	/**
	 * @return the totalClonedLinesOfFiles
	 */
	public HashMap<String, Integer> getTotalClonedLinesOfFiles() {
		return totalClonedLinesOfFiles;
	}

	/**
	 * @param totalClonedLinesOfFiles the totalClonedLinesOfFiles to set
	 */
	public void setTotalClonedLinesOfFiles(
			HashMap<String, Integer> totalClonedLinesOfFiles) {
		this.totalClonedLinesOfFiles = totalClonedLinesOfFiles;
	}

	/**
	 * @return the allClonedLinesOfFiles
	 */
	public HashMap<String, ArrayList<int[]>> getAllClonedLinesOfFiles() {
		return allClonedLinesOfFiles;
	}

	/**
	 * @param allClonedLinesOfFiles the allClonedLinesOfFiles to set
	 */
	public void setAllClonedLinesOfFiles(
			HashMap<String, ArrayList<int[]>> allClonedLinesOfFiles) {
		this.allClonedLinesOfFiles = allClonedLinesOfFiles;
	}

	/**
	 * @return the numOfClones
	 */
	public int getNumOfClones() {
		return numOfClones;
	}

	/**
	 * @param numOfClones the numOfClones to set
	 */
	public void setNumOfClones(int numOfClones) {
		this.numOfClones = numOfClones;
	}

	/**
	 * @return the numOfPortions
	 */
	public int getNumOfPortions() {
		return numOfPortions;
	}

	/**
	 * @param numOfPortions the numOfPortions to set
	 */
	public void setNumOfPortions(int numOfPortions) {
		this.numOfPortions = numOfPortions;
	}

	
}
