package cleman;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Random;

import org.eclipse.jdt.core.dom.ASTNode;

import report.RptSourceFile;

public class ReportPairs implements Serializable {
	protected static final long serialVersionUID = 1L;
	
	public static final int INDEX_TOTAL = 0, INDEX_IMPORT = 1, INDEX_METHOD = 2, INDEX_STATEMENT = 3, INDEX_GOOD = 4;
	public static double THRES_MATCH = 0.7;
	private String projectURL;
	private String reportName;
	private long revision;
	private String logFile;
	private HashSet<ClonePair> clonePairs = new HashSet<ClonePair>();
	public int numOfLOC = 0;
	public int numOfCLOC = 0;
	public int numOfClones = 0;
	public int numOfPortions = 0;
	public HashMap<String, ArrayList<Integer>> linesOfFiles = new HashMap<String, ArrayList<Integer>>();
	public HashMap<String, Integer> totalClonedLinesOfFiles = new HashMap<String, Integer>();
	public HashMap<String, ArrayList<int[]>> allClonedLinesOfFiles = new HashMap<String, ArrayList<int[]>>(); // 0-based
	
	public ReportPairs() {
		// TODO Auto-generated constructor stub
	}
	
	public String getProjectURL() {
		return projectURL;
	}

	public void setProjectURL(String projectURL) {
		this.projectURL = projectURL;
	}

	public String getReportName() {
		return reportName;
	}

	public void setReportName(String reportName) {
		this.reportName = reportName;
	}

	public long getRevision() {
		return revision;
	}

	public void setRevision(long revision) {
		this.revision = revision;
	}

	public String getLogFile() {
		return logFile;
	}

	public void setLogFile(String logFile) {
		this.logFile = logFile;
	}

	public HashSet<ClonePair> getClonePairs() {
		return clonePairs;
	}

	public void setClonePairs(HashSet<ClonePair> clonePairs) {
		this.clonePairs = clonePairs;
	}

	public static void splitBellonReferences(String logName)
	{
		String PROJECT = "COPY \"project\" FROM stdin;", REFERENCE = "COPY \"reference\" FROM stdin;";
		ArrayList<String> systemNames = new ArrayList<String>();
		ArrayList<Integer> systemSizes = new ArrayList<Integer>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null && !line.equals(PROJECT));
	    	while ((line = in.readLine()) != null && !line.trim().isEmpty()) {
	    		String[] parts = line.split("\t");
	    		systemNames.add(parts[1]);
	    		systemSizes.add(Integer.parseInt(parts[4]));
	    	}
	    	while ((line = in.readLine()) != null && !line.equals(REFERENCE));
	    	BufferedWriter[] outs = new BufferedWriter[systemNames.size()];
	    	for(int i = 0; i < outs.length; i++)
	    	{
	    		outs[i] = new BufferedWriter(new FileWriter("references_" + systemNames.get(i) + "_" + systemSizes.get(i) + ".txt"));
	    	}
	    	while ((line = in.readLine()) != null)
	    	{
	    		String[] parts = line.split("\t");
	    		int index = parts[2].indexOf('/');
	    		parts[2] = parts[2].substring(index+1);
	    		parts[2] = parts[2].replace('/', '\\');
	    		index = parts[5].indexOf('/');
	    		parts[5] = parts[5].substring(index+1);
	    		parts[5] = parts[5].replace('/', '\\');
	    		int sysid = Integer.parseInt(parts[1])-1;
	    		outs[sysid].append(parts[2]);
	    		for(int i = 3; i < parts.length; i++)
	    			outs[sysid].append("\t" + parts[i]);
	    		outs[sysid].append("\r\n");
	    	}
	    	for(int i = 0; i < outs.length; i++)
	    	{
	    		outs[i].flush();
	    		outs[i].close();
	    	}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	public static ReportPairs loadBellonReferences(String logName)
	{
		ReportPairs report = new ReportPairs();
		report.setReportName("Bellon");
		report.setLogFile(logName);
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null)
	    	{
	    		String[] parts = line.split("\t");
	    		CloneFragment cf1 = new CloneFragment(), cf2 = new CloneFragment();
	    		cf1.fileName = parts[0];
	    		cf1.startLine = Integer.parseInt(parts[1]);
	    		cf1.endLine = Integer.parseInt(parts[2]);
	    		cf2.fileName = parts[3];
	    		cf2.startLine = Integer.parseInt(parts[4]);
	    		cf2.endLine = Integer.parseInt(parts[5]);
	    		ClonePair cp = new ClonePair();
	    		cp.cf1 = cf1; cp.cf2 = cf2; cp.type = Integer.parseInt(parts[6]);
	    		report.getClonePairs().add(cp);
	    	}
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return report;
	}
	public static int parseHex(char ch)
	{
		if(ch >= '0' && ch <= '9')
			return ch - '0';
		ch = Character.toLowerCase(ch);
		return 10 + ch - 'a';
	}
	public static int parseHex(String str)
	{
		int v = 0;
		for(int i = 0; i < str.length(); i++)
			v = v * 16 + parseHex(str.charAt(i));
		return v;
	}
	private static HashMap<String, ArrayList<Integer>> readTokensOfFiles(String root, File file)
	{
		HashMap<String, ArrayList<Integer>> tokensOfFiles = new HashMap<String, ArrayList<Integer>>();
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				tokensOfFiles.putAll(readTokensOfFiles(root, sub));
		}
		else if(file.getName().endsWith(".java.2_0_0_0.default.ccfxprep"))
		{
			ArrayList<Integer> tokensOfFile = new ArrayList<Integer>();
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	String line = "";
		    	while ((line = in.readLine()) != null) {
		    		int end = line.indexOf('.');
		    		tokensOfFile.add(parseHex(line.substring(0, end)));
		    	}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
			String fileName = file.getAbsolutePath().substring(root.length()+1, (int)(file.getAbsolutePath().length()-".java.2_0_0_0.default.ccfxprep".length()));
			tokensOfFiles.put(fileName, tokensOfFile);
		}
		
		return tokensOfFiles;
	}
	private static HashMap<String, ArrayList<int[]>> readTokenRangesOfFiles(String root, File file)
	{
		HashMap<String, ArrayList<int[]>> tokensOfFiles = new HashMap<String, ArrayList<int[]>>();
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				tokensOfFiles.putAll(readTokenRangesOfFiles(root, sub));
		}
		else if(file.getName().endsWith(".java.2_0_0_0.default.ccfxprep"))
		{
			ArrayList<int[]> tokensOfFile = new ArrayList<int[]>();
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	String line = "";
		    	while ((line = in.readLine()) != null) {
		    		String[] parts = line.split("\t");
		    		int[] range = new int[2];
		    		String strLen = !Character.isDigit(parts[1].charAt(0)) ? parts[1].substring(1) : parts[1];
		    		range[1] = parseHex(strLen);
		    		parts = parts[0].split("\\.");
		    		range[0] = parseHex(parts[2]);
		    		range[1] += range[0] - 1;
		    		tokensOfFile.add(range);
		    	}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
			String fileName = file.getAbsolutePath().substring(root.length()+1, (int)(file.getAbsolutePath().length()-".java.2_0_0_0.default.ccfxprep".length()));
			tokensOfFiles.put(fileName, tokensOfFile);
		}
		
		return tokensOfFiles;
	}
	public static ReportPairs loadCCFinderReport(String logName)
	{
		int[] counts = new int[5];
		ReportPairs report = new ReportPairs();
		report.setReportName("CCFinderX");
		report.setLogFile(logName);
		String srcDir = "";
		HashMap<String, RptSourceFile> importRanges = ReportLines.getImportRanges(new File(logName).getParent(), "src");
		HashMap<String, ArrayList<Integer>> tokensOfFiles = new HashMap<String, ArrayList<Integer>>();
		ArrayList<String> fileNames = new ArrayList<String>();
		HashMap<String, ArrayList<Boolean>> linesOfFiles = new HashMap<String, ArrayList<Boolean>>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("option: -n "))
		    	{
		    		srcDir = line.substring("option: -n ".length()).trim();
		    		tokensOfFiles = readTokensOfFiles(srcDir + "\\.ccfxprepdir", new File(srcDir + "\\.ccfxprepdir"));
		    		break;
		    	}
	    	}
	    	
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("source_files {"))
		    		break;
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.trim().equals("}"))
		    		break;
		    	String[] parts = line.split("\t");
		    	fileNames.add(parts[1].trim());
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("clone_pairs {"))
		    		break;
	    	}
	    	HashSet<String> examinedPairs = new HashSet<String>();
	    	while ((line = in.readLine()) != null) { 
	    		if(line.trim().equals("}"))
		    		break;
	    		String[] parts = line.split("\t");
		    	if(examinedPairs.contains(parts[1].trim() + " " + parts[2].trim()))
		    		continue;
		    	examinedPairs.add(parts[2].trim() + " " + parts[1].trim());
		    	CloneFragment[] cfs = new CloneFragment[2];
		    	for(int i = 1; i <= 2; i++)
		    	{
			    	String cloneInfo = parts[i].trim();
		    		int start = 0, end = cloneInfo.indexOf('.');
		    		int fid = Integer.parseInt(cloneInfo.substring(start, end));
		    		String fileName = fileNames.get(fid-1).substring(srcDir.length()+1);
		    		start = end+1;
	    			end = cloneInfo.indexOf('-', start);
	    			int sl = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start, end)));
	    			start = end + 1;
	    			int el = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start)));
	    			cfs[i-1] = new CloneFragment();
	    			cfs[i-1].fileName = fileName;
	    			cfs[i-1].startLine = sl;
	    			cfs[i-1].endLine = el;
		    	}
		    	if(!cfs[0].isOverlaped(cfs[1]))
		    	{
			    	counts[INDEX_TOTAL]++;
		    		RptSourceFile rsf1 = importRanges.get(cfs[0].fileName);
	    			int fragmentType1 = ReportPairs.INDEX_GOOD;
	    			if (cfs[0].startLine <= rsf1.endImportLine)
	    				fragmentType1 = ReportPairs.INDEX_IMPORT;
	    			else
	    			{
	    				if(Util.isCrossMethod(rsf1.cu, cfs[0].startLine, cfs[0].endLine))
	    					fragmentType1 = ReportPairs.INDEX_METHOD;
	    			}
	    			RptSourceFile rsf2 = importRanges.get(cfs[1].fileName);
	    			int fragmentType2 = INDEX_GOOD;
	    			if (cfs[1].startLine <= rsf2.endImportLine)
	    				fragmentType2 = INDEX_IMPORT;
	    			else
	    			{
	    				if(Util.isCrossMethod(rsf2.cu, cfs[1].startLine, cfs[1].endLine))
	    					fragmentType2 = INDEX_METHOD;
	    			}
	    			if(fragmentType1 == INDEX_IMPORT || fragmentType2 == INDEX_IMPORT)
	    				counts[INDEX_IMPORT]++;
	    			else if(fragmentType1 == INDEX_METHOD || fragmentType2 == INDEX_METHOD)
	    				counts[INDEX_METHOD]++;
	    			else
					{
    					counts[INDEX_GOOD]++;
	    				ClonePair cp = new ClonePair(cfs[0], cfs[1]);
		    			report.getClonePairs().add(cp);
		    			ArrayList<Boolean> lines = linesOfFiles.get(cfs[0].fileName);
    	    			if(lines == null)
    	    				lines = new ArrayList<Boolean>();
    	    			int cl = lines.size();
        				for(int ii = cl-1; ii < cfs[0].endLine; ii++)
        					lines.add(false);
    	    			for(int ii = 0; ii < cfs[0].endLine-cfs[0].startLine+1; ii++)
    	    			{
    	    				lines.set(cfs[0].startLine + ii - 1, true);
    	    			}
    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
    	    			linesOfFiles.put(cfs[0].fileName, lines);
    	    			
    	    			lines = linesOfFiles.get(cfs[1].fileName);
    	    			if(lines == null)
    	    				lines = new ArrayList<Boolean>();
    	    			cl = lines.size();
        				for(int ii = cl-1; ii < cfs[1].endLine; ii++)
        					lines.add(false);
    	    			for(int ii = 0; ii < cfs[1].endLine-cfs[1].startLine+1; ii++)
    	    			{
    	    				lines.set(cfs[1].startLine + ii - 1, true);
    	    			}
    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
    	    			linesOfFiles.put(cfs[1].fileName, lines);
					}
		    	}
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	System.out.println(report.getReportName());
    	System.out.println("Total: " + counts[INDEX_TOTAL] + "\tImport: " 
    			+ counts[INDEX_IMPORT] + "\tMethod: " + counts[INDEX_METHOD] + "\tGood: " + counts[INDEX_GOOD]);
    	for(String fileName : linesOfFiles.keySet())
    	{
    		ArrayList<Boolean> lines = linesOfFiles.get(fileName);
			ArrayList<int[]> allClonedLinesOfFile = new ArrayList<int[]>();
    		int j = 0;
    		
    		while(j < lines.size())
    		{
    			while(j < lines.size() && !lines.get(j))
    				j++;
    			if(j < lines.size())
    			{
    				int start = j++;
	    			while(j < lines.size() && lines.get(j))
	    				j++;
	    			int[] pair = new int[2];
	    			pair[0] = start;
	    			pair[1] = j-1;
	    			allClonedLinesOfFile.add(pair);
	    			report.numOfCLOC += j - start;
    			}
    		}
    		report.allClonedLinesOfFiles.put(fileName, allClonedLinesOfFile);
    	}
    	System.out.println("Clone LOC: " + report.numOfCLOC);
    	
		return report;
	}
	public static ReportPairs loadDeckardReport(String logName)
	{
		int[] counts = new int[5];
		for(int i = 0; i < counts.length; i++)
			counts[i] = 0;
		HashMap<String, RptSourceFile> importRanges = ReportLines.getImportRanges(new File(logName).getParent(), "src");
		String DK_LINE = "LINE:", DK_FILE = "FILE ";
		ReportPairs report = new ReportPairs();
		report.setReportName("Deckard");
		report.setLogFile(logName);
		HashMap<String, ArrayList<Boolean>> linesOfFiles = new HashMap<String, ArrayList<Boolean>>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) {
	    		while((line = in.readLine()) != null && line.trim().isEmpty())
		    		line = in.readLine();
	    		if(line == null)
	    			break;
	    		ArrayList<CloneFragment> group = new ArrayList<CloneFragment>();
	    		do
	    		{
	    			int start = line.indexOf(DK_FILE);
	    			if(start == -1)
	    				continue;
	    			start += DK_FILE.length();
	    			int end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			String fileName = line.substring(start, end);
	    			fileName = Util.replaceAll(fileName, '/', '\\');
	    			start = line.indexOf(DK_LINE, end+1);
	    			if(start == -1)
	    				continue;
	    			start += DK_LINE.length();
	    			end = line.indexOf(':', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int sl = Integer.parseInt(line.substring(start, end));
	    			/*if(sl <= importRanges.get(fileName).endImportLine)
	    				continue;*/
	    			start = end + 1;
	    			end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int el = sl + Integer.parseInt(line.substring(start, end)) - 1;
	    			CloneFragment cf = new CloneFragment();
	    			cf.fileName = fileName;
	    			cf.startLine = sl; cf.endLine = el;
	    			group.add(cf);
	    		} while ((line = in.readLine()) != null && !line.trim().isEmpty());
	    		for(int i = 0; i < group.size()-1; i++)
	    		{
	    			CloneFragment cf1 = group.get(i);
	    			RptSourceFile rsf1 = importRanges.get(cf1.fileName);
	    			int fragmentType1 = INDEX_GOOD;
	    			if (cf1.startLine <= rsf1.endImportLine)
	    				fragmentType1 = INDEX_IMPORT;
	    			else
	    			{
	    				if(Util.isCrossMethod(rsf1.cu, cf1.startLine, cf1.endLine))
	    					fragmentType1 = INDEX_METHOD;
	    			}
	    			for(int j = i+1; j < group.size(); j++)
	    			{
	    				CloneFragment cf2 = group.get(j);
	    				if(!cf1.isOverlaped(cf2))
	    				{
	    					counts[INDEX_TOTAL]++;
	    					RptSourceFile rsf2 = importRanges.get(cf2.fileName);
	    	    			int fragmentType2 = INDEX_GOOD;
	    	    			if (cf2.startLine <= rsf2.endImportLine)
	    	    				fragmentType2 = INDEX_IMPORT;
	    	    			else
	    	    			{
	    	    				if(Util.isCrossMethod(rsf2.cu, cf2.startLine, cf2.endLine))
	    	    					fragmentType2 = INDEX_METHOD;
	    	    			}
	    	    			if(fragmentType1 == INDEX_IMPORT || fragmentType2 == INDEX_IMPORT)
	    	    				counts[INDEX_IMPORT]++;
	    	    			else if(fragmentType1 == INDEX_METHOD || fragmentType2 == INDEX_METHOD)
	    	    				counts[INDEX_METHOD]++;
	    	    			else
	    					{
		    					ClonePair cp = new ClonePair(cf1, cf2);
		    					report.getClonePairs().add(cp);
		    					counts[INDEX_GOOD]++;
		    					ArrayList<Boolean> lines = linesOfFiles.get(cf1.fileName);
		    	    			if(lines == null)
		    	    				lines = new ArrayList<Boolean>();
		    	    			int cl = lines.size();
		        				for(int ii = cl-1; ii < cf1.endLine; ii++)
		        					lines.add(false);
		    	    			for(int ii = 0; ii < cf1.endLine-cf1.startLine+1; ii++)
		    	    			{
		    	    				lines.set(cf1.startLine + ii - 1, true);
		    	    			}
		    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
		    	    			linesOfFiles.put(cf1.fileName, lines);
		    	    			
		    	    			lines = linesOfFiles.get(cf2.fileName);
		    	    			if(lines == null)
		    	    				lines = new ArrayList<Boolean>();
		    	    			cl = lines.size();
		        				for(int ii = cl-1; ii < cf2.endLine; ii++)
		        					lines.add(false);
		    	    			for(int ii = 0; ii < cf2.endLine-cf2.startLine+1; ii++)
		    	    			{
		    	    				lines.set(cf2.startLine + ii - 1, true);
		    	    			}
		    	    			//System.out.println("Got from " + fileName + " " + sl + "-->" + (sl+len-1));
		    	    			linesOfFiles.put(cf2.fileName, lines);
	    					}
	    				}
	    			}
	    		}
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	System.out.println(report.getReportName());
    	System.out.println("Total: " + counts[INDEX_TOTAL] + "\tImport: " 
    			+ counts[INDEX_IMPORT] + "\tMethod: " + counts[INDEX_METHOD] + "\tGood: " + counts[INDEX_GOOD]);
    	for(String fileName : linesOfFiles.keySet())
    	{
    		ArrayList<Boolean> lines = linesOfFiles.get(fileName);
			ArrayList<int[]> allClonedLinesOfFile = new ArrayList<int[]>();
    		int j = 0;
    		
    		while(j < lines.size())
    		{
    			while(j < lines.size() && !lines.get(j))
    				j++;
    			if(j < lines.size())
    			{
    				int start = j++;
	    			while(j < lines.size() && lines.get(j))
	    				j++;
	    			int[] pair = new int[2];
	    			pair[0] = start;
	    			pair[1] = j-1;
	    			allClonedLinesOfFile.add(pair);
	    			report.numOfCLOC += j - start;
    			}
    		}
    		report.allClonedLinesOfFiles.put(fileName, allClonedLinesOfFile);
    	}
    	System.out.println("Clone LOC: " + report.numOfCLOC);
    	
		return report;
	}
	
	public void compareTo(ReportPairs other, HashSet<String> goods, HashSet<String> oks)
	{
		HashSet<ClonePair> goodMatches1 = new HashSet<ClonePair>(), goodMatches2 = new HashSet<ClonePair>();
		HashSet<ClonePair> okMatches1 = new HashSet<ClonePair>(), okMatches2 = new HashSet<ClonePair>();
		HashMap<ClonePair, HashSet<ClonePair>> goodMatches = new HashMap<ClonePair, HashSet<ClonePair>>();
		for(ClonePair cp1 : this.clonePairs)
		{
			HashSet<ClonePair> gms = new HashSet<ClonePair>();
			for(ClonePair cp2 : other.getClonePairs())
			{
				double[] matches = cp1.match(cp2);
				if(matches[0] >= THRES_MATCH)
				{
					goodMatches1.add(cp1);
					goodMatches2.add(cp2);
					gms.add(cp2);
					goods.add(cp1.cf1.fileName + "|" + cp1.cf1.startLine + "|" + cp1.cf1.endLine
							+ "<->"
							+cp1.cf2.fileName + "|" + cp1.cf2.startLine + "|" + cp1.cf2.endLine);
				}
				if(matches[1] >= THRES_MATCH)
				{
					okMatches1.add(cp1);
					okMatches2.add(cp2);
					oks.add(cp1.cf1.fileName + "|" + cp1.cf1.startLine + "|" + cp1.cf1.endLine
							+ "<->"
							+cp1.cf2.fileName + "|" + cp1.cf2.startLine + "|" + cp1.cf2.endLine);
				}
			}
			if(!gms.isEmpty())
				goodMatches.put(cp1, gms);
		}
		System.out.println(this.reportName + ": " + this.clonePairs.size() + "\t" + goodMatches1.size() + " good " + okMatches1.size() + " ok");
		System.out.println(other.getReportName() + ": " + other.getClonePairs().size() + "\t" + goodMatches2.size() + " good " + okMatches2.size() + " ok");
		
		if(!goodMatches.isEmpty())
		{
			int count = 0;
			File dir = new File("good_matches_" + other.getReportName());
			if(!dir.exists())
				dir.mkdirs();
			for(ClonePair ref : goodMatches.keySet())
			{
				HashSet<ClonePair> gms = goodMatches.get(ref);
				try {
					BufferedWriter out = new BufferedWriter(new FileWriter(dir.getAbsolutePath() + "\\" + (++count) + ".txt"));
					out.append(ref.cf1.fileName + "\t" + ref.cf1.startLine + "\t" + ref.cf1.endLine);
					out.append("\t");
					out.append(ref.cf2.fileName + "\t" + ref.cf2.startLine + "\t" + ref.cf2.endLine);
					out.append("\r\n");
					out.append("\r\n");
					for(ClonePair cp : gms)
					{
						out.append(cp.cf1.fileName + "\t" + cp.cf1.startLine + "\t" + cp.cf1.endLine);
						out.append("\t");
						out.append(cp.cf2.fileName + "\t" + cp.cf2.startLine + "\t" + cp.cf2.endLine);
						out.append("\r\n");
					}
					out.flush();
					out.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
				
			}
		}
	}
	public void getCCFCloneCode(String logName)
	{
		File logFile = new File(logName);
		ReportPairs report = new ReportPairs();
		report.setReportName("CCFinderX");
		report.setLogFile(logName);
		String srcDir = "";
		HashMap<String, ArrayList<int[]>> tokensOfFiles = new HashMap<String, ArrayList<int[]>>();
		ArrayList<String> fileNames = new ArrayList<String>();
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("option: -n "))
		    	{
		    		srcDir = line.substring("option: -n ".length()).trim();
		    		tokensOfFiles = readTokenRangesOfFiles(srcDir + "\\.ccfxprepdir", new File(srcDir + "\\.ccfxprepdir"));
		    		break;
		    	}
	    	}
	    	
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("source_files {"))
		    		break;
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.trim().equals("}"))
		    		break;
		    	String[] parts = line.split("\t");
		    	fileNames.add(parts[1].trim());
	    	}
	    	while ((line = in.readLine()) != null) { 
		    	if(line.startsWith("clone_pairs {"))
		    		break;
	    	}
	    	HashSet<String> examinedPairs = new HashSet<String>();
	    	while ((line = in.readLine()) != null) { 
	    		if(line.trim().equals("}"))
		    		break;
	    		String[] parts = line.split("\t");
		    	if(examinedPairs.contains(parts[1].trim() + " " + parts[2].trim()))
		    		continue;
		    	examinedPairs.add(parts[2].trim() + " " + parts[1].trim());
		    	CloneFragment[] cfs = new CloneFragment[2];
		    	for(int i = 1; i <= 2; i++)
		    	{
			    	String cloneInfo = parts[i].trim();
		    		int start = 0, end = cloneInfo.indexOf('.');
		    		int fid = Integer.parseInt(cloneInfo.substring(start, end));
		    		String fileName = fileNames.get(fid-1).substring(srcDir.length()+1);
		    		start = end+1;
	    			end = cloneInfo.indexOf('-', start);
	    			int sl = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start, end)))[0];
	    			start = end + 1;
	    			int el = tokensOfFiles.get(fileName).get(Integer.parseInt(cloneInfo.substring(start)))[1];
	    			cfs[i-1] = new CloneFragment();
	    			cfs[i-1].fileName = fileName;
	    			cfs[i-1].startLine = sl;
	    			cfs[i-1].endLine = el;
		    	}
		    	ClonePair cp = new ClonePair(cfs[0], cfs[1]);
    			report.getClonePairs().add(cp);
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	ArrayList<ClonePair> clonePairs = new ArrayList<ClonePair>(report.getClonePairs());
    	HashSet<Integer> samples = new HashSet<Integer>();
		Random rand = new Random();
		for(int i = 0; i < Main.numOfSamples; i++)
		{
			int sample;
			do
			{
				sample = rand.nextInt(clonePairs.size());
			} while(samples.contains(sample));
			samples.add(sample);
		}
		File reportDir = new File(logFile.getParentFile().getAbsolutePath() + "\\clonesCCF2Check");
		if(!reportDir.exists())
			reportDir.mkdirs();
		for(int sample : samples)
		{
			try {
				File clonePairDir = new File(reportDir.getAbsolutePath() + "\\" + sample);
				if(!clonePairDir.exists())
					clonePairDir.mkdirs();
				ClonePair clonePair = clonePairs.get(sample);
				BufferedWriter out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\1.txt"));
				out.append(clonePair.cf1.fileName + "\t" + clonePair.cf1.startLine + "\t" + clonePair.cf1.endLine + "\r\n" + "\r\n");
				int len = clonePair.cf1.endLine - clonePair.cf1.startLine + 1;
				char[] buf = new char[len];
				BufferedReader in = new BufferedReader(new FileReader(logFile.getParentFile().getAbsolutePath()+ "\\" + clonePair.cf1.fileName));
				in.skip(clonePair.cf1.startLine);
				int numOfRead = in.read(buf, 0, len);
				if(numOfRead != len)
					System.err.println("Reading out of bound in cf1: " + numOfRead + " - " + len);
				out.append(String.valueOf(buf));
				in.close();
				out.flush();
				out.close();
				out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\2.txt"));
				out.append(clonePair.cf2.fileName + "\t" + clonePair.cf2.startLine + "\t" + clonePair.cf2.endLine + "\r\n" + "\r\n");
				in = new BufferedReader(new FileReader(logFile.getParentFile().getAbsolutePath()+ "\\" + clonePair.cf2.fileName));
				len = clonePair.cf2.endLine - clonePair.cf2.startLine + 1;
				buf = new char[len];
				in.skip(clonePair.cf2.startLine);
				numOfRead = in.read(buf, 0, len);
				if(numOfRead != len)
					System.err.println("Reading out of bound in cf2: " + numOfRead + " - " + len);
				out.append(String.valueOf(buf));
				in.close();
				out.flush();
				out.close();
				out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\diff.bat"));
				out.append("\"F:\\Program Files\\TortoiseSVN\\bin\\TortoiseMerge.exe\" 1.txt 2.txt");
				out.flush();
				out.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	public int[] getDeckardCloneCode(String logName)
	{
		int[] counts = new int[5];
		for(int i = 0; i < counts.length; i++)
			counts[i] = 0;
		File logFile = new File(logName);
		HashMap<String, RptSourceFile> importRanges = ReportLines.getImportRanges(logFile.getParent(), "src");
		String DK_LINE = "LINE:", DK_FILE = "FILE ";
		ReportPairs report = new ReportPairs();
		report.setReportName("Deckard");
		report.setLogFile(logName);
		try {
			BufferedReader in = new BufferedReader(new FileReader(logName));
	    	String line = "";
	    	while ((line = in.readLine()) != null) {
	    		while((line = in.readLine()) != null && line.trim().isEmpty())
		    		line = in.readLine();
	    		if(line == null)
	    			break;
	    		ArrayList<CloneFragment> group = new ArrayList<CloneFragment>();
	    		do
	    		{
	    			int start = line.indexOf(DK_FILE);
	    			if(start == -1)
	    				continue;
	    			start += DK_FILE.length();
	    			int end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			String fileName = line.substring(start, end);
	    			fileName = Util.replaceAll(fileName, '/', '\\');
	    			start = line.indexOf(DK_LINE, end+1);
	    			if(start == -1)
	    				continue;
	    			start += DK_LINE.length();
	    			end = line.indexOf(':', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int sl = Integer.parseInt(line.substring(start, end));
	    			/*if(sl <= importRanges.get(fileName))
	    				continue;*/
	    			start = end + 1;
	    			end = line.indexOf(' ', start);
	    			if(end == -1 || end >= line.length()-1)
	    				continue;
	    			int el = sl + Integer.parseInt(line.substring(start, end)) - 1;
	    			CloneFragment cf = new CloneFragment();
	    			cf.fileName = fileName;
	    			cf.startLine = sl; cf.endLine = el;
	    			group.add(cf);
	    		} while ((line = in.readLine()) != null && !line.trim().isEmpty());
	    		for(int i = 0; i < group.size()-1; i++)
	    		{
	    			CloneFragment cf1 = group.get(i);
	    			RptSourceFile rsf1 = importRanges.get(cf1.fileName);
	    			int fragmentType1 = INDEX_GOOD;
	    			if (cf1.startLine <= rsf1.endImportLine)
	    				fragmentType1 = INDEX_IMPORT;
	    			else
	    			{
	    				if(Util.isCrossMethod(rsf1.cu, cf1.startLine, cf1.endLine))
	    					fragmentType1 = INDEX_METHOD;
	    			}
	    			for(int j = i+1; j < group.size(); j++)
	    			{
	    				CloneFragment cf2 = group.get(j);
	    				if(!cf1.isOverlaped(cf2))
	    				{
	    					counts[INDEX_TOTAL]++;
	    					RptSourceFile rsf2 = importRanges.get(cf2.fileName);
	    	    			int fragmentType2 = INDEX_GOOD;
	    	    			if (cf2.startLine <= rsf2.endImportLine)
	    	    				fragmentType2 = INDEX_IMPORT;
	    	    			else
	    	    			{
	    	    				if(Util.isCrossMethod(rsf2.cu, cf2.startLine, cf2.endLine))
	    	    					fragmentType2 = INDEX_METHOD;
	    	    			}
	    	    			if(fragmentType1 == INDEX_IMPORT || fragmentType2 == INDEX_IMPORT)
	    	    				counts[INDEX_IMPORT]++;
	    	    			else if(fragmentType1 == INDEX_METHOD || fragmentType2 == INDEX_METHOD)
	    	    				counts[INDEX_METHOD]++;
	    	    			else
	    					{
		    					ClonePair cp = new ClonePair(cf1, cf2);
		    					report.getClonePairs().add(cp);
		    					counts[INDEX_GOOD]++;
	    					}
	    				}
	    			}
	    		}
	    	}
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	ArrayList<ClonePair> clonePairs = new ArrayList<ClonePair>(report.getClonePairs());
    	HashSet<Integer> samples = new HashSet<Integer>();
		Random rand = new Random();
		for(int i = 0; i < Main.numOfSamples; i++)
		{
			int sample;
			do
			{
				sample = rand.nextInt(clonePairs.size());
			} while(samples.contains(sample));
			samples.add(sample);
		}
		File reportDir = new File(logFile.getParentFile().getAbsolutePath() + "\\clonesDeckard2Check");
		if(!reportDir.exists())
			reportDir.mkdirs();
		for(int sample : samples)
		{
			try {
				File clonePairDir = new File(reportDir.getAbsolutePath() + "\\" + sample);
				if(!clonePairDir.exists())
					clonePairDir.mkdirs();
				ClonePair clonePair = clonePairs.get(sample);
				BufferedWriter out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\1.txt"));
				out.append(clonePair.cf1.fileName + "\t" + clonePair.cf1.startLine + "\t" + clonePair.cf1.endLine + "\r\n" + "\r\n");
				BufferedReader in = new BufferedReader(new FileReader(logFile.getParentFile().getAbsolutePath()+ "\\" + clonePair.cf1.fileName));
				for(int i = 1; i < clonePair.cf1.startLine; i++)
					in.readLine();
				for(int i = clonePair.cf1.startLine; i <= clonePair.cf1.endLine; i++)
				{
					out.append(in.readLine() + "\r\n");
				}
				in.close();
				out.flush();
				out.close();
				out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\2.txt"));
				out.append(clonePair.cf2.fileName + "\t" + clonePair.cf2.startLine + "\t" + clonePair.cf2.endLine + "\r\n" + "\r\n");
				in = new BufferedReader(new FileReader(logFile.getParentFile().getAbsolutePath()+ "\\" + clonePair.cf2.fileName));
				for(int i = 1; i < clonePair.cf2.startLine; i++)
					in.readLine();
				for(int i = clonePair.cf2.startLine; i <= clonePair.cf2.endLine; i++)
				{
					out.append(in.readLine() + "\r\n");
				}
				in.close();
				out.flush();
				out.close();
				out = new BufferedWriter(new FileWriter(clonePairDir.getAbsolutePath() + "\\diff.bat"));
				out.append("\"F:\\Program Files\\TortoiseSVN\\bin\\TortoiseMerge.exe\" 1.txt 2.txt");
				out.flush();
				out.close();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		return counts;
	}
}
