package cleman;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Stack;

import org.eclipse.jdt.core.dom.*;

import treeMapping.MapVisitor;
import treeMapping.TreeMapper;


/**
 * @author Nguyen Anh Hoan
 *
 */
public class SequenceIDVisitor extends ASTVisitor {
	public static final String propertySequenceID = "sid";
	private int sid = 0;
	private int lastNode = 0;	//for end token of vector
	private Stack<Integer> stackCurrentNode = new Stack<Integer>();	//for start token of vector
	private ArrayList<Integer> fragments = new ArrayList<Integer>();
	
    public SequenceIDVisitor(ArrayList<Integer> fragments)
    {
    	this.fragments = fragments;
    }
    public void preVisit(ASTNode node) {
    	int nodeType = node.getNodeType(); 
    	if (nodeType == ASTNode.JAVADOC || nodeType == ASTNode.BLOCK_COMMENT || nodeType == ASTNode.LINE_COMMENT)
    		return;
    	stackCurrentNode.push(++lastNode); 
    }
    public void postVisit(ASTNode node) {
    	int nodeType = node.getNodeType();
    	if (nodeType == ASTNode.JAVADOC || nodeType == ASTNode.BLOCK_COMMENT || nodeType == ASTNode.LINE_COMMENT)
    		return;
    	int currentNode = stackCurrentNode.pop();
    	if(nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.BLOCK && nodeType != ASTNode.PARENTHESIZED_EXPRESSION && 
    			lastNode-currentNode+1 >= Main.minFragmentSize)
		{
    		Fragment fragment = Fragment.all.get(fragments.get(sid));
    		//fragment.setAstNode(node);
    		//Fragment.prev.put(fragment.getId(), fragment);
			if(node.getProperty(TreeMapper.propertyRevisionMap) == null)
    		{
    			Fragment.deleteds.put(fragment.getId(), fragment);
    			Fragment.all.remove(fragment.getId());
    		}
    		node.setProperty(propertySequenceID, sid++);
		}
    	
    }
 }
