package cleman;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.tmatesoft.svn.core.SVNDirEntry;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.io.SVNRepository;

/**
 * This class will be used to hold the files information for the application.
 * @author Jafar Al-Kofahi
 *
 */
public class SourceFile implements Serializable, Cloneable {
	private static final long serialVersionUID = 1L;
	
	public static int nextID = 1;
	//public static HashSet<SourceFile> all = new HashSet<SourceFile>();
	
	public static HashMap<String, SourceFile> all = new HashMap<String, SourceFile>(), prev = new HashMap<String, SourceFile>();
	//public static HashSet<SourceFile> addeds = new HashSet<SourceFile>();
	/**
	 * This will contain the file ID, which will be the hashCode of the file path
	 */
	public int fileID;
	/**
	 * This will contain the full path for this file
	 */
	public String fileLocation; 
	/**
	 * The last modified time for this file 
	 */
	public long lastModified;
	/**
	 * This file size in KB
	 */
	public long fileSize;
	/**
	 * This will contain the total number of code lines in the file  
	 */
	//public int fileCodeLength;
	/**
	 * This will contain the total number of code lines in the file  
	 */
	public int fileCommentLength;
	/**
	 * This DS will contain all vectors that belong to this file.
	 */
	public ArrayList<Integer> fragments = new ArrayList<Integer>();
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.fileID;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		SourceFile other = (SourceFile) obj;
		if (fileID != other.fileID)
			return false;
		return true;
	}
	
	public SourceFile(SourceFile other)
	{
		this.fileID = other.fileID;
		this.fileLocation = other.fileLocation;
		this.lastModified = other.lastModified;
		this.fileSize = other.fileSize;
		this.fileCommentLength = other.fileCommentLength;
		this.fragments = other.fragments;
	}
	
	public SourceFile(File file)
	{
		String fileName = file.getAbsolutePath();
		this.fileID = nextID++;
		String relativePath = fileName.substring(Main.applicationPath.length()+1);
		this.fileLocation = relativePath;
		StringBuffer strBuf = new StringBuffer();
		ArrayList<Integer> lines = new ArrayList<Integer>();
		lines.add(0);
    	try {
			BufferedReader in = new BufferedReader(new FileReader(fileName));
	    	String line = "";
	    	int charCount = 0, lineCount = 0;
	    	while ((line = in.readLine()) != null) { 
	    		lineCount++;
				lines.add(charCount);
				charCount += line.length() + 2;
				//line.replaceAll("/" + '*' + '*', "/" + '*' + " ");
	    		
	    		/*while (line.indexOf("/**") > -1) {
	    			int i = line.indexOf("/**");
	    			line = line.substring(0, i) + "/* " + line.substring(i + 3);
	    		}
	    		strBuf.append(line + "\r\n");*/
	    		char[] chs = line.toCharArray();
	    		for (int i = 0; i < chs.length - 2; i++)
	    			if (chs[i] == '/' && chs[i+1] == '*') chs[i+2] = ' '; 
	    		strBuf.append(chs);
	    		strBuf.append("\r\n");
	    	}
	    	lines.set(0, lineCount);
	    	//this.fileCodeLength = lineCount;
			//CheckReport.files.put(fileName, lines);
	    	in.close();
		}
    	catch (IOException e) {
    		e.printStackTrace();
    	}
    	ASTParser parser = ASTParser.newParser(AST.JLS3);
    	parser.setSource(strBuf.toString().toCharArray());
    	ASTNode ast = parser.createAST(null); 
    	VectorVisitor visitor = new VectorVisitor(this);
    	visitor.lines = new ArrayList<Integer>(lines);
    	ast.accept(visitor);
    	this.fragments = new ArrayList<Integer>(visitor.fragments);
    	/*for(Fragment f : visitor.fragments)
    		Fragment.all.put(f.getId(), f);*/
    	visitor = null;
	}
	public SourceFile(String file, ArrayList<Integer> prevFragments, String  pfileLocation)
	{
		
		BufferedReader in = new BufferedReader(new StringReader(file));
		ArrayList<Integer> p_lines = new ArrayList<Integer>();
		String line = "";
    	int charCount = 0, lineCount = 0;
    	StringBuffer strBuf = new StringBuffer();
    	try {
			while ((line = in.readLine()) != null) { 
				lineCount++;
				p_lines.add(charCount);
				charCount += line.length() + 2;
				//line.replaceAll("/" + '*' + '*', "/" + '*' + " ");
				
				/*while (line.indexOf("/**") > -1) {
					int i = line.indexOf("/**");
					line = line.substring(0, i) + "/* " + line.substring(i + 3);
				}
				strBuf.append(line + "\r\n");*/
				char[] chs = line.toCharArray();
				for (int i = 0; i < chs.length - 2; i++)
					if (chs[i] == '/' && chs[i+1] == '*') chs[i+2] = ' '; 
				strBuf.append(chs);
				strBuf.append("\r\n");
				
			}
			in.close();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
    	
    	p_lines.set(0, lineCount);
    	
		
		
		
		this.fileID = nextID++;
		//CheckReport.files.put(fileName, lines);
    	try{
		ASTParser parser = ASTParser.newParser(AST.JLS3);
    	  
    	this.fileLocation=pfileLocation;
    	parser.setSource(file.toCharArray());
    	ASTNode ast = parser.createAST(null); 
    	
    	VectorVisitor visitor = new VectorVisitor(this);
    	visitor.lines=p_lines;
    	visitor.prevFragments = prevFragments;
    	ast.accept(visitor);
    	this.fragments = new ArrayList<Integer>(visitor.fragments);
    	
    	visitor = null;
    	} catch (Exception e)
    	{
    		System.out.println(e);
    	}
	}
	public SourceFile(ASTNode ast, ArrayList<Integer> prevFragments, String filecontent)
	{
		
		BufferedReader in = new BufferedReader(new StringReader(filecontent));
		ArrayList<Integer> p_lines = new ArrayList<Integer>();
		String line = "";
    	int charCount = 0, lineCount = 0;
    	StringBuffer strBuf = new StringBuffer();
    	try {
			while ((line = in.readLine()) != null) { 
				lineCount++;
				p_lines.add(charCount);
				charCount += line.length() + 2;
				//line.replaceAll("/" + '*' + '*', "/" + '*' + " ");
				
				/*while (line.indexOf("/**") > -1) {
					int i = line.indexOf("/**");
					line = line.substring(0, i) + "/* " + line.substring(i + 3);
				}
				strBuf.append(line + "\r\n");*/
				char[] chs = line.toCharArray();
				for (int i = 0; i < chs.length - 2; i++)
					if (chs[i] == '/' && chs[i+1] == '*') chs[i+2] = ' '; 
				strBuf.append(chs);
				strBuf.append("\r\n");
				
			}
			in.close();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
    	
    	p_lines.set(0, lineCount);
    	
		
		
		this.fileID = nextID++;
		//CheckReport.files.put(fileName, lines);
    	VectorVisitor visitor = new VectorVisitor(this);
    	
    	visitor.prevFragments = prevFragments;
    	ast.accept(visitor);
    	this.fragments = new ArrayList<Integer>(visitor.fragments);
    	/*for(Fragment f : visitor.fragments)
    		Fragment.all.put(f.getId(), f);*/
    	visitor = null;
	}
	/**
	 * Generate fragments for all files in the project specified by the path
	 * @param dirPath
	 * @return
	 */
	public static void genAllFiles(File file) {
		try {
			//File file = new File(dirPath);
			if (file.isFile()) {
				if (file.getName().endsWith(".java")) {
					SourceFile sf = new SourceFile(file);
					sf.fileSize = file.length();
					all.put(sf.fileLocation, sf);
				}
			}
			else {
				File[] fileList = file.listFiles();
				if(fileList != null &&  fileList.length > 0)
					for (File f : fileList) {
						if (f.isDirectory() || f.getName().endsWith(".java")) {
							genAllFiles(f);
						}
					}
			}
		}
		catch (Exception ex) {
			ex.printStackTrace();
		}
	}
	public static void genAllFiles(SVNRepository repository, String path, long revision) {
		try {
			Collection entries = repository.getDir(path, revision, null, (Collection)null);
			Iterator iterator = entries.iterator();
			while (iterator.hasNext()) {
				SVNDirEntry entry = (SVNDirEntry)iterator.next();
				//System.err.println(entry.getRelativePath() + "\t" + entry.getURL());
				/*System.err.println( "/" + (path.equals( "" ) ? "" : path + "/" ) + entry.getName( ) + 
						" ( author: '" + entry.getAuthor( ) + "'; revision: " + entry.getRevision( ) + 
						"; date: " + entry.getDate( ) + ")" );*/
				if(entry.getKind() == SVNNodeKind.FILE)
				{
					if(entry.getName().endsWith(".java")
						&& !entry.getName().equals("Parser.java")
						&& !entry.getName().equals("Interpreter.java")
					)
					{
						ByteArrayOutputStream out = new ByteArrayOutputStream();
						repository.getFile((path.equals("") ? "" : path + "/") + entry.getName(), revision, null, out);
						//repository.getFile(path + "/" + entry.getName(), revision, null, out);
						String fileLocation = (path.equals("") ? "" : path + "/") + entry.getName();
						
						SourceFile sf = new SourceFile(out.toString(), null, fileLocation);
						//sf.fileLocation = (path.equals("") ? "" : path + "/") + entry.getName();
						sf.fileSize = out.size();
						all.put(sf.fileLocation, sf);
						//System.err.println(sf.fileLocation);
					}
				}
				if (entry.getKind() == SVNNodeKind.DIR) {
					genAllFiles(repository, (path.equals("")) ? entry.getName() : path + "/" + entry.getName(), revision);
				}
			}
		}
		catch (Exception e) {
			/*if(!e.getMessage().contains("path not found"))
				e.printStackTrace();
			else
				System.err.println("Path not found: " + e.getMessage());*/
			e.printStackTrace();
		}
	}
	/*void printFragments() {
		System.out.println("File " + this.fileLocation + "\t" + fragments.size() + "fragments");
    	for (int fragmentID  : fragments) {
    		Data.fragments.get(fragmentID).printFragment();
    	}
	}*/
}
