/**
 * 
 */
package cleman;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.tmatesoft.svn.core.SVNDirEntry;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.io.SVNRepository;

/**
 * @author Hoan Nguyen
 *
 */
public class Util {
	public static String replaceAll(String str, char char1, char char2)
	{
		StringBuffer buf = new StringBuffer();
		for(int i = 0; i < str.length(); i++)
			if(str.charAt(i) == char1)
				buf.append(char2);
			else
				buf.append(str.charAt(i));
		
		return buf.toString();
	}
	public static String getFileContent(File file)
	{
		StringBuffer strBuf = new StringBuffer();
		try {
			BufferedReader in = new BufferedReader(new FileReader(file));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
	    		strBuf.append(line + "\n");
	    	}
	    	in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return strBuf.toString();
	}
	public static ASTNode getClosestMethodNode(ASTNode node)
	{
		ASTNode temp = node;
		while (temp != null && temp.getNodeType() != ASTNode.METHOD_DECLARATION)
			temp = temp.getParent();
		
		return temp;
	}
	public static boolean isCrossMethod(CompilationUnit root, int start, int end)
	{
		ASTPathVisitor visitor = new ASTPathVisitor(root, start, end);
		root.accept(visitor);
		//if(visitor.startMethod != null && visitor.endMethod != null && visitor.startMethod != visitor.endMethod)
		if(visitor.startMethod != visitor.endMethod)
		{
			if (visitor.startMethod != null && visitor.startMethod.getBody() != null)
			{
				ASTLeafVisitor leafVisitor = new ASTLeafVisitor(root);
				visitor.startMethod.getBody().accept(leafVisitor);
				if(leafVisitor.start < start)
					return true;
			}
			if (visitor.endMethod != null && visitor.endMethod.getBody() != null)
			{
				ASTLeafVisitor leafVisitor = new ASTLeafVisitor(root);
				visitor.endMethod.getBody().accept(leafVisitor);
				if(leafVisitor.end > end)
					return true;
			}
		}
		return false;
	}
	
	public static int countLOC(File file, String extension)
	{
		int numOfLines = 0;
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				numOfLines += countLOC(sub, extension);
		}
		else if(file.getName().endsWith("." + extension))
		{
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	while (in.readLine() != null) {
		    		numOfLines++;
		    	}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		return numOfLines;
	}
	
	public static int[] countLOC(SVNRepository repository, String path, long revision) 
	{
		int[] count = {0, 0};
		try {
			Collection entries = repository.getDir(path, revision, null, (Collection)null);
			Iterator iterator = entries.iterator();
			while (iterator.hasNext()) {
				SVNDirEntry entry = (SVNDirEntry)iterator.next();
				if (entry.getKind() == SVNNodeKind.DIR) {
					int[] sub = countLOC(repository, (path.equals("")) ? entry.getName() : path + "/" + entry.getName(), revision);
					count[0] += sub[0];
					count[1] += sub[1];
				}
				else if(entry.getName().endsWith(".java"))
				{
					count[0]++;
					ByteArrayOutputStream out = new ByteArrayOutputStream();
					repository.getFile(path + "/" + entry.getName(), revision, null, out);
					//System.out.println(path + "/" + entry.getName());
					String content = out.toString();
					int index = -1;
					while((index = content.indexOf('\n', index+1)) > -1)
						count[1]++;
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		return count;
	}
	
	private static class ASTPathVisitor extends ASTVisitor {
		private CompilationUnit root;
		private int start, end;
		public MethodDeclaration startMethod = null, endMethod = null;
		
		public ASTPathVisitor(CompilationUnit root, int start, int end) {
			this.root = root;
			this.start = start;
			this.end = end;
		}
		public boolean visit(TypeDeclaration node)
		{
			if(startMethod != null && endMethod != null)
				return false;
			return true;
		}
		public boolean visit(MethodDeclaration node)
		{
			if(startMethod != null && endMethod != null)
				return false;
			int start = root.getLineNumber(node.getStartPosition()), end = root.getLineNumber(node.getStartPosition() + node.getLength() - 1);
			if(start <= this.start && end >= this.start)
				startMethod = node;
			if(start <= this.end && end >= this.end)
				endMethod = node;
			
			return false;
		}
	}
	
	private static class ASTLeafVisitor extends ASTVisitor {
		private CompilationUnit root;
		public int start = -1, end;
		private Stack<Integer> numOfChildren = new Stack<Integer>();
		
		public ASTLeafVisitor(CompilationUnit root)
		{
			this.root = root;
		}
		
		public void preVisit(ASTNode node)
		{
			if(!this.numOfChildren.isEmpty())
			{
				int c = this.numOfChildren.pop();
				this.numOfChildren.push(c+1);
			}
			this.numOfChildren.push(0);
		}
		
		public void postVisit(ASTNode node)
		{
			int c = this.numOfChildren.pop();
			if(c == 0)
			{
				end = this.root.getLineNumber(node.getStartPosition());
				if(start == -1)
					start = end;
			}
		}
	}
}
