package cleman;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Stack;

import org.eclipse.jdt.core.dom.*;

import treeMapping.MapVisitor;
import treeMapping.TreeMapper;


/**
 * @author Nguyen Anh Hoan
 *
 */
public class VectorVisitor extends ASTVisitor {
	public static final String PROPERTY_CHILDREN = "c";
	public static final String PROPERTY_NUM_NODES = "n";
	public static final String PROPERTY_VECTORS = "v";
	public static byte[] indexer = new byte[84];	//category of this node type (Class, Method, Block, Statement, ..)
	public ArrayList<Integer> lines = null;
	SourceFile sourceFile;
	//String fileName;
    public ArrayList<Integer> fragments = new ArrayList<Integer>();	//fragments of one file
    public ArrayList<Integer> prevFragments = new ArrayList<Integer>();
    /*
	 * Stack of children's n-gram vectors
	 */
	private Stack<ArrayList<HashMap<Integer, Integer>>> stackChildrenVectors = new Stack<ArrayList<HashMap<Integer, Integer>>>();
	/*
	 * Stack of VERTICAL n-grams starting from the ROOT of the subtrees 
	 */
	private Stack<ArrayList<HashMap<Integer, Integer>>> stackChildrenRootVGrams = new Stack<ArrayList<HashMap<Integer, Integer>>>();
	/*
	 * Stack of HORIZONTAL sequences of children
	 */
	private Stack<ArrayList<Byte>> stackChildrenHSequences = new Stack<ArrayList<Byte>>();
	
	private Stack<HashSet<Fragment>> stackChildrenFragments = new Stack<HashSet<Fragment>>();
    
	private int lastNode = 0;	//for end token of vector
	private Stack<Integer> stackCurrentNode = new Stack<Integer>();	//for start token of vector

    public VectorVisitor() {

    }
    public VectorVisitor(String fileName) {
    	//this.fileName = fileName;
    }
    public VectorVisitor(SourceFile sourceFile) {
    	this.sourceFile = sourceFile;
    	//this.fileName = sourceFile.fileLocation;
    }
    static {
    	int index=0;
		for(byte i=0; i<84; i++) {
			if (i==0 || i==6  || i==11 || i==12 || i==17 || i==19 || i==20 || i==22 || i==26 || i==28 ||  i==30 || i==35 || 
					  i==46 || i==47 || i==52 || i==53 || i==54 || i==54 || i==63 || i==64 || (i>=71 && i<=83)
					  //|| i == 42
					  //|| i == ASTNode.BLOCK
					  ){
						indexer[i] = Fragment.NotConsideredFrags;
			}
			else
			{
				int gram = i << 24;
				Fragment.gram2Index.put(gram, index);
				Fragment.index2Gram.put(index++, gram);
				
				switch (i) {
				//case ASTNode.COMPILATION_UNIT: indexer[i] = Fragment.ClassFragment; break;
				//case ASTNode.TYPE_DECLARATION_STATEMENT: indexer[i] = Fragment.ClassFragment; break;
				case ASTNode.TYPE_DECLARATION: indexer[i] = Fragment.ClassFragment; break;
				case ASTNode.METHOD_DECLARATION: indexer[i] = Fragment.MethodFragment; break;
				//case ASTNode.BLOCK: indexer[i] = Fragment.BlockFragment; break;
				case ASTNode.DO_STATEMENT: indexer[i] = Fragment.LoopStatementFragment; break;
				case ASTNode.FOR_STATEMENT: indexer[i] = Fragment.LoopStatementFragment; break;
				case ASTNode.ENHANCED_FOR_STATEMENT: indexer[i] = Fragment.LoopStatementFragment; break;
				case ASTNode.WHILE_STATEMENT: indexer[i] = Fragment.LoopStatementFragment; break;
				case ASTNode.IF_STATEMENT: indexer[i] = Fragment.IfStatementFragment; break;
				//case ASTNode.SWITCH_CASE: indexer[i] = Fragment.IfStatementFragment; break;
				case ASTNode.SWITCH_STATEMENT: indexer[i] = Fragment.SwitchStatementFragment; break;
				
				case ASTNode.CONDITIONAL_EXPRESSION: indexer[i] = Fragment.Expression; break;
				case ASTNode.INFIX_EXPRESSION: indexer[i] = Fragment.Expression; break;
				case ASTNode.POSTFIX_EXPRESSION: indexer[i] = Fragment.Expression; break;
				case ASTNode.PREFIX_EXPRESSION: indexer[i] = Fragment.Expression; break;
				case ASTNode.PARENTHESIZED_EXPRESSION: indexer[i] = Fragment.Expression; break;
				case ASTNode.INSTANCEOF_EXPRESSION: indexer[i] = Fragment.Expression; break;
				
				case ASTNode.METHOD_INVOCATION: indexer[i] = Fragment.MethodState; break;
				case ASTNode.SUPER_METHOD_INVOCATION: indexer[i] = Fragment.MethodState; break;
				case ASTNode.CLASS_INSTANCE_CREATION: indexer[i] = Fragment.MethodState; break;
				case ASTNode.CONSTRUCTOR_INVOCATION: indexer[i] = Fragment.MethodState; break;
				case ASTNode.SUPER_CONSTRUCTOR_INVOCATION: indexer[i] = Fragment.MethodState; break;
				
				//case ASTNode.VARIABLE_DECLARATION_EXPRESSION: indexer[i] = Fragment.DeclarationExp; break;
				//case ASTNode.VARIABLE_DECLARATION_FRAGMENT: indexer[i] = Fragment.VarState; break;
				case ASTNode.VARIABLE_DECLARATION_STATEMENT: indexer[i] = Fragment.DeclarationState; break;
				case ASTNode.FIELD_DECLARATION: indexer[i] = Fragment.DeclarationState; break;
				
				case ASTNode.SIMPLE_NAME: indexer[i] = Fragment.SimpleName; break;
				case ASTNode.BOOLEAN_LITERAL: indexer[i] = Fragment.Literal; break;
				case ASTNode.CHARACTER_LITERAL: indexer[i] = Fragment.Literal; break;
				case ASTNode.STRING_LITERAL: indexer[i] = Fragment.Literal; break;
				case ASTNode.NUMBER_LITERAL: indexer[i] = Fragment.Literal; break;
				
				case ASTNode.ARRAY_ACCESS: indexer[i] = Fragment.ArrayState; break;
				case ASTNode.ARRAY_CREATION: indexer[i] = Fragment.ArrayState; break;
				case ASTNode.ARRAY_INITIALIZER: indexer[i] = Fragment.ArrayState; break;
				case ASTNode.ARRAY_TYPE: indexer[i] = Fragment.ArrayState; break;
				
				case ASTNode.ASSERT_STATEMENT: indexer[i] = Fragment.AssertState; break;
				case ASTNode.ASSIGNMENT: indexer[i] = Fragment.AssignState; break;
				
				case ASTNode.MEMBER_REF: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.METHOD_REF: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.METHOD_REF_PARAMETER: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.PRIMITIVE_TYPE: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.QUALIFIED_NAME: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.TAG_ELEMENT: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.TEXT_ELEMENT: indexer[i] = Fragment.OtherFragments; break;
				case ASTNode.TYPE_PARAMETER: indexer[i] = Fragment.OtherFragments; break;
				default: indexer[i] = Fragment.OtherStatementFragment; break;
				}
			}
		}
    }
    public void preVisit(ASTNode node) {
    	int aNode = node.getNodeType(); 
    	if (aNode == ASTNode.JAVADOC || aNode == ASTNode.BLOCK_COMMENT || aNode == ASTNode.LINE_COMMENT)
    		return;
    	
    	stackCurrentNode.push(++lastNode); 
    	
    	ArrayList<HashMap<Integer, Integer>> childrenVector = new ArrayList<HashMap<Integer, Integer>>();
    	stackChildrenVectors.push(childrenVector);
    	ArrayList<HashMap<Integer, Integer>> childrenGram = new ArrayList<HashMap<Integer, Integer>>();
    	stackChildrenRootVGrams.push(childrenGram);
    	ArrayList<Byte> childrenSequence = new ArrayList<Byte>();
    	stackChildrenHSequences.push(childrenSequence);
    	HashSet<Fragment> childrenFragments = new HashSet<Fragment>();
    	stackChildrenFragments.push(childrenFragments);
    	
    	node.setProperty(PROPERTY_NUM_NODES, 0);
    	node.setProperty(PROPERTY_CHILDREN, new ArrayList<ASTNode>());
    	
    	if(aNode == ASTNode.BLOCK)
    		node.setProperty(PROPERTY_VECTORS, new ArrayList<HashMap<Integer, Integer>>());
    	/*for (int i = 0; i < currentDepth; i++)
    		System.out.print("\t");
    	System.out.print("{" + indexer[aNode] + "\t" + node.getClass().getSimpleName());
    	if (aNode == ASTNode.SIMPLE_NAME)
    		System.out.println(": " + node.toString());
    	else
    		System.out.println();*/
    }
    public void postVisit(ASTNode node) {
    	int aNode = node.getNodeType();
    	if (aNode == ASTNode.JAVADOC || aNode == ASTNode.BLOCK_COMMENT || aNode == ASTNode.LINE_COMMENT)
    		return;
    	
    	/*for (int i = 0; i < currentDepth; i++)
    		System.out.print("\t");
    	System.out.println("}");*/
    	
    	buildFragment(node);
    }
    private void buildFragment(ASTNode node) {
    	int nodeType = node.getNodeType();
    	int currentNode = stackCurrentNode.pop();
    	node.setProperty(PROPERTY_NUM_NODES, lastNode-currentNode+1);

    	ArrayList<HashMap<Integer, Integer>> childrenVectors = stackChildrenVectors.pop();
    	ArrayList<HashMap<Integer, Integer>> childrenRootVGrams = stackChildrenRootVGrams.pop();
    	ArrayList<Byte> childrenHSequence = stackChildrenHSequences.pop();
    	/*
    	 * VERTICAL n-grams starting from this node
    	 */
    	HashMap<Integer, Integer> myRootVGrams = new HashMap<Integer, Integer>();
    	
    	HashMap<Integer, Integer> vector = new HashMap<Integer, Integer>();
    	/*
    	 * Adding vectors of all children
    	 */
    	if (!childrenVectors.isEmpty()) {
    		vector.putAll(new HashMap<Integer, Integer>(childrenVectors.get(0)));
    		for(int i = 1; i < childrenVectors.size(); i++) {
    			HashMap<Integer, Integer> childVector = childrenVectors.get(i);
    			for (int index : childVector.keySet()) {
    				if (vector.containsKey(index))
    					vector.put(index, (vector.get(index) + childVector.get(index)));
    				else 
    					vector.put(index, childVector.get(index));
    			}
    		}
    	}
    	/*
    	 * Adding all new horizontal n-grams
    	 */
    	if (nodeType != ASTNode.BLOCK && nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.PARENTHESIZED_EXPRESSION)
	    	for (int i = 0; i < childrenHSequence.size()-1; i++) {
	    		int gram = childrenHSequence.get(i);
	    		for (int j = 2; j <= Fragment.maxSizeOfGram; j++) {
	    			if (i+j-1 < childrenHSequence.size()) {
	    				gram = (gram << 4) + childrenHSequence.get(i+j-1);
	    				int index;
	    				if (Fragment.gram2Index.containsKey(gram))
	    					index = Fragment.gram2Index.get(gram);
	    				else {
	    					index = Fragment.gram2Index.size();
	    					Fragment.gram2Index.put(gram, index);
	    					Fragment.index2Gram.put(index, gram);
	    				}
	    				if (vector.containsKey(index))
	    					vector.put(index, vector.get(index) + 1);
	    				else 
	    					vector.put(index, 1);
	    			}
	    			else break;
	    		}
	    	}
    	/*
    	 * This node is also a single node type in the vector
    	 */
		if ((indexer[nodeType] != Fragment.NotConsideredFrags)) {
    	//if(indexer[nodeType] <= 11) {
			int gram = nodeType << 24;
			int index =	Fragment.gram2Index.get(gram);
			if (vector.containsKey(index))
				vector.put(index, vector.get(index) + 1);
			else 
				vector.put(index, 1);
		}
		/*
		 * This node is also a 1-gram in the vector
		 */
		if(indexer[nodeType] <= 11) {
			int gram = (1<<24) + indexer[nodeType];
			int tmpIndex;
			if (Fragment.gram2Index.containsKey(gram))
				tmpIndex = Fragment.gram2Index.get(gram);
			else {
				tmpIndex = Fragment.gram2Index.size();
				Fragment.gram2Index.put(gram, tmpIndex);
				Fragment.index2Gram.put(tmpIndex, gram);
			}
			if (myRootVGrams.containsKey(tmpIndex))
				myRootVGrams.put(tmpIndex, myRootVGrams.get(tmpIndex) + 1);
			else 
				myRootVGrams.put(tmpIndex, 1);
		}
		/*
		 * Building all n-grams starting from this node (will be used by its parent)
		 */
		if (!childrenRootVGrams.isEmpty()) {
			if(indexer[nodeType] <= 11) {
				for (HashMap<Integer, Integer> childGram : childrenRootVGrams) {
					for (int index : childGram.keySet()) {
						int gram = Fragment.index2Gram.get(index);
						int sizeOfChildGram = getSizeOfGram(gram);
						gram = (indexer[nodeType] << (4 * sizeOfChildGram)) | gram;
						int tmpIndex;
						if (Fragment.gram2Index.containsKey(gram))
							tmpIndex = Fragment.gram2Index.get(gram);
	    				else {
	    					tmpIndex = Fragment.gram2Index.size();
	    					Fragment.gram2Index.put(gram, tmpIndex);
	    					Fragment.index2Gram.put(tmpIndex, gram);
	    				}
	    				if (vector.containsKey(tmpIndex))
	    					vector.put(tmpIndex, vector.get(tmpIndex) + childGram.get(index));
	    				else 
	    					vector.put(tmpIndex, childGram.get(index));
	    				if (sizeOfChildGram+1 < Fragment.maxSizeOfGram) {
	    					if (myRootVGrams.containsKey(tmpIndex))
	    						myRootVGrams.put(tmpIndex, myRootVGrams.get(tmpIndex) + childGram.get(index));
		    				else 
		    					myRootVGrams.put(tmpIndex, childGram.get(index));
	    				}
	    			}
				}
			}
			else {
				for (HashMap<Integer, Integer> childGram : childrenRootVGrams) {
					for (int index : childGram.keySet()) {
    					if (myRootVGrams.containsKey(index))
    						myRootVGrams.put(index, myRootVGrams.get(index) + childGram.get(index));
	    				else 
	    					myRootVGrams.put(index, childGram.get(index));
	    			}
				}
			}
		}
		HashSet<Fragment> childrenFragments = stackChildrenFragments.pop();
		if((nodeType == ASTNode.EXPRESSION_STATEMENT || nodeType == ASTNode.BLOCK || nodeType == ASTNode.PARENTHESIZED_EXPRESSION) && !stackChildrenFragments.isEmpty())
		{
			HashSet<Fragment> temp = stackChildrenFragments.pop();
    		temp.addAll(childrenFragments);
    		stackChildrenFragments.push(temp);
		}
		if(node.getParent() != null && node.getParent().getNodeType() == ASTNode.BLOCK)
		{
			ArrayList<HashMap<Integer, Integer>> vectors = (ArrayList<HashMap<Integer, Integer>>)node.getParent().getProperty(PROPERTY_VECTORS);
			vectors.add(vector);
			node.getParent().setProperty(PROPERTY_VECTORS, vectors);
		}
		/*
		 * Build the corresponding fragment
		 */
		if(nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.BLOCK && nodeType != ASTNode.PARENTHESIZED_EXPRESSION && 
				lastNode-currentNode+1 >= Main.minFragmentSize)
		{
			Fragment fragment;
			ASTNode mappedNode = (ASTNode)node.getProperty(TreeMapper.propertyRevisionMap);
			int mappedSID = -1;
			if(mappedNode == null || mappedNode.getProperty(SequenceIDVisitor.propertySequenceID) == null)
				fragment = new Fragment();
			else
			{
				//System.err.println(mappedNode.getClass().getSimpleName() + ": " + mappedNode.getProperty(SequenceIDVisitor.propertySequenceID));
				mappedSID = (Integer)mappedNode.getProperty(SequenceIDVisitor.propertySequenceID);
				fragment = new Fragment(prevFragments.get(mappedSID));
				/*String preContent = Fragment.all.get(prevFragments.get(mappedSID)).getCode();
				String mappedContent = mappedNode.toString();
				if(!mappedContent.equals(preContent))
				{
					System.err.println("Error in getting mapped fragment in building");
					System.err.println(preContent + "\nbecomes\n" + mappedContent);
				}*/
			}
			fragment.setSourceFile(this.sourceFile);
			//fragment.setAstNode(node);
	    	if(Main.doReport)
	    		fragment.setCode(node.toString());
	    	fragment.setStartChar(node.getStartPosition());
          // fragment.setAstNode(node);
	    //	node.getRoot().get
	    	fragment.setLength(node.getLength());
	    	//fragment.setStartLine(binSearch(node.getStartPosition(), this.lines));
	    	//fragment.setEndLine(binSearch(node.getStartPosition()+node.getLength()-1, this.lines));
	    	fragment.setGramVector(vector);
	    	fragment.setType((byte)((indexer[nodeType] < 8) ? indexer[nodeType] : 0));
	    	fragment.setVectorLength();
	    	if(mappedNode == null || mappedNode.getProperty(SequenceIDVisitor.propertySequenceID) == null)
			{
				Fragment.addeds.put(fragment.getId(), fragment);
				//Fragment.all.remove(fragment.getId());
			}
			else if((Integer)node.getProperty(MapVisitor.propertyStatus) != MapVisitor.STATUS_UNCHANGED) 
				//&& !vector.equals(prevFragments.get(mappedSID).getSparseGramVector()))
			{
				Fragment.modifieds.put(fragment.getId(), fragment);
				Fragment.all.remove(fragment.getId());
			}
			else
			{
				Fragment prevFragment = Fragment.all.get(prevFragments.get(mappedSID));
				if(prevFragment == null)
				{
					System.err.println("Error in building fragment: no prev fragment " + node.getClass().getSimpleName());
					System.err.println(Fragment.prev.get(prevFragments.get(mappedSID)).toString());
					System.err.println(node.toString());
					System.err.println(mappedNode.toString());
					/*if(Fragment.prev.get(prevFragments.get(mappedSID)).getCode().equals(node.toString()))
						prevFragment = Fragment.prev.get(prevFragments.get(mappedSID));*/
				}
				fragment.setBuckets(new HashSet<Bucket>(prevFragment.getBuckets()));
				fragment.setClones(new HashSet<Integer>(prevFragment.getClones()));
				Fragment.unchangeds.put(fragment.getId(), fragment);
				Fragment.all.remove(fragment.getId());
			}
				
	    	if(!stackChildrenFragments.isEmpty())
	    	{
	    		HashSet<Fragment> temp = stackChildrenFragments.pop();
	    		temp.add(fragment);
	    		stackChildrenFragments.push(temp);
	    	}
	    	fragments.add(fragment.getId());	//add to this file's fragments
	    	/*System.out.println("In " + this.sourceFile.fileLocation);
	    	for(int id : vector.keySet())
    		{
	    		int gram = Fragment.index2Gram.get(id);
    			System.out.print(Fragment.nGramToString(gram) + ":" + vector.get(id) + " ");
    		}
    		System.out.println();*/
	    	
		}
		else if(Fragment.mergeWindowSize > 1 && nodeType == ASTNode.BLOCK && lastNode-currentNode+1 >= Main.minFragmentSize)
		{
			ArrayList<ASTNode> childrenNodes = (ArrayList<ASTNode>)node.getProperty(PROPERTY_CHILDREN);
			int[] childrenRootGrams = new int[childrenNodes.size()];
			for(int i = 0; i < childrenNodes.size(); i++)
			{
				ASTNode childNode = childrenNodes.get(i);
				int type = indexer[childNode.getNodeType()];
				if(childNode.getNodeType() == ASTNode.EXPRESSION_STATEMENT)
				{
					ExpressionStatement exprNode = (ExpressionStatement)childNode;
					type = indexer[exprNode.getExpression().getNodeType()];
				}
				if(type <= 11)
				{
					childrenRootGrams[i] = type;
				}
				else 
					childrenRootGrams[i] = -1;
			}
			for(int start = 0; start < childrenNodes.size()-1; start++)
    		{
    			int end = start + 1;
    			while(end < childrenNodes.size() && end < start+Fragment.mergeWindowSize)
    			{
    				int numNodes = 0;
    				for(int j = start; j <= end; j++)
    				{
    					ASTNode child = childrenNodes.get(j);
    					numNodes += (Integer)child.getProperty(PROPERTY_NUM_NODES);
    				}
    				//System.out.println("In " + this.sourceFile.fileLocation + " " + start + " " + end + " " + numNodes);
    				if(numNodes >= Main.minFragmentSize)
    				{
    					Fragment mergeFragment = new Fragment();
    					mergeFragment.setSourceFile(sourceFile);
    					mergeFragment.setStartChar(childrenNodes.get(start).getStartPosition());
    					mergeFragment.setLength(childrenNodes.get(end).getStartPosition() - mergeFragment.getStartChar() + childrenNodes.get(end).getLength());
    					//mergeFragment.setStartLine(binSearch(childrenNodes.get(start).getStartPosition(), this.lines));
    					//mergeFragment.setEndLine(binSearch(childrenNodes.get(end).getStartPosition()+childrenNodes.get(end).getLength()-1, this.lines));
   					
    					mergeFragment.setType(Fragment.BlockFragment);
    					HashMap<Integer, Integer> mergeVector = new HashMap<Integer, Integer>();
    			    	/*
    			    	 * Adding vectors of all children
    			    	 */
    					ArrayList<HashMap<Integer, Integer>> vectors = (ArrayList<HashMap<Integer, Integer>>)node.getProperty(PROPERTY_VECTORS);
			    		mergeVector.putAll(new HashMap<Integer, Integer>(vectors.get(start)));
			    		for(int l = start+1; l <= end; l++) {
			    			HashMap<Integer, Integer> childVector = vectors.get(l);
			    			for (int index : childVector.keySet()) {
			    				if (mergeVector.containsKey(index))
			    					mergeVector.put(index, (mergeVector.get(index) + childVector.get(index)));
			    				else 
			    					mergeVector.put(index, childVector.get(index));
			    			}
			    		}
			    		// Adding horizontal grams
			    		ArrayList<Integer> childrenGrams = new ArrayList<Integer>();
			    		for(int i = start; i <= end; i++)
			    			if(childrenRootGrams[i] > 0)
			    				childrenGrams.add(childrenRootGrams[i]);
			    		for (int i = 0; i < childrenGrams.size()-1; i++) {
				    		int gram = childrenGrams.get(i);
				    		for (int j = 2; j <= Fragment.maxSizeOfGram; j++) {
				    			if (i+j-1 < childrenGrams.size()) {
				    				gram = (gram << 4) + childrenGrams.get(i+j-1);
				    				int index;
				    				if (Fragment.gram2Index.containsKey(gram))
				    					index = Fragment.gram2Index.get(gram);
				    				else {
				    					index = Fragment.gram2Index.size();
				    					Fragment.gram2Index.put(gram, index);
				    					Fragment.index2Gram.put(index, gram);
				    				}
				    				if (mergeVector.containsKey(index))
				    					mergeVector.put(index, mergeVector.get(index) + 1);
				    				else 
				    					mergeVector.put(index, 1);
				    			}
				    			else break;
				    		}
				    	}
			    		mergeFragment.setGramVector(mergeVector);
			    		mergeFragment.setVectorLength();
			    		fragments.add(mergeFragment.getId());
			    		Fragment.addeds.put(mergeFragment.getId(), mergeFragment);
			    		/*for(int id : mergeVector.keySet())
			    		{
			    			int gram = Fragment.index2Gram.get(id);
			    			System.out.print(Fragment.nGramToString(gram) + ":" + mergeVector.get(id) + " ");
			    		}
			    		System.out.println();*/
    				}
    				end++;
    			}
    		}
		}
		/*
		 * Pushing to the stacks respectively
		 */
    	if (!stackChildrenVectors.isEmpty()) {	//if not root
    		if (!vector.isEmpty()) {
	    		ArrayList<HashMap<Integer, Integer>> parentVectors = stackChildrenVectors.pop();	//get siblings
	    		parentVectors.add(vector); 	//join them (append this node type)
	    		stackChildrenVectors.push(parentVectors);	//back home
	    		ArrayList<ASTNode> childrenNodes = (ArrayList<ASTNode>)node.getParent().getProperty(PROPERTY_CHILDREN);
	    		childrenNodes.add(node);
	    		node.getParent().setProperty(PROPERTY_CHILDREN, childrenNodes);
    		}
    		if (!myRootVGrams.isEmpty()) {
    			ArrayList<HashMap<Integer, Integer>> parentGrams = stackChildrenRootVGrams.pop();	//get siblings
        		parentGrams.add(myRootVGrams); 	//join them (append this node type)
        		stackChildrenRootVGrams.push(parentGrams);	//back home
    		}
    		if (indexer[nodeType] <= 11) {
    			ArrayList<Byte> sequence = stackChildrenHSequences.pop();	//get siblings
    			sequence.add(indexer[nodeType]);	//	join them (append this node type)
    			stackChildrenHSequences.push(sequence);
    		}
    		else if (nodeType == ASTNode.EXPRESSION_STATEMENT || nodeType == ASTNode.BLOCK || nodeType == ASTNode.PARENTHESIZED_EXPRESSION) {
    			ArrayList<Byte> sequence = stackChildrenHSequences.pop();
    			sequence.addAll(childrenHSequence);
    			stackChildrenHSequences.push(sequence);
    		}
    	}
    }
    byte getSizeOfGram(int gram) {
    	byte i = 0;
    	int temp = gram  & ((1<<24) - 1);
    	//gram = Math.abs(gram);
    	while (temp != 0) {
    		i++;
    		temp /= 16;
    	}
    	return i;
    }
    /**
	* 
	* @param x
	* @param al
	* @return
	*/
	public static  int binSearch(int x, ArrayList<Integer> al) {
				
		if (al.size() < 1) return al.size()-1;
		int low = 0, high = al.size()-1, mid;
		while(low < high) {
			mid = (low + high) / 2;
            if(al.get(mid) < x)
                low = mid + 1;
            else if(al.get(mid) > x)
                high = mid - 1;
            else
                return mid;
        }
		if (al.get(low) > x) return low - 1;
		else return low;
	}

 }
