/**
 * 
 */
package repositoryConnecting;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Scanner;

import org.tmatesoft.svn.core.SVNDepth;
import org.tmatesoft.svn.core.SVNDirEntry;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.SVNURL;
import org.tmatesoft.svn.core.auth.ISVNAuthenticationManager;
import org.tmatesoft.svn.core.internal.io.dav.DAVRepositoryFactory;
import org.tmatesoft.svn.core.io.SVNRepository;
import org.tmatesoft.svn.core.io.SVNRepositoryFactory;
import org.tmatesoft.svn.core.wc.ISVNDiffStatusHandler;
import org.tmatesoft.svn.core.wc.SVNClientManager;
import org.tmatesoft.svn.core.wc.SVNDiffClient;
import org.tmatesoft.svn.core.wc.SVNDiffStatus;
import org.tmatesoft.svn.core.wc.SVNRevision;
import org.tmatesoft.svn.core.wc.SVNStatusType;
import org.tmatesoft.svn.core.wc.SVNWCUtil;

import cleman.SourceFile;

/**
 * @author Nguyen Anh Hoan
 *
 */
public class SVNChangedFiles {
	private SVNRepository repository = null;
	private ISVNAuthenticationManager authManager;
	private SVNClientManager clientManager = null;
	private SVNURL url;
	private String username;
	private String password;
	
	/**
	 * 
	 */
	public SVNChangedFiles(SVNRepository rep)
	{
		this.repository = rep;
		this.url = rep.getLocation();
	}
	/**
	 * 
	 */
	public SVNChangedFiles(String url, String username, String password)
	{
		try {
			this.url = SVNURL.parseURIEncoded(url);
			this.username = username;
			this.password = password;
			DAVRepositoryFactory.setup();
		}
		catch (SVNException e){
			e.printStackTrace();
		}
		
	}
	/**
	 * 
	 */
	public boolean connect()
	{
		try {
			authManager = SVNWCUtil.createDefaultAuthenticationManager(username, password);
			repository = SVNRepositoryFactory.create(this.url);
			repository.setAuthenticationManager(authManager);
			repository.testConnection();
			//System.out.println("Connect OK");
		}
		catch (SVNException e) {
			//e.printStackTrace();
			return false;
		}
		return true;
	}
	public String getDiff(long r1, long r2)
	{
		clientManager = SVNClientManager.newInstance(SVNWCUtil.createDefaultOptions(true), authManager);
		SVNDiffClient diffClient = clientManager.getDiffClient();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			diffClient.doDiff(url, SVNRevision.create(r1), url , SVNRevision.create(r2), SVNDepth.INFINITY, false, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		Scanner scanner = new Scanner(out.toString());
		StringBuffer buff = new StringBuffer();
		while(scanner.hasNextLine())
		{
			buff.append(scanner.nextLine() + "\n");
		}
		
		return buff.toString();
	}
	public String getDiff(String fileName, long r1, long r2)
	{
		clientManager = SVNClientManager.newInstance(SVNWCUtil.createDefaultOptions(true), authManager);
		SVNDiffClient diffClient = clientManager.getDiffClient();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			diffClient.doDiff(url.appendPath(fileName, false), SVNRevision.create(r1), 
					url.appendPath(fileName, false), SVNRevision.create(r2), SVNDepth.INFINITY, false, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		Scanner scanner = new Scanner(out.toString());
		StringBuffer buff = new StringBuffer();
		while(scanner.hasNextLine())
		{
			buff.append(scanner.nextLine() + "\n");
		}
		
		return buff.toString();
	}
	public boolean checkExistence(String path, long revision)
	{
		try {
			return repository.checkPath(path, revision) == SVNNodeKind.FILE;
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		
		return false;
	}
	public String getFile(String path, long revision)
	{
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			repository.getFile(path, revision, null, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		
		return out.toString();
	}
	public ArrayList<SVNDiffStatus> getDiffStatus(long r1, long r2, HashSet<String> delFiles, HashSet<String> addFiles, HashSet<String> modFiles)
	{
		clientManager = SVNClientManager.newInstance ();
		SVNDiffClient diffClient = clientManager.getDiffClient ();

		SVNDiffStatusHandler handler = new SVNDiffStatusHandler(r1, r2);
		try {
			diffClient.doDiffStatus (this.url, SVNRevision.create(r1), this.url, SVNRevision.create(r2), SVNDepth.INFINITY, true, handler);
		} catch (SVNException e) {
			e.printStackTrace();
		}
		for(SVNDiffStatus diffStatus : handler.changes)
		{
			SVNStatusType status = diffStatus.getModificationType ();
			String path = diffStatus.getPath ();
			if(path.endsWith("/Interpreter.java") || path.endsWith("/Parser.java"))
				continue;
			if (status == SVNStatusType.STATUS_ADDED && path.endsWith(".java"))
			{
				addFiles.add(path);
			}
			else if (status == SVNStatusType.STATUS_DELETED)
			{
				if(diffStatus.getKind() == SVNNodeKind.DIR)
				{
					delFiles.addAll(getFiles(repository, path, r1));
				}
				else if(path.endsWith(".java"))
					delFiles.add(path); 
			}
			else if (status == SVNStatusType.STATUS_MODIFIED && path.endsWith(".java"))
			{
				modFiles.add(path);
			}
		}
		
		return handler.changes;
	}
	
	private class SVNDiffStatusHandler implements ISVNDiffStatusHandler
	{
		public ArrayList<SVNDiffStatus> changes = new ArrayList<SVNDiffStatus>();
		private long r1, r2;
		
		public SVNDiffStatusHandler(long r1, long r2)
		{
			this.r1 = r1;
			this.r2 = r2;
		}
		public void handleDiffStatus (SVNDiffStatus diffStatus) throws SVNException
		{
			changes.add(diffStatus);
		}
    }
	public HashSet<String> getFiles(SVNRepository repository, String path, long revision)
	{
		HashSet<String> result = new HashSet<String>();
		try {
			Collection entries = repository.getDir(path, revision, null, (Collection)null);
			Iterator iterator = entries.iterator();
			while (iterator.hasNext()) {
				SVNDirEntry entry = (SVNDirEntry)iterator.next();
				if(entry.getKind() == SVNNodeKind.FILE && entry.getName().endsWith(".java"))
					result.add(path + "/" + entry.getName());
				else if (entry.getKind() == SVNNodeKind.DIR) {
					result.addAll(getFiles(repository, (path.equals("")) ? entry.getName() : path + "/" + entry.getName(), revision));
				}
			}
		} catch(Exception e) {
			e.printStackTrace();
		}
		
		return result;
	}
}
