/**
 * 
 */
package repositoryConnecting;

import java.io.ByteArrayOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Scanner;
import java.util.Set;

import org.tmatesoft.svn.core.SVNDepth;
import org.tmatesoft.svn.core.SVNException;
import org.tmatesoft.svn.core.SVNLogEntry;
import org.tmatesoft.svn.core.SVNLogEntryPath;
import org.tmatesoft.svn.core.SVNNodeKind;
import org.tmatesoft.svn.core.SVNURL;
import org.tmatesoft.svn.core.auth.ISVNAuthenticationManager;
import org.tmatesoft.svn.core.internal.io.dav.DAVRepositoryFactory;
import org.tmatesoft.svn.core.io.SVNRepository;
import org.tmatesoft.svn.core.io.SVNRepositoryFactory;
import org.tmatesoft.svn.core.wc.SVNClientManager;
import org.tmatesoft.svn.core.wc.SVNDiffClient;
import org.tmatesoft.svn.core.wc.SVNRevision;
import org.tmatesoft.svn.core.wc.SVNWCUtil;

/**
 * @author Nguyen Anh Hoan
 *
 */
public class SVNConnector {
	public SVNRepository repository = null;
	private ISVNAuthenticationManager authManager;
	private SVNClientManager clientManager = null;
	private SVNURL url;
	private String username;
	private String password;
	
	/**
	 * 
	 */
	public SVNConnector(SVNRepository rep)
	{
		this.repository = rep;
		this.url = rep.getLocation();
	}
	/**
	 * 
	 */
	public SVNConnector(String url, String username, String password)
	{
		try {
			this.url = SVNURL.parseURIEncoded(url);
			this.username = username;
			this.password = password;
			DAVRepositoryFactory.setup();
		}
		catch (SVNException e){
			e.printStackTrace();
		}
		
	}
	/**
	 * 
	 */
	public boolean connect()
	{
		try {
			authManager = SVNWCUtil.createDefaultAuthenticationManager(username, password);
			repository = SVNRepositoryFactory.create(this.url);
			repository.setAuthenticationManager(authManager);
			repository.testConnection();
			//System.out.println("Connect OK");
		}
		catch (SVNException e) {
			//e.printStackTrace();
			return false;
		}
		return true;
	}
	public String getDiff(long r1, long r2)
	{
		clientManager = SVNClientManager.newInstance(SVNWCUtil.createDefaultOptions(true), authManager);
		SVNDiffClient diffClient = clientManager.getDiffClient();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			diffClient.doDiff(url, SVNRevision.create(r1), url , SVNRevision.create(r2), SVNDepth.INFINITY, false, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		Scanner scanner = new Scanner(out.toString());
		StringBuffer buff = new StringBuffer();
		while(scanner.hasNextLine())
		{
			buff.append(scanner.nextLine() + "\n");
		}
		
		return buff.toString();
	}
	public String getDiff(String fileName, long r1, long r2)
	{
		clientManager = SVNClientManager.newInstance(SVNWCUtil.createDefaultOptions(true), authManager);
		SVNDiffClient diffClient = clientManager.getDiffClient();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			diffClient.doDiff(url.appendPath(fileName, false), SVNRevision.create(r1), 
					url.appendPath(fileName, false), SVNRevision.create(r2), SVNDepth.INFINITY, false, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		Scanner scanner = new Scanner(out.toString());
		StringBuffer buff = new StringBuffer();
		while(scanner.hasNextLine())
		{
			buff.append(scanner.nextLine() + "\n");
		}
		
		return buff.toString();
	}
	public boolean checkExistence(String path, long revision)
	{
		try {
			return repository.checkPath(path, revision) == SVNNodeKind.FILE;
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		
		return false;
	}
	public String getFile(String path, long revision)
	{
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			repository.getFile(path, revision, null, out);
		}
		catch (SVNException e) {
			e.printStackTrace();
		}
		
		return out.toString();
	}
	public ArrayList<Integer> getFixRevisions(long start, long end)
	{
		ArrayList<Integer> revs = new ArrayList<Integer>();
		Collection logEntries = null;
        try {
            logEntries = repository.log(new String[] {""}, null, start, end, true, true);
        } catch (SVNException svne) {
            System.out.println("error while collecting log information for '"
                    + url + "': " + svne.getMessage());
            System.exit(1);
        }
        for (Iterator entries = logEntries.iterator(); entries.hasNext();) {
            /*
             * gets a next SVNLogEntry
             */
            SVNLogEntry logEntry = (SVNLogEntry) entries.next();
            String message = logEntry.getMessage().toLowerCase();
            if(message.contains("fix") || message.contains("bug"))
            {
            	if (logEntry.getChangedPaths().size() > 0) {
                    /*
                     * keys are changed paths
                     */
                    Set changedPathsSet = logEntry.getChangedPaths().keySet();

                    for (Iterator changedPaths = changedPathsSet.iterator(); changedPaths
                            .hasNext();) {
                        /*
                         * obtains a next SVNLogEntryPath
                         */
                        SVNLogEntryPath entryPath = (SVNLogEntryPath) logEntry
                                .getChangedPaths().get(changedPaths.next());
                        /*
                         * SVNLogEntryPath.getPath returns the changed path itself;
                         * 
                         * SVNLogEntryPath.getType returns a charecter describing
                         * how the path was changed ('A' - added, 'D' - deleted or
                         * 'M' - modified);
                         * 
                         * If the path was copied from another one (branched) then
                         * SVNLogEntryPath.getCopyPath &
                         * SVNLogEntryPath.getCopyRevision tells where it was copied
                         * from and what revision the origin path was at.
                         */
                        if(entryPath.getPath().endsWith(".java"))
                        {
                        	revs.add((int)logEntry.getRevision());
                        	System.out.println(logEntry.getRevision());
                        	break;
                        }
                    }
                }
            }
        }
        Collections.sort(revs);
		
		return revs;
	}
}
