package treeMapping;

import java.util.ArrayList;

import org.eclipse.jdt.core.dom.*;

import util.FileIO;
import util.StringProcessor;

public class LineDiff {
	private MapSourceFile file1, file2;
	
	private class TreeVisitor extends ASTVisitor {
		private ArrayList<String> lineContents;
		private ArrayList<Integer> lines;
		private boolean isLeaf = true;
		
		public TreeVisitor(ArrayList<Integer> lines)
		{
			this.lines = lines;
			this.lineContents = new ArrayList<String>();
			for(int i = 0; i < lines.size(); i++)
				lineContents.add("");
		}
		
		public ArrayList<String> getLineContents() {
			return lineContents;
		}

		public void preVisit(ASTNode node)
		{
			isLeaf = true;
		}
		public void postVisit(ASTNode node)
		{
			int nodeType = node.getNodeType();
			if(nodeType != ASTNode.BLOCK && nodeType != ASTNode.EXPRESSION_STATEMENT && nodeType != ASTNode.PARENTHESIZED_EXPRESSION)
			{
				int startLine = binSearch(node.getStartPosition(), lines);
		    	String code = "";
		    	if(isLeaf)
				{
		    		//if(nodeType != ASTNode.BOOLEAN_LITERAL && nodeType != ASTNode.CHARACTER_LITERAL && nodeType != ASTNode.NUMBER_LITERAL)
		    			code = node.toString();
		    		/*else
		    			code = node.getClass().getSimpleName();*/
		    		isLeaf = false;
				}
				else
				{
					code = node.getClass().getSimpleName();
				}
		    	lineContents.set(startLine, lineContents.get(startLine) + " " + code);
			}
			else
				isLeaf = false;
		}
		/*
		 * 
		 */
		private int binSearch(int x, ArrayList<Integer> al) {
			if (al.size() <= 1) return al.size()-1;
			int low = 0, high = al.size()-1, mid;
			while(low < high) {
				mid = (low + high) / 2;
				if(al.get(mid) < x)
					low = mid + 1;
				else if(al.get(mid) > x)
					high = mid - 1;
				else
					return mid;
			}
			if (al.get(low) > x) return low - 1;
			else return low;
		}
	}
	
	public LineDiff(MapSourceFile file1, MapSourceFile file2)
	{
		this.file1 = file1;
		this.file2 = file2;
	}
	
	public ArrayList<ArrayList<Integer>> doDiff() {
		TreeVisitor visitor = new TreeVisitor(file1.getLines());
		file1.getAst().accept(visitor);
		ArrayList<String> lines1 = visitor.getLineContents();
		visitor = new TreeVisitor(file2.getLines());
		file2.getAst().accept(visitor);
		ArrayList<String> lines2 = visitor.getLineContents();
		
		ArrayList<Integer> lcsM = new ArrayList<Integer>();
		ArrayList<Integer> lcsN = new ArrayList<Integer>();
		StringProcessor.doLCS(lines1, lines2, 16, lcsM, lcsN);
		
		ArrayList<ArrayList<Integer>> result = new ArrayList<ArrayList<Integer>>();
		result.add(lcsM);
		result.add(lcsN);

		return result;
	}
	
	public ArrayList<ArrayList<Integer>> doDiffO()
	{
		TreeVisitor visitor = new TreeVisitor(file1.getLines());
		file1.getAst().accept(visitor);
		ArrayList<String> lines1 = visitor.getLineContents();
		visitor = new TreeVisitor(file2.getLines());
		file2.getAst().accept(visitor);
		ArrayList<String> lines2 = visitor.getLineContents();
		
		int lenM = lines1.size(), lenN = lines2.size();
		int[][] d = new int[lenM+1][lenN+1];
		String[] codeM = new String[lenM+1];
		String[] codeN = new String[lenN+1];
		String[][] p = new String[lenM+1][lenN+1];
		d[0][0] = 0;
		for(int i = 1; i <= lenM; i++)
		{
			d[i][0] = 0;
			codeM[i] = lines1.get(i-1).toString();
		}
		for(int i = 1; i <= lenN; i++)
		{
			d[0][i] = 0;
			codeN[i] = lines2.get(i - 1).toString();
		}
		for(int i = 1; i <= lenM; i++)
		{
			for(int j = 1; j <= lenN; j++)
			{
				if(codeM[i].equals(codeN[j]) 
						&& ((i > 1 && j > 1 && codeM[i-1].equals(codeN[j-1])) || (i < lenM && j < lenN && codeM[i+1].equals(codeN[j+1])))
						)
				{
					d[i][j] = d[i-1][j-1] + 1;
					p[i][j] = "LU";
				}
				else if(d[i-1][j] >= d[i][j-1])
				{
					d[i][j] = d[i-1][j];
					p[i][j] = "U";
				}
				else
				{
					d[i][j] = d[i][j-1];
					p[i][j] = "L";
				}
			}
		}
		ArrayList<Integer> lcsM = new ArrayList<Integer>();
		ArrayList<Integer> lcsN = new ArrayList<Integer>();
		printLCS(p, lenM, lenN, lcsM, lcsN);
		
		ArrayList<ArrayList<Integer>> result = new ArrayList<ArrayList<Integer>>();
		result.add(lcsM);
		result.add(lcsN);
		
		return result;
	}
	/**
	 * 
	 */
	/*private void printLCS(String[][] p, int i, int j, ArrayList<Integer> lcsM, ArrayList<Integer> lcsN)
	{
		if(i == 0 || j == 0)
			return;
		if(p[i][j].equals("LU"))
		{
			printLCS(p, i-1, j-1, lcsM, lcsN);
			lcsM.add(i-1);
			lcsN.add(j-1);
		}
		else if(p[i][j].equals("U"))
			printLCS(p, i-1, j, lcsM, lcsN);
		else
			printLCS(p, i, j-1, lcsM, lcsN);
	}*/
	private void printLCS(String[][] p, int lenM, int lenN, ArrayList<Integer> lcsM, ArrayList<Integer> lcsN)
	{
		int i = lenM, j = lenN;
		while(i > 0 && j > 0)
		{
			if(p[i][j].equals("LU"))
			{
				lcsM.add(0, i-1);
				lcsN.add(0, j-1);
				i--; j--;
			}
			else if(p[i][j].equals("U"))
				i--;
			else
				j--;
		}
	}
}
