package treeMapping;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Scanner;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.Comment;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.core.runtime.IProgressMonitor;

public class MapSourceFile implements Serializable, Cloneable {
	private static final long serialVersionUID = 1L;
	
	public static int nextID = 1;
	/**
	 * This will contain the file ID, which will be the hashCode
	 */
	private int fileID;
	private String filePath;
	private String fileContent;
	private ArrayList<Integer> lines = new ArrayList<Integer>();
	private ASTNode ast;
	private ArrayList<Integer> changedLines;
	private ArrayList<Integer> unchangedLines;
	private HashSet<Integer> commentLines = new HashSet<Integer>();
	private HashMap<ASTNode, ArrayList<ASTNode>> tree = new HashMap<ASTNode, ArrayList<ASTNode>>();
	private ArrayList<ASTNode> unchangedLeaves = new ArrayList<ASTNode>();
	private ArrayList<ASTNode> allLeaves = new ArrayList<ASTNode>();
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.fileID;
	}
	public MapSourceFile(ASTNode ast)
	{
		this.ast = ast;
		this.fileID = nextID++;
		this.fileContent = ast.toString();
		buildLinesWoChange();
	}
	public MapSourceFile(String fileContent)
	{
		this.fileID = nextID++;
		this.fileContent = fileContent;
		buildLines();
	}
	public String getFilePath() {
		return filePath;
	}
	public void setFilePath(String filePath) {
		this.filePath = filePath;
	}
	/**
	 * @return the lines
	 */
	public ArrayList<Integer> getLines() {
		return lines;
	}
	/**
	 * @param lines the lines to set
	 */
	public void setLines(ArrayList<Integer> lines) {
		this.lines = lines;
	}
	/**
	 * @return the ast
	 */
	public ASTNode getAst() {
		return ast;
	}
	/**
	 * @param ast the ast to set
	 */
	public void setAst(ASTNode ast) {
		this.ast = ast;
	}
	/**
	 * @return the changedLines
	 */
	public ArrayList<Integer> getChangedLines() {
		return changedLines;
	}
	/**
	 * @param changedLines the changedLines to set
	 */
	public void setChangedLines(ArrayList<Integer> changedLines) {
		this.changedLines = changedLines;
	}
	public ArrayList<Integer> getUnchangedLines() {
		return unchangedLines;
	}
	public void setUnchangedLines(ArrayList<Integer> unchangedLines) {
		this.unchangedLines = unchangedLines;
	}
	public HashSet<Integer> getCommentLines() {
		return commentLines;
	}
	public void setCommentLines(HashSet<Integer> commentLines) {
		this.commentLines = commentLines;
	}
	public HashMap<ASTNode, ArrayList<ASTNode>> getTree() {
		return tree;
	}
	public ArrayList<ASTNode> getUnchangedLeaves() {
		return unchangedLeaves;
	}
	public ArrayList<ASTNode> getAllLeaves() {
		return allLeaves;
	}
	/**
	 * @return the fileContent
	 */
	public String getFileContent() {
		return fileContent;
	}
	/**
	 * @param fileContent the fileContent to set
	 */
	public void setFileContent(String fileContent) {
		this.fileContent = fileContent;
	}
	/**
	 * 
	 */
	public void buildLines()
	{
		lines = new ArrayList<Integer>();
		int charCount = -1;
    	do
    	{
    		charCount++;
    		lines.add(charCount);
    		charCount = fileContent.indexOf("\n", charCount);
    	} while (charCount > -1);
		
		ASTParser parser = ASTParser.newParser(AST.JLS3);
    	parser.setSource(fileContent.toCharArray());
    	ast = parser.createAST(null);
	}
	public void buildLinesWoChange()
	{
		int charCount = 0;
		Scanner scanner = new Scanner(fileContent);
		String line = "";
    	lines = new ArrayList<Integer>();
    	while ((scanner.hasNextLine())) {
    		line = scanner.nextLine();
    		lines.add(charCount);
    		charCount += line.length() + 1;
    	}
		scanner.close();
	}
	/**
	 * 
	 */
	public void buildBlockCommentLines()
	{
		if(ast == null)
			return;
		int startChar = 0;
		int startLine = 0;
		int endChar = -1;
		int endLine = 0;
		StringBuffer buff = new StringBuffer();
		List comments = ((CompilationUnit)ast).getCommentList();
		if (comments != null) {
			for (int i=0; i < comments.size(); ++i) {
				Comment comment = (Comment) comments.get(i);
				//System.out.println(comment.toString() + "\t" + comment.getStartPosition() + "\t" + comment.getLength());
				startChar = comment.getStartPosition();
				if(startChar < endChar)
					System.err.println("Something wrong");
				buff.append(fileContent.substring(endChar+1, startChar));
				startLine = binSearch(startChar, lines);
				endChar = comment.getStartPosition() + comment.getLength() - 1;
				endLine = binSearch(endChar, lines);
				if(endLine > startLine)
					for(int j = startLine; j <= endLine; j++)
						commentLines.add(j);
				if(comment.isDocComment())
				{
					buff.append("/* ");
					buff.append(fileContent.substring(startChar+3, endChar+1));
				}
				else
					buff.append(fileContent.substring(startChar, endChar+1));
			}
			buff.append(fileContent.substring(endChar + 1));
			fileContent = buff.toString();
			ASTParser parser = ASTParser.newParser(AST.JLS3);
	    	parser.setSource(fileContent.toCharArray());
	    	ast = parser.createAST(null);
		}
	}
	/**
	 * 
	 */
	public void buildUnchangedLines()
	{
		unchangedLines = new ArrayList<Integer>();
		for(int i = 0; i <= lines.size(); i++)
			unchangedLines.add(i);
		if(changedLines != null)
		{
			for(int i = changedLines.size()-1; i >= 0; i--)
			{
				unchangedLines.remove(changedLines.get(i) - 1);
			}
		}
	}
	/**
	 * 
	 */
	public void deleteChangedLinesInComments()
	{
		if(ast == null || changedLines == null)
			return;
		int k = 0;
		int startChar = 0;
		int startLine = 0;
		int endChar = -1;
		int endLine = 0;
		StringBuffer buff = new StringBuffer();
		List comments = ((CompilationUnit)ast).getCommentList();
		if (comments != null) {
			for (int i=0; i < comments.size(); ++i) {
				Comment comment = (Comment) comments.get(i);
				//System.out.println(comment.toString() + "\t" + comment.getStartPosition() + "\t" + comment.getLength());
				startChar = comment.getStartPosition();
				if(startChar < endChar)
					System.err.println("Something wrong");
				buff.append(fileContent.substring(endChar+1, startChar));
				startLine = binSearch(startChar, lines);
				endChar = comment.getStartPosition() + comment.getLength() - 1;
				endLine = binSearch(endChar, lines);
				/*for(int j = startLine; j < endLine; j++)
					buff.append("\n");*/
				if(comment.isDocComment())
				{
					buff.append("/* ");
					buff.append(fileContent.substring(startChar+3, endChar+1));
				}
				else
					buff.append(fileContent.substring(startChar, endChar+1));
				while(k < changedLines.size() && changedLines.get(k) < startLine + 1)
				{
					if(fileContent.substring(lines.get(changedLines.get(k)-1), lines.get(changedLines.get(k))).trim().isEmpty())
						changedLines.remove(k);
					else
						k++;
				}
				if(k < changedLines.size())
				{
					if(changedLines.get(k) == startLine + 1)
					{
						if(lines.get(startLine) == startChar || fileContent.substring(lines.get(startLine), startChar).trim().isEmpty())
							changedLines.remove(k);
					}
					while(k < changedLines.size() && changedLines.get(k) <= endLine + 1)
					{
						if(changedLines.get(k) > startLine + 1)
							changedLines.remove(k);
						else
						{
							int line = changedLines.get(k);
							if(line < lines.size() && fileContent.substring(lines.get(line-1), lines.get(line)).trim().isEmpty())
								changedLines.remove(k);
							else
								k++;
						}
					}
				}
			}
		}
		while(k < changedLines.size()-1)
		{
			int changedLine = changedLines.get(k);
			String line;
			if(changedLine == lines.size())
				line = fileContent.substring(lines.get(changedLine-1));
			else
				line = fileContent.substring(lines.get(changedLine-1), lines.get(changedLine));  
			if(line.trim().isEmpty())
				changedLines.remove(k);
			else
				k++;
		}
		buff.append(fileContent.substring(endChar + 1));
		fileContent = buff.toString();
		buildLines();
	}
	/**
	 * 
	 */
	public void buildTree()
	{
		MapVisitor mapVisitor = new MapVisitor();
    	/*if(unchangedLines.size() == lines.size())
    		System.out.println("No changes made");*/
    	//mapVisitor.setChangedLines(new ArrayList<Integer>(changedLines));
		mapVisitor.setUnchangedLines(new HashSet<Integer>(unchangedLines));
    	mapVisitor.setLines(lines);
    	ast.accept(mapVisitor);
    	tree = new HashMap<ASTNode, ArrayList<ASTNode>>(mapVisitor.getTree());
    	unchangedLeaves = new ArrayList<ASTNode>(mapVisitor.getUnchangedLeaves());
    	allLeaves = new ArrayList<ASTNode>(mapVisitor.getAllLeaves());
	}
	/**
	 * 
	 */
	public void printTree(ASTNode root)
	{
		System.out.println(root.getProperty(MapVisitor.propertyLocationID) + 
				" line: " + root.getProperty(MapVisitor.propertyStartLine) + 
				" status: " + root.getProperty(MapVisitor.propertyStatus) +
				" vector: " + (root.getProperty(MapVectorVisitor.propertyFragment) != null));
		if(root.getProperty(MapVectorVisitor.propertyFragment) == null)
			System.out.println(root + "@" + root.getParent());
		ArrayList<ASTNode> children = tree.get(root);
		if(!children.isEmpty())
			for(ASTNode child : children)
				printTree(child);
	}
	/**
	 * 
	 * @param x
	 * @param al
	 * @return
	 */
	private int binSearch(int x, ArrayList<Integer> al) {
		if (al.size() <= 1) return al.size()-1;
		int low = 0, high = al.size()-1, mid;
		while(low < high) {
			mid = (low + high) / 2;
            if(al.get(mid) < x)
                low = mid + 1;
            else if(al.get(mid) > x)
                high = mid - 1;
            else
                return mid;
        }
		if (al.get(low) > x) return low - 1;
		else return low;
	}
}
