/**
 * 
 */
package util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;

import cleman.Data;
import cleman.Hash;

/**
 * @author hoan
 *
 */
public class FileIO {
	public static PrintStream logStream;
	static {
		try {
			logStream = new PrintStream(new FileOutputStream("log.txt"));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	
	public static String getSimpleFileName(String fileName)
	{
		char separator = '/';
		if(fileName.lastIndexOf('\\') > -1)
			separator = '\\';
		int start = fileName.lastIndexOf(separator) + 1;
		int end = fileName.lastIndexOf('.');
		if(end <= start)
			end = fileName.length();
		
		return fileName.substring(start, end);
	}
	
	/*public static String getSimpleClassName(String className)
	{
		String name = className.substring(className.lastIndexOf('.') + 1);
		int start = 0;
		while (start < name.length() && !Character.isJavaIdentifierPart(name.charAt(start)))
			start++;
		int end = name.length()-1;
		while (end >= 0 && !Character.isJavaIdentifierPart(name.charAt(end)))
			end--;
		if (end < start)
			return null;
		return name.substring(start, end+1);
	}*/
	public static String getSimpleClassName(String className)
	{
		String name = className.substring(className.lastIndexOf('.') + 1);
		return name;
	}
	
	public static String getSVNRepoRootName(String url)
	{
		String name = "";
		int end = url.length() - 1;
		while (url.charAt(end) == '/' && end >= 0)
			end--;
		if (end >= 0)
		{
			int start = url.lastIndexOf('/', end);
			if (start <= end)
				name = url.substring(start+1, end+1);
		}
		
		return name;
	}
	
	public static String[] splitFileName(String fileName)
	{
		char separator = '/';
		if(fileName.lastIndexOf('\\') > -1)
			separator = '\\';
		int start = fileName.lastIndexOf(separator) + 1;
		int end = fileName.lastIndexOf('.');
		if(end <= start)
			end = fileName.length() + 1;
		String[] names = new String[2];
		names[0] = fileName.substring(0, start-1);
		names[1] = fileName.substring(start, end);
		return names;
	}
	
	/*public static String getFileContent(File file)
	{
		StringBuffer strBuf = new StringBuffer();
		try {
			BufferedReader in = new BufferedReader(new FileReader(file));
	    	String line = "";
	    	while ((line = in.readLine()) != null) { 
	    		strBuf.append(line + "\r\n");
	    	}
	    	in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return strBuf.toString();
	}*/
	
	public static String readStringFromFile(String inputFile) {
		try {
			BufferedInputStream in = new BufferedInputStream(new FileInputStream(inputFile));
			byte[] bytes = new byte[(int) new File(inputFile).length()];
			in.read(bytes);
			in.close();
			return new String(bytes);
		}
		catch (Exception e) {
			//e.printStackTrace();
			return null;
		}
	}
	
	public static void writeStringToFile(String string, String outputFile) {
		try {
			BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));
			writer.write(string);
			writer.flush();
			writer.close();
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}
	
	/*
	 * Read/write an object
	 */
	
	public static void writeObjectToFile(Object object, String objectFile, boolean append) {
		try {
			ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(objectFile, append)));
			out.writeObject(object);
			out.flush();
			out.close();
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(0);
		}
	}
	
	public static Object readObjectFromFile(String objectFile) {
		try {
			ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(objectFile)));
			Object object = in.readObject();
			in.close();
			return object;
		}
		catch (Exception e) {
			//e.printStackTrace();
			return null;
		}
	}
	
	public static int countLOC(File file, String extension)
	{
		int numOfLines = 0;
		if(file.isDirectory())
		{
			for(File sub : file.listFiles())
				numOfLines += countLOC(sub, extension);
		}
		else if(file.getName().endsWith("." + extension))
		{
			try {
				BufferedReader in = new BufferedReader(new FileReader(file));
		    	while (in.readLine() != null) {
		    		numOfLines++;
		    	}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		return numOfLines;
	}
	
	public static String getHtmlPageContent(String url, String query, String charset) throws MalformedURLException, IOException
	{
		URLConnection connection = new URL(url + "?" + query).openConnection();
		connection.setRequestProperty("Accept-Charset", charset);
		//System.out.println(connection.getReadTimeout());
		connection.setReadTimeout(10000);
		InputStream response = connection.getInputStream();
		StringBuilder sb = new StringBuilder();
		BufferedInputStream in = new BufferedInputStream(response);
		byte[] bytes = new byte[10000];
		int len = in.read(bytes);
		while (len != -1)
		{
			//System.out.println(len);
			//System.out.println(new String(bytes, 0, len));
			sb.append(new String(bytes, 0, len));
			//Thread.sleep(100);
			len = in.read(bytes);
		}
		in.close();
		//System.out.println(len);
		//System.out.println(sb.toString());
		/*Scanner sc = new Scanner(response);
		while (sc.hasNextLine())
		{
			System.out.println(sc.nextLine());
		}*/
		/*BufferedReader in = new BufferedReader(new InputStreamReader(response));
		String inputLine;
        while ((inputLine = in.readLine()) != null)
        {
            System.out.println(inputLine);
            System.out.println("Here!!!");
        }
        in.close();*/
		return sb.toString();
	}
	
	public static ArrayList<String> getAllFilesInFolder(String folder) {
		ArrayList<String> allFiles = new ArrayList<String>();
		for (File file : new File(folder).listFiles()) {
			if (file.isFile())
			{
				System.out.println(file.getName() + ":" + file.length());
				allFiles.add(file.getPath());
			}
			else
				allFiles.addAll(getAllFilesInFolder(file.getPath()));
		}
		return allFiles;
	}

	/*This method will be used to read all data stored on HDD, for hash tables, CG, and vectors*/
	public static void loadData(String inputDataPath)
	{
		try 
		{
			Data data = (Data)loadObjectFromFile(inputDataPath + "/clemanX.dat");
			Hash.b = data.getB();
			Hash.ran = data.getRan();
			Hash.randGaussian = data.getRandGaussian();
			//Hash.aGaussian = data.getAGaussian();
			Hash.aGaussian = (double[][][])loadObjectFromFile(inputDataPath + "/aGaussian.dat");
			
			//Fragment.index2Gram = (HashMap<Integer, Integer>)loadObjectFromFile(applicationPath + "\\index2Gram.txt");
			//Fragment.gram2Index = (HashMap<Integer, Integer>)loadObjectFromFile(applicationPath + "\\gram2Index.txt");
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	/*This method will be used to write from a file*/
	public static void saveObjectToFile(Object c, String wFileName)
	{
		File data = new File(wFileName);
		//To remove the previous existed file
		if(data.exists()) data.delete();
		try {
			data.createNewFile();	  
			ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(data));
			out.writeObject(c);
			out.flush();
			out.close();
		}
		catch (IOException e){
			e.printStackTrace();
		}
	}
	public static void saveData(String applicationPath)
	{
		Data data = new Data();
		//data.setAGaussian(Hash.aGaussian);
		data.setB(Hash.b);
		data.setRan(Hash.ran);
		data.setRandGaussian(Hash.randGaussian);
		/*data.setBuckets(Bucket.all);
		data.setFiles(SourceFile.all);
		data.setFragments(Fragment.all);
		data.setGroups(Group.all);
		data.setNextFileID(SourceFile.nextID);
		data.setNextFragmentID(Fragment.nextID);
		data.setNextGroupID(Group.nextID);*/
		
		/*data.setGram2Index(Fragment.gram2Index);
		data.setIndex2Gram(Fragment.index2Gram);*/
		/*int[] grams = new int[Fragment.index2Gram.size()];
		for(int i = 0; i < grams.length; i++)
			grams[i] = Fragment.index2Gram.get(i);
		data.setGrams(grams);*/
		
		saveObjectToFile(data, applicationPath + "\\clemanX.dat");
		saveObjectToFile(Hash.aGaussian, applicationPath + "\\aGaussian.dat");
		//saveObjectToFile(Fragment.index2Gram, applicationPath + "\\index2Gram.txt");
		//saveObjectToFile(Fragment.gram2Index, applicationPath + "\\gram2Index.txt");
	}
	/*This method will be used to read from a file*/
	public static Object loadObjectFromFile(String rFileName)
	{	   
		Object result = null;
		File data = new File(rFileName);
		if (data.exists()) {
			try{
				ObjectInputStream in = new ObjectInputStream(new FileInputStream(data)); 			     
				result =  in.readObject();
				in.close();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		return result;
	}
	
}
